@@ -95,11 +95,31 @@ typedef uint32_t svBitVecVal;
  */
 #define SV_MASK(N) (~(-1 << (N)))
 
-#define SV_GET_UNSIGNED_BITS(VALUE, N) ((N) == 32 ? (VALUE) : ((VALUE)&SV_MASK(N)))
+#define SV_GET_UNSIGNED_BITS(VALUE, N) ((N) == 32 ? (VALUE) : ((VALUE) & SV_MASK(N)))
 
 #define SV_GET_SIGNED_BITS(VALUE, N) \
     ((N) == 32 ? (VALUE)             \
-               : (((VALUE) & (1 << (N))) ? ((VALUE) | ~SV_MASK(N)) : ((VALUE)&SV_MASK(N))))
+               : (((VALUE) & (1 << (N))) ? ((VALUE) | ~SV_MASK(N)) : ((VALUE) & SV_MASK(N))))
+
+#ifndef VPI_TIME
+#    define VPI_TIME
+typedef struct t_vpi_time {
+    int32_t type;
+    uint32_t high, low;
+    double real;
+} s_vpi_time, *p_vpi_time;
+
+#    define vpiScaledRealTime 1
+#    define vpiSimTime 2
+#    define vpiSuppressTime 3
+#endif
+
+/* time value */
+typedef s_vpi_time svTimeVal;
+
+/* time value types */
+#define sv_scaled_real_time vpiScaledRealTime
+#define sv_sim_time vpiSimTime
 
 /*
  * Implementation-dependent representation.
@@ -179,6 +199,7 @@ XXTERN int svSizeOfArray(const svOpenArrayHandle);
  * or NULL if index outside the range or null pointer
  */
 XXTERN void* svGetArrElemPtr(const svOpenArrayHandle, int indx1, ...);
+
 /* specialized versions for 1-, 2- and 3-dimensional arrays: */
 XXTERN void* svGetArrElemPtr1(const svOpenArrayHandle, int indx1);
 XXTERN void* svGetArrElemPtr2(const svOpenArrayHandle, int indx1, int indx2);
@@ -190,6 +211,7 @@ XXTERN void* svGetArrElemPtr3(const svOpenArrayHandle, int indx1, int indx2, int
  * The user is responsible for allocating an array to hold the
  * canonical representation.
  */
+
 /* s=source, d=destination */
 /* From user space into simulator storage */
 XXTERN void svPutBitArrElemVecVal(const svOpenArrayHandle d, const svBitVecVal* s, int indx1, ...);
@@ -218,6 +240,7 @@ XXTERN void svGetLogicArrElem2VecVal(svLogicVecVal* d, const svOpenArrayHandle s
                                      int indx2);
 XXTERN void svGetLogicArrElem3VecVal(svLogicVecVal* d, const svOpenArrayHandle s, int indx1,
                                      int indx2, int indx3);
+
 XXTERN svBit svGetBitArrElem(const svOpenArrayHandle s, int indx1, ...);
 XXTERN svBit svGetBitArrElem1(const svOpenArrayHandle s, int indx1);
 XXTERN svBit svGetBitArrElem2(const svOpenArrayHandle s, int indx1, int indx2);
@@ -248,7 +271,7 @@ XXTERN svScope svGetScope(void);
 
 /*
  * Set context for subsequent export function execution.
- * This function must be called before calling an export function, unless
+ * This function shall be called before calling an export function, unless
  * the export function is called while executing an import function. In that
  * case the export function shall inherit the scope of the surrounding import
  * function. This is known as the "default scope".
@@ -268,7 +291,7 @@ XXTERN svScope svGetScopeFromName(const char* scopeName);
 
 /*
  * Store an arbitrary user data pointer for later retrieval by svGetUserData()
- * The userKey is generated by the user. It must be guaranteed by the user to
+ * The userKey is generated by the user. It needs to be guaranteed by the user to
  * be unique from all other userKey's for all unique data storage requirements
  * It is recommended that the address of static functions or variables in the
  * user's C code be used as the userKey.
@@ -300,31 +323,55 @@ XXTERN void* svGetUserData(const svScope scope, void* userKey);
  * modified. Whether this information is available or not is implementation-
  * specific. Note that the string provided (if any) is owned by the SV
  * implementation and is valid only until the next call to any SV function.
- * Applications must not modify this string or free it
+ * Applications shall not modify this string or free it.
  */
 XXTERN int svGetCallerInfo(const char** fileName, int* lineNumber);
 
 /*
  * Returns 1 if the current execution thread is in the disabled state.
- * Disable protocol must be adhered to if in the disabled state.
+ * Disable protocol shall be adhered to if in the disabled state.
  */
 XXTERN int svIsDisabledState(void);
 
 /*
  * Imported functions call this API function during disable processing to
  * acknowledge that they are correctly participating in the DPI disable protocol.
- * This function must be called before returning from an imported function that is
+ * This function shall be called before returning from an imported function that is
  * in the disabled state.
  */
 XXTERN void svAckDisabledState(void);
 
 /*
-**********************************************************
-* DEPRECATED PORTION OF FILE STARTS FROM HERE.
-* IEEE-1800-compliant tools may not provide
-* support for the following functionality.
-**********************************************************
-*/
+ * Retrieve the current simulation time, scaled to the time unit of the scope.
+ * If scope is NULL, then time is scaled to the simulation time unit.
+ * It is an error to call svGetTime() with an invalid svScope.
+ * This function returns -1 for all error cases, 0 upon success.
+ */
+XXTERN int svGetTime(const svScope scope, svTimeVal* time);
+
+/*
+ * Retrieve the time unit for scope.
+ * If scope is NULL, then simulation time unit is retrieved.
+ * It is an error to call svGetTimeUnit() with an invalid svScope.
+ * This function returns -1 for all error cases, 0 upon success.
+ */
+XXTERN int svGetTimeUnit(const svScope scope, int32_t* time_unit);
+
+/*
+ * Retrieve the time precision for scope.
+ * If scope is NULL, then simulation time unit is retrieved.
+ * It is an error to call svGetTimePrecision() with an invalid svScope.
+ * This function returns -1 for all error cases, 0 upon success.
+ */
+XXTERN int svGetTimePrecision(const svScope scope, int32_t* time_precision);
+
+/*
+ **********************************************************
+ * DEPRECATED PORTION OF FILE STARTS FROM HERE.
+ * IEEE-1800-compliant tools may not provide
+ * support for the following functionality.
+ **********************************************************
+ */
 
 /*
  * Canonical representation of packed arrays
@@ -402,8 +449,9 @@ XXTERN void svPutPartSelectLogic(svLogicPackedArrRef d, const svLogicVec32 s, in
 /*
  * Functions for open array translation between simulator and canonical
  * representations. These functions copy the whole packed array in either
- * direction. The user is responsible for allocating an array in the
- * canonical representation.
+* direction. The user is responsible for allocating an array in the
+* canonical representation.
+
  */
 
 /* s=source, d=destination */
@@ -436,10 +484,10 @@ XXTERN void svGetLogicArrElem3Vec32(svLogicVec32* d, const svOpenArrayHandle s,
                                     int indx2, int indx3);
 
 /*
-**********************************************************
-* DEPRECATED PORTION OF FILE ENDS HERE.
-**********************************************************
-*/
+ **********************************************************
+ * DEPRECATED PORTION OF FILE ENDS HERE.
+ **********************************************************
+ */
 
 #undef DPI_EXTERN
 
