@@ -2,7 +2,8 @@
 //! @file DiagnosticClient.h
 //! @brief Client interface for diagnostic rendering
 //
-// File is under the MIT license; see LICENSE for details
+// SPDX-FileCopyrightText: Michael Popoloski
+// SPDX-License-Identifier: MIT
 //------------------------------------------------------------------------------
 #pragma once
 
@@ -11,7 +12,7 @@
 
 namespace slang {
 
-class DiagnosticClient {
+class SLANG_EXPORT DiagnosticClient {
 public:
     virtual ~DiagnosticClient() = default;
     virtual void report(const ReportedDiagnostic& diagnostic) = 0;
@@ -22,9 +23,9 @@ protected:
     const DiagnosticEngine* engine = nullptr;
     const SourceManager* sourceManager = nullptr;
 
-    void getIncludeStack(BufferID buffer, SmallVector<SourceLocation>& stack) const;
-    string_view getSourceLine(SourceLocation location, size_t col) const;
-    static string_view getSeverityString(DiagnosticSeverity severity);
+    void getIncludeStack(BufferID buffer, SmallVectorBase<SourceLocation>& stack) const;
+    std::string_view getSourceLine(SourceLocation location, size_t col) const;
+    static std::string_view getSeverityString(DiagnosticSeverity severity);
 };
 
 } // namespace slang
