@@ -2,10 +2,12 @@
 //! @file DiagnosticEngine.h
 //! @brief Primary interface for managing diagnostic reporting
 //
-// File is under the MIT license; see LICENSE for details
+// SPDX-FileCopyrightText: Michael Popoloski
+// SPDX-License-Identifier: MIT
 //------------------------------------------------------------------------------
 #pragma once
 
+#include <filesystem>
 #include <memory>
 #include <string>
 #include <typeindex>
@@ -13,6 +15,7 @@
 
 #include "slang/diagnostics/Diagnostics.h"
 #include "slang/util/Hash.h"
+#include "slang/util/TypeTraits.h"
 
 namespace slang {
 
@@ -20,12 +23,12 @@ class DiagArgFormatter;
 class DiagnosticClient;
 class SourceManager;
 
-struct ReportedDiagnostic {
+struct SLANG_EXPORT ReportedDiagnostic {
     const Diagnostic& originalDiagnostic;
-    span<const SourceLocation> expansionLocs;
-    span<const SourceRange> ranges;
+    std::span<const SourceLocation> expansionLocs;
+    std::span<const SourceRange> ranges;
     SourceLocation location;
-    string_view formattedMessage;
+    std::string_view formattedMessage;
     DiagnosticSeverity severity = DiagnosticSeverity::Ignored;
     bool shouldShowIncludeStack = false;
 
@@ -41,7 +44,7 @@ struct ReportedDiagnostic {
 /// - Register one or more DiagnosticClient derived classes with addClient()
 /// - Issue diagnostics by calling issue()
 ///
-class DiagnosticEngine {
+class SLANG_EXPORT DiagnosticEngine {
 public:
     /// Constructs a new diagnostic engine, using the specified source manager
     /// for reporting source-based information from diagnostics.
@@ -120,18 +123,18 @@ public:
     void setMessage(DiagCode code, const std::string& message);
 
     /// Gets the message currently mapped for the given diagnostic.
-    string_view getMessage(DiagCode code) const;
+    std::string_view getMessage(DiagCode code) const;
 
     /// Gets the option string that can be used to refer to a particular diagnostic.
     /// If the diagnostic has no option string provided, this returns an empty string.
-    string_view getOptionName(DiagCode code) const;
+    std::string_view getOptionName(DiagCode code) const;
 
     /// Finds a diagnostic given an option name. If no matching diagnostic is found,
     /// returns an empty diagnostic code.
-    span<const DiagCode> findFromOptionName(string_view optionName) const;
+    std::span<const DiagCode> findFromOptionName(std::string_view optionName) const;
 
     /// Finds the diagnostic group with the given name, if it exists. Otherwise returns nullptr.
-    const DiagGroup* findDiagGroup(string_view name) const;
+    const DiagGroup* findDiagGroup(std::string_view name) const;
 
     /// Clears out all custom mappings for diagnostics, reverting built-ins back to
     /// their defaults and removing all user-specified diagnostics.
@@ -141,18 +144,28 @@ public:
     /// severity type.
     void clearMappings(DiagnosticSeverity severity);
 
+    /// Adds paths for which all warnings will be supressed. This applies to
+    /// all natural warnings, regardless of whether they've been upgraded to an error.
+    std::error_code addIgnorePaths(std::string_view pattern);
+
+    /// Adds paths for which all warnings will be supressed, when looking at the
+    /// original location for macro expansions (i.e. where the macro was originally defined).
+    /// This applies to all natural warnings, regardless of whether they've been upgraded
+    /// to an error.
+    std::error_code addIgnoreMacroPaths(std::string_view pattern);
+
     /// Sets a custom formatter function for the given type. This is used to
     /// provide formatting for diagnostic arguments of a custom type.
     template<typename ForType>
     void setFormatter(std::shared_ptr<DiagArgFormatter> formatter) {
-        formatters[std::type_index(typeid(ForType))] = std::move(formatter);
+        formatters[SLANG_TYPEOF(ForType)] = std::move(formatter);
     }
 
     /// Sets a custom formatter for the given type that should apply by default to
     /// all new DiagnosticEngine instances that get created.
     template<typename ForType>
     static void setDefaultFormatter(std::shared_ptr<DiagArgFormatter> formatter) {
-        defaultFormatters[std::type_index(typeid(ForType))] = std::move(formatter);
+        defaultFormatters[SLANG_TYPEOF(ForType)] = std::move(formatter);
     }
 
     /// Formats the given diagnostic using its arguments and the currently mapped
@@ -165,7 +178,7 @@ public:
     /// Sets diagnostic options from the given option strings, typically from a list of -W
     /// arguments passed to a command line invocation. Any errors encountered while parsing
     /// the options are returned via the diagnostics set.
-    Diagnostics setWarningOptions(span<const std::string> options);
+    Diagnostics setWarningOptions(std::span<const std::string> options);
 
     /// Sets diagnostic options from the `pragma diagnostic entries in all of the various
     /// source files tracked by the engine's source manager. Any errors encountered
@@ -185,60 +198,37 @@ public:
     /// If @a mapOriginalLocations is set to true, the returned source ranges will
     /// be specified in their original textual locations. Otherwise they will
     /// remain as macro locations.
-    void mapSourceRanges(SourceLocation loc, span<const SourceRange> ranges,
-                         SmallVector<SourceRange>& mapped, bool mapOriginalLocations = true) const;
+    void mapSourceRanges(SourceLocation loc, std::span<const SourceRange> ranges,
+                         SmallVectorBase<SourceRange>& mapped,
+                         bool mapOriginalLocations = true) const;
 
     /// A helper method used as a shortcut to turn all of the specified diagnostics into
     /// a human-friendly string. This is mostly intended to be used for things like tests.
-    static std::string reportAll(const SourceManager& sourceManager, span<const Diagnostic> diags);
+    static std::string reportAll(const SourceManager& sourceManager,
+                                 std::span<const Diagnostic> diags);
 
 private:
     // An entry added by `pragma diagnostic at the given source offset which
     // sets a diagnostic to the given severity.
     struct DiagnosticMapping {
         size_t offset = 0;
-        optional<DiagnosticSeverity> severity;
+        std::optional<DiagnosticSeverity> severity;
 
-        DiagnosticMapping(size_t offset, optional<DiagnosticSeverity> severity) noexcept :
+        DiagnosticMapping(size_t offset, std::optional<DiagnosticSeverity> severity) noexcept :
             offset(offset), severity(severity) {}
     };
 
-    optional<DiagnosticSeverity> findMappedSeverity(DiagCode code, SourceLocation location) const;
-    void issueImpl(const Diagnostic& diagnostic, DiagnosticSeverity severity);
+    std::optional<DiagnosticSeverity> findMappedSeverity(DiagCode code,
+                                                         SourceLocation location) const;
+    bool issueImpl(const Diagnostic& diagnostic, DiagnosticSeverity severity);
 
     template<typename TDirective>
-    void setMappingsFromPragmasImpl(BufferID buffer, span<const TDirective> directives,
+    void setMappingsFromPragmasImpl(BufferID buffer, std::span<const TDirective> directives,
                                     Diagnostics& diags);
 
     // The source manager used to resolve locations into file names.
     const SourceManager& sourceManager;
 
-    // A global mapping from diagnostic to a configured severity it should have.
-    flat_hash_map<DiagCode, DiagnosticSeverity> severityTable;
-
-    // A global mapping from diagnostic to the message it should display.
-    flat_hash_map<DiagCode, std::string> messageTable;
-
-    // A set of buffers for which we have reported an include stack,
-    // so that we don't do it more than once.
-    flat_hash_set<BufferID> reportedIncludeStack;
-
-    // A map from diagnostic -> source file -> list of in-source diagnostic mappings.
-    // These correspond to `pragma diagnostic entries in the source code.
-    flat_hash_map<DiagCode, flat_hash_map<BufferID, std::vector<DiagnosticMapping>>> diagMappings;
-
-    // A list of all registered clients that receive issued diagnostics.
-    std::vector<std::shared_ptr<DiagnosticClient>> clients;
-
-    // A map from typeid to a formatter for that type. Used to register custom
-    // formatters for subsystem-specific types.
-    using FormatterMap = flat_hash_map<std::type_index, std::shared_ptr<DiagArgFormatter>>;
-    mutable FormatterMap formatters;
-
-    // A set of default formatters that will be assigned to each new DiagnosticEngine instance
-    // that gets created. These can later be overridden on a per-instance basis.
-    static FormatterMap defaultFormatters;
-
     // Tracking for the number of errors and warnings we've issued.
     int numErrors = 0;
     int numWarnings = 0;
@@ -254,6 +244,36 @@ private:
     // Tracking for whether we've already issued an error for going over
     // the configured error limit (we only want to do that once).
     bool issuedOverLimitErr = false;
+
+    // A global mapping from diagnostic to a configured severity it should have.
+    flat_hash_map<DiagCode, DiagnosticSeverity> severityTable;
+
+    // A global mapping from diagnostic to the message it should display.
+    flat_hash_map<DiagCode, std::string> messageTable;
+
+    // A set of buffers for which we have reported an include stack,
+    // so that we don't do it more than once.
+    flat_hash_set<BufferID> reportedIncludeStack;
+
+    // A map from diagnostic -> source file -> list of in-source diagnostic mappings.
+    // These correspond to `pragma diagnostic entries in the source code.
+    flat_hash_map<DiagCode, flat_hash_map<BufferID, std::vector<DiagnosticMapping>>> diagMappings;
+
+    // A list of path patterns in which to suppress warnings.
+    std::vector<std::filesystem::path> ignoreWarnPatterns;
+    std::vector<std::filesystem::path> ignoreMacroWarnPatterns;
+
+    // A list of all registered clients that receive issued diagnostics.
+    std::vector<std::shared_ptr<DiagnosticClient>> clients;
+
+    // A map from type_index to a formatter for that type. Used to register custom
+    // formatters for subsystem-specific types.
+    using FormatterMap = flat_hash_map<SLANG_TYPEINDEX, std::shared_ptr<DiagArgFormatter>>;
+    mutable FormatterMap formatters;
+
+    // A set of default formatters that will be assigned to each new DiagnosticEngine instance
+    // that gets created. These can later be overridden on a per-instance basis.
+    static FormatterMap defaultFormatters;
 };
 
 } // namespace slang
