@@ -2,27 +2,28 @@
 //! @file TextDiagnosticClient.h
 //! @brief Diagnostic client that formats to a text string
 //
-// File is under the MIT license; see LICENSE for details
+// SPDX-FileCopyrightText: Michael Popoloski
+// SPDX-License-Identifier: MIT
 //------------------------------------------------------------------------------
 #pragma once
 
+#include <fmt/color.h>
 #include <functional>
 #include <string>
 
 #include "slang/diagnostics/DiagnosticClient.h"
 
-namespace fmt {
-inline namespace v9 {
-enum class terminal_color : uint8_t;
-}
-} // namespace fmt
-
 namespace slang {
 
 class FormatBuffer;
-class Symbol;
 
-class TextDiagnosticClient : public DiagnosticClient {
+namespace ast {
+class Symbol;
+}
+
+enum class ShowHierarchyPathOption { Auto, Always, Never };
+
+class SLANG_EXPORT TextDiagnosticClient : public DiagnosticClient {
 public:
     fmt::terminal_color noteColor;
     fmt::terminal_color warningColor;
@@ -42,7 +43,7 @@ public:
     void showOptionName(bool show) { includeOptionName = show; }
     void showIncludeStack(bool show) { includeFileStack = show; }
     void showMacroExpansion(bool show) { includeExpansion = show; }
-    void showHierarchyInstance(bool show) { includeHierarchy = show; }
+    void showHierarchyInstance(ShowHierarchyPathOption option) { includeHierarchy = option; }
 
     template<typename TFunc>
     void setSymbolPathCB(TFunc&& func) {
@@ -69,14 +70,15 @@ private:
     bool includeOptionName = true;
     bool includeFileStack = true;
     bool includeExpansion = true;
-    bool includeHierarchy = true;
+    ShowHierarchyPathOption includeHierarchy = ShowHierarchyPathOption::Auto;
 
-    using SymbolPathCB = std::function<std::string(const Symbol&)>;
+    using SymbolPathCB = std::function<std::string(const ast::Symbol&)>;
     SymbolPathCB symbolPathCB;
     static SymbolPathCB defaultSymbolPathCB;
 
-    void formatDiag(SourceLocation loc, span<const SourceRange> ranges, DiagnosticSeverity severity,
-                    string_view message, string_view optionName);
+    void formatDiag(SourceLocation loc, std::span<const SourceRange> ranges,
+                    DiagnosticSeverity severity, std::string_view message,
+                    std::string_view optionName);
 };
 
 } // namespace slang
