@@ -2,13 +2,17 @@
 //! @file SVInt.h
 //! @brief Arbitrary precision integer support
 //
-// File is under the MIT license; see LICENSE for details
+// SPDX-FileCopyrightText: Michael Popoloski
+// SPDX-License-Identifier: MIT
 //-----------------------------------------------------------------------------
 #pragma once
 
-#include <ostream>
+#include <bit>
+#include <climits>
+#include <concepts>
+#include <iosfwd>
+#include <optional>
 
-#include "slang/numeric/MathUtils.h"
 #include "slang/util/SmallVector.h"
 
 namespace slang {
@@ -25,7 +29,7 @@ using bitwidth_t = uint32_t;
 // clang-format on
 
 /// Specifies the base of an integer (for converting to/from a string)
-ENUM_SIZED(LiteralBase, uint8_t, LB)
+SLANG_ENUM_SIZED(LiteralBase, uint8_t, LB)
 #undef LB
 
 bool literalBaseFromChar(char base, LiteralBase& result);
@@ -33,7 +37,7 @@ bool literalBaseFromChar(char base, LiteralBase& result);
 /// Represents a single 4-state bit. The usual bit values of 0 and 1 are
 /// augmented with X (unknown) and Z (high impedance). Both X and Z are
 /// considered "unknown" for most computation purposes.
-struct logic_t {
+struct SLANG_EXPORT logic_t {
     // limited from 0 to 15, plus x or z
     uint8_t value;
 
@@ -94,12 +98,12 @@ struct logic_t {
         return !isUnknown() && value != 0;
     }
 
-    friend bool exactlyEqual(logic_t lhs, logic_t rhs) {
+    SLANG_EXPORT friend bool exactlyEqual(logic_t lhs, logic_t rhs) {
         return lhs.value == rhs.value;
     }
 
     char toChar() const;
-    friend std::ostream& operator<<(std::ostream& os, const logic_t& rhs);
+    SLANG_EXPORT friend std::ostream& operator<<(std::ostream& os, const logic_t& rhs);
 
     static constexpr uint8_t X_VALUE = 1 << 7;
     static constexpr uint8_t Z_VALUE = 1 << 6;
@@ -110,7 +114,7 @@ struct logic_t {
 /// POD base class for SVInt that contains all data members. The purpose of this
 /// is so that other types can manage the backing memory for SVInts with really
 /// large bit widths.
-class SVIntStorage {
+class SLANG_EXPORT SVIntStorage {
 public:
     SVIntStorage() : val(0), bitWidth(1), signFlag(false), unknownFlag(false) {}
     SVIntStorage(bitwidth_t bits, bool signFlag, bool unknownFlag) :
@@ -145,7 +149,7 @@ public:
 /// words are allocated adjacent in memory. The bits in these extra words indicate whether the
 /// corresponding bits in the low words are unknown or normal.
 ///
-class SVInt : SVIntStorage {
+class SLANG_EXPORT SVInt : SVIntStorage {
 public:
     /// Simple default constructor for convenience, results in a 1 bit zero value.
     SVInt() {}
@@ -159,32 +163,37 @@ public:
     }
 
     /// Construct from a given integer value. Uses only the bits necessary to hold the value.
-    template<typename T, typename = std::enable_if_t<std::is_integral_v<T> || std::is_enum_v<T>>>
+    template<std::integral T>
     SVInt(T value) : SVIntStorage(0, false, false) {
         val = (uint64_t)value;
-        if constexpr (IsSignedHelper<T>::type::value) {
+        if constexpr (std::is_signed_v<T>) {
             signFlag = true;
             if (value < 0)
-                bitWidth = bitwidth_t(64 - slang::countLeadingOnes64(val) + 1);
+                bitWidth = bitwidth_t(64 - std::countl_one(val) + 1);
             else
-                bitWidth = bitwidth_t(64 - slang::countLeadingZeros64(val) + 1);
+                bitWidth = bitwidth_t(64 - std::countl_zero(val) + 1);
+        }
+        else if constexpr (std::is_same_v<T, bool>) {
+            bitWidth = 1;
         }
         else {
-            bitWidth = clog2(value + 1);
-            if (bitWidth == 0) {
-                if (value == 0)
-                    bitWidth = 1;
-                else
-                    bitWidth = 64;
-            }
+            bitWidth = (bitwidth_t)std::bit_width(std::make_unsigned_t<T>(value));
         }
+
+        if (bitWidth == 0) {
+            if (value == 0)
+                bitWidth = 1;
+            else
+                bitWidth = 64;
+        }
+
         clearUnusedBits();
     }
 
     /// Construct from a 64-bit value that can be given an arbitrarily large number of bits (sign
     /// extended if necessary).
     SVInt(bitwidth_t bits, uint64_t value, bool isSigned) : SVIntStorage(bits, isSigned, false) {
-        ASSERT(bits > 0 && bits <= MAX_BITS);
+        SLANG_ASSERT(bits > 0 && bits <= MAX_BITS);
         if (isSingleWord())
             val = value;
         else
@@ -193,9 +202,9 @@ public:
     }
 
     /// Construct from numeric data already in memory as a range of bytes.
-    SVInt(bitwidth_t bits, span<const byte> bytes, bool isSigned) :
+    SVInt(bitwidth_t bits, std::span<const byte> bytes, bool isSigned) :
         SVIntStorage(bits, isSigned, false) {
-        ASSERT(bits > 0 && bits <= MAX_BITS);
+        SLANG_ASSERT(bits > 0 && bits <= MAX_BITS);
         initSlowCase(bytes);
     }
 
@@ -238,8 +247,8 @@ public:
 
     /// Checks whether it's possible to convert the value to a simple built-in
     /// integer type and if so returns it.
-    template<typename T, typename = std::enable_if_t<std::is_integral_v<T> || std::is_enum_v<T>>>
-    optional<T> as() const {
+    template<std::integral T>
+    std::optional<T> as() const {
         bitwidth_t bits = getMinRepresentedBits();
         if (unknownFlag || bits > sizeof(T) * CHAR_BIT)
             return std::nullopt;
@@ -295,11 +304,15 @@ public:
     static SVInt createFillZ(bitwidth_t bitWidth, bool isSigned);
 
     [[nodiscard]] size_t hash() const;
-    void writeTo(SmallVector<char>& buffer, LiteralBase base) const;
-    void writeTo(SmallVector<char>& buffer, LiteralBase base, bool includeBase,
+    void writeTo(SmallVectorBase<char>& buffer, LiteralBase base,
+                 bitwidth_t abbreviateThresholdBits = DefaultStringAbbreviationThresholdBits) const;
+    void writeTo(SmallVectorBase<char>& buffer, LiteralBase base, bool includeBase,
                  bitwidth_t abbreviateThresholdBits = MAX_BITS) const;
-    std::string toString() const;
-    std::string toString(LiteralBase base) const;
+    std::string toString(
+        bitwidth_t abbreviateThresholdBits = DefaultStringAbbreviationThresholdBits,
+        bool exactUnknowns = false) const;
+    std::string toString(LiteralBase base, bitwidth_t abbreviateThresholdBits =
+                                               DefaultStringAbbreviationThresholdBits) const;
     std::string toString(LiteralBase base, bool includeBase,
                          bitwidth_t abbreviateThresholdBits = MAX_BITS) const;
 
@@ -347,7 +360,7 @@ public:
     /// unknown values, so make sure you know what you're doing with it.
     bitwidth_t countLeadingZeros() const {
         if (isSingleWord())
-            return slang::countLeadingZeros64(val) - (BITS_PER_WORD - bitWidth);
+            return bitwidth_t(std::countl_zero(val)) - (BITS_PER_WORD - bitWidth);
         return countLeadingZerosSlowCase();
     }
 
@@ -355,7 +368,7 @@ public:
     /// unknown values, so make sure you know what you're doing with it.
     bitwidth_t countLeadingOnes() const {
         if (isSingleWord())
-            return slang::countLeadingOnes64(val << (BITS_PER_WORD - bitWidth));
+            return bitwidth_t(std::countl_one(val << (BITS_PER_WORD - bitWidth)));
         return countLeadingOnesSlowCase();
     }
 
@@ -496,15 +509,15 @@ public:
 
     /// Constructs from a string (in SystemVerilog syntax). This is mostly for convenience;
     /// any errors will assert instead of being handled gracefully.
-    static SVInt fromString(string_view str);
+    static SVInt fromString(std::string_view str);
 
     /// Construct from an array of digits.
     static SVInt fromDigits(bitwidth_t bits, LiteralBase base, bool isSigned, bool anyUnknown,
-                            span<logic_t const> digits);
+                            std::span<logic_t const> digits);
 
     /// Construct from a floating point value.
-    static SVInt fromDouble(bitwidth_t bits, double value, bool isSigned);
-    static SVInt fromFloat(bitwidth_t bits, float value, bool isSigned);
+    static SVInt fromDouble(bitwidth_t bits, double value, bool isSigned, bool round = true);
+    static SVInt fromFloat(bitwidth_t bits, float value, bool isSigned, bool round = true);
 
     /// Evaluates a conditional expression; i.e. condition ? left : right
     static SVInt conditional(const SVInt& condition, const SVInt& lhs, const SVInt& rhs);
@@ -517,26 +530,26 @@ public:
     static logic_t logicalEquiv(const SVInt& lhs, const SVInt& rhs);
 
     /// Concatenates one or more integers into one output integer.
-    static SVInt concat(span<SVInt const> operands);
+    static SVInt concat(std::span<SVInt const> operands);
 
     /// Stream formatting operator. Guesses a nice base to use and writes the string representation
     /// into the stream.
-    friend std::ostream& operator<<(std::ostream& os, const SVInt& rhs);
+    SLANG_EXPORT friend std::ostream& operator<<(std::ostream& os, const SVInt& rhs);
 
     /// Stricter equality, taking into account unknown bits.
-    friend bool exactlyEqual(const SVInt& lhs, const SVInt& rhs);
+    SLANG_EXPORT friend bool exactlyEqual(const SVInt& lhs, const SVInt& rhs);
 
     /// Wildcard based equality, with unknown bits as wildcards.
     /// This method only looks for wildcard bits on the rhs, as needed by the conditional operator.
-    friend logic_t condWildcardEqual(const SVInt& lhs, const SVInt& rhs);
+    SLANG_EXPORT friend logic_t condWildcardEqual(const SVInt& lhs, const SVInt& rhs);
 
     /// Wildcard based equality, with unknown bits as wildcards.
     /// This method implements matching as required by casex statements.
-    friend bool caseXWildcardEqual(const SVInt& lhs, const SVInt& rhs);
+    SLANG_EXPORT friend bool caseXWildcardEqual(const SVInt& lhs, const SVInt& rhs);
 
     /// Wildcard based equality, with Z bits as wildcards.
     /// This method implements matching as required by casez statements.
-    friend bool caseZWildcardEqual(const SVInt& lhs, const SVInt& rhs);
+    SLANG_EXPORT friend bool caseZWildcardEqual(const SVInt& lhs, const SVInt& rhs);
 
     enum {
         BITS_PER_WORD = sizeof(uint64_t) * CHAR_BIT,
@@ -549,7 +562,7 @@ public:
     static const SVInt One;
 
     /// The default threshold, in bits, to use for abbreviating toString results
-    /// when calling one of the simple toString() methods that doesn't take a user
+    /// when calling one of the simple toString() methods without passing a user
     /// provided threshold.
     static constexpr bitwidth_t DefaultStringAbbreviationThresholdBits = 128;
 
@@ -564,7 +577,7 @@ private:
     // Initialization routines for various cases.
     void initSlowCase(logic_t bit);
     void initSlowCase(uint64_t value);
-    void initSlowCase(span<const byte> bytes);
+    void initSlowCase(std::span<const byte> bytes);
     void initSlowCase(const SVIntStorage& other);
 
     uint64_t* getRawData() { return isSingleWord() ? &val : pVal; }
@@ -595,10 +608,10 @@ private:
     static constexpr uint32_t whichBit(bitwidth_t bitIndex) { return bitIndex % BITS_PER_WORD; }
     static constexpr uint64_t maskBit(bitwidth_t bitIndex) { return 1ULL << whichBit(bitIndex); }
 
-    static SVInt fromDecimalDigits(bitwidth_t bits, bool isSigned, span<logic_t const> digits);
+    static SVInt fromDecimalDigits(bitwidth_t bits, bool isSigned, std::span<logic_t const> digits);
 
     static SVInt fromPow2Digits(bitwidth_t bits, bool isSigned, bool anyUnknown, uint32_t radix,
-                                uint32_t shift, span<logic_t const> digits);
+                                uint32_t shift, std::span<logic_t const> digits);
 
     // Split an integer's data into 32-bit words.
     static void splitWords(const SVInt& value, uint32_t* dest, uint32_t numWords);
@@ -625,16 +638,6 @@ private:
         uint32_t value = (bitWidth + BITS_PER_WORD - 1) / BITS_PER_WORD;
         return unknown ? value * 2 : value;
     }
-
-    template<typename T, typename = void>
-    struct IsSignedHelper {
-        using type = std::is_signed<T>;
-    };
-
-    template<typename T>
-    struct IsSignedHelper<T, std::enable_if_t<std::is_enum_v<T>>> {
-        using type = std::is_signed<std::underlying_type_t<T>>;
-    };
 };
 
 inline logic_t operator||(const SVInt& lhs, logic_t rhs) {
@@ -675,7 +678,7 @@ inline logic_t SVInt::logicalEquiv(const SVInt& lhs, const SVInt& rhs) {
 }
 
 /// Returns the ceiling of the log_2 of the value. If value is zero, returns zero.
-inline uint32_t clog2(const SVInt& v) {
+inline SLANG_EXPORT uint32_t clog2(const SVInt& v) {
     if (v == 0)
         return 0;
     return v.getBitWidth() - (v - SVInt::One).countLeadingZeros();
@@ -684,7 +687,7 @@ inline uint32_t clog2(const SVInt& v) {
 inline namespace literals {
 
 inline SVInt operator""_si(const char* str, size_t size) {
-    return SVInt::fromString(string_view(str, size));
+    return SVInt::fromString(std::string_view(str, size));
 }
 
 } // namespace literals
