@@ -2,23 +2,28 @@
 //! @file NumberParser.h
 //! @brief Helper type to parse numeric literals
 //
-// File is under the MIT license; see LICENSE for details
+// SPDX-FileCopyrightText: Michael Popoloski
+// SPDX-License-Identifier: MIT
 //------------------------------------------------------------------------------
 #pragma once
 
+#include <cmath>
+
 #include "slang/diagnostics/Diagnostics.h"
 #include "slang/diagnostics/NumericDiags.h"
 #include "slang/numeric/SVInt.h"
 #include "slang/parsing/Token.h"
 #include "slang/syntax/SyntaxFacts.h"
+#include "slang/text/CharInfo.h"
 #include "slang/text/SourceLocation.h"
+#include "slang/util/LanguageVersion.h"
 #include "slang/util/SmallVector.h"
 
-namespace slang {
+namespace slang::parsing {
 
 class NumberParser {
 public:
-    NumberParser(Diagnostics& diagnostics, BumpAllocator& alloc);
+    NumberParser(Diagnostics& diagnostics, BumpAllocator& alloc, LanguageVersion languageVersion);
 
     struct IntResult {
         Token size;
@@ -26,23 +31,24 @@ public:
         Token value;
         bool isSimple = true;
 
-        static IntResult simple(Token value) { return { Token(), Token(), value, true }; }
+        static IntResult simple(Token value) { return {Token(), Token(), value, true}; }
 
         static IntResult vector(Token size, Token base, Token value) {
-            return { size, base, value, false };
+            return {size, base, value, false};
         }
     };
 
     template<typename TStream>
     IntResult parseSimpleInt(TStream& stream) {
         auto token = stream.expect(TokenKind::IntegerLiteral);
-        if (token.intValue() > INT32_MAX)
-            reportIntOverflow(token);
+        checkIntOverflow(token);
         return IntResult::simple(token);
     }
 
-    template<typename TStream>
+    template<typename TStream, bool RequireSameLine = false>
     IntResult parseInteger(TStream& stream) {
+        const bool isNegated = stream.getLastConsumed().kind == TokenKind::Minus;
+
         Token sizeToken;
         Token baseToken;
 
@@ -52,21 +58,33 @@ public:
             startVector(baseToken, Token());
         }
         else {
-            if (!stream.peek(TokenKind::IntegerBase)) {
-                if (token.intValue() > INT32_MAX)
-                    reportIntOverflow(token);
+            auto createSimple = [&] {
+                checkIntOverflow(token);
                 return IntResult::simple(token);
+            };
+
+            if constexpr (RequireSameLine) {
+                if (!stream.peekSameLine())
+                    return createSimple();
             }
 
+            if (!stream.peek(TokenKind::IntegerBase))
+                return createSimple();
+
             sizeToken = token;
             baseToken = stream.consume();
             startVector(baseToken, sizeToken);
         }
 
+        if constexpr (RequireSameLine) {
+            if (!stream.peekSameLine())
+                return reportMissingDigits(sizeToken, baseToken, Token());
+        }
+
         // At this point we expect to see vector digits, but they could be split out into other
         // token types because of hex literals.
         auto first = stream.peek();
-        if (!SyntaxFacts::isPossibleVectorDigit(first.kind))
+        if (!syntax::SyntaxFacts::isPossibleVectorDigit(first.kind))
             return reportMissingDigits(sizeToken, baseToken, first);
 
         int count = 0;
@@ -91,10 +109,15 @@ public:
                 break;
             }
 
-            next = stream.peek();
-        } while (SyntaxFacts::isPossibleVectorDigit(next.kind) && next.trivia().empty());
+            if constexpr (RequireSameLine) {
+                if (!stream.peekSameLine())
+                    break;
+            }
 
-        return IntResult::vector(sizeToken, baseToken, finishValue(first, count == 1));
+            next = stream.peek();
+        } while (syntax::SyntaxFacts::isPossibleVectorDigit(next.kind) && next.trivia().empty());
+
+        return IntResult::vector(sizeToken, baseToken, finishValue(first, count == 1, isNegated));
     }
 
     template<typename TStream>
@@ -107,18 +130,51 @@ public:
                     << real_t(std::numeric_limits<double>::denorm_min());
             }
             else {
-                ASSERT(!std::isfinite(literal.realValue()));
+                SLANG_ASSERT(!std::isfinite(literal.realValue()));
                 addDiag(diag::RealLiteralOverflow, literal.location())
                     << real_t(std::numeric_limits<double>::max());
             }
         }
+
+        // Enforce extra rules about where digits and underscores
+        // are allowed to go in a real literal.
+        auto raw = literal.rawText();
+        auto len = raw.length();
+        for (size_t i = 0; i < len; i++) {
+            char c = raw[i];
+            size_t j = i + 1;
+
+            auto checkForDigit = [&](DiagCode missingCode) {
+                if (j == len || !isDecimalDigit(raw[j])) {
+                    auto code = (j < len && raw[j] == '_') ? diag::DigitsLeadingUnderscore
+                                                           : missingCode;
+                    addDiag(code, literal.location() + j);
+                }
+            };
+
+            if (c == '.') {
+                checkForDigit(diag::MissingFractionalDigits);
+            }
+            else if (c == 'e' || c == 'E') {
+                if (j < len && (raw[j] == '+' || raw[j] == '-'))
+                    j++;
+
+                checkForDigit(diag::MissingExponentDigits);
+            }
+        }
+
         return literal;
     }
 
 private:
+    void checkIntOverflow(Token token) {
+        if (languageVersion < LanguageVersion::v1800_2023 && token.intValue() > INT32_MAX)
+            reportIntOverflow(token);
+    }
+
     void startVector(Token baseToken, Token sizeToken);
     int append(Token token, bool isFirst);
-    Token finishValue(Token firstToken, bool singleToken);
+    Token finishValue(Token firstToken, bool singleToken, bool isNegated);
     void addDigit(logic_t digit, int maxValue);
     Diagnostic& addDiag(DiagCode code, SourceLocation location);
     IntResult reportMissingDigits(Token sizeToken, Token baseToken, Token first);
@@ -126,6 +182,7 @@ private:
 
     bitwidth_t sizeBits = 0;
     LiteralBase literalBase = LiteralBase::Binary;
+    LanguageVersion languageVersion;
     SourceLocation firstLocation;
     bool signFlag = false;
     bool hasUnknown = false;
@@ -133,8 +190,8 @@ private:
     SVInt decimalValue;
     Diagnostics& diagnostics;
     BumpAllocator& alloc;
-    SmallVectorSized<logic_t, 16> digits;
-    SmallVectorSized<char, 64> text;
+    SmallVector<logic_t> digits;
+    SmallVector<char> text;
 };
 
-} // namespace slang
+} // namespace slang::parsing
