@@ -2,36 +2,42 @@
 //! @file SyntaxPrinter.h
 //! @brief Support for printing syntax nodes and tokens
 //
-// File is under the MIT license; see LICENSE for details
+// SPDX-FileCopyrightText: Michael Popoloski
+// SPDX-License-Identifier: MIT
 //------------------------------------------------------------------------------
 #pragma once
 
 #include <string>
 
-#include "slang/syntax/SyntaxNode.h"
+#include "slang/parsing/Token.h"
 
 namespace slang {
+class SourceManager;
+}
 
+namespace slang::syntax {
+
+class SyntaxNode;
 class SyntaxTree;
 
 /// Provides support for printing tokens, trivia, or whole syntax trees
 /// back to source code.
-class SyntaxPrinter {
+class SLANG_EXPORT SyntaxPrinter {
 public:
     SyntaxPrinter() = default;
     explicit SyntaxPrinter(const SourceManager& sourceManager);
 
     /// Append raw text to the buffer.
     /// @return a reference to this object, to allow chaining additional method calls.
-    SyntaxPrinter& append(string_view text);
+    SyntaxPrinter& append(std::string_view text);
 
     /// Print the provided @a trivia to the internal buffer.
     /// @return a reference to this object, to allow chaining additional method calls.
-    SyntaxPrinter& print(Trivia trivia);
+    SyntaxPrinter& print(parsing::Trivia trivia);
 
     /// Print the provided @a token to the internal buffer.
     /// @return a reference to this object, to allow chaining additional method calls.
-    SyntaxPrinter& print(Token token);
+    SyntaxPrinter& print(parsing::Token token);
 
     /// Print the provided @a node to the internal buffer.
     /// @return a reference to this object, to allow chaining additional method calls.
@@ -110,4 +116,4 @@ private:
     bool squashNewlines = true;
 };
 
-} // namespace slang
+} // namespace slang::syntax
