@@ -2,46 +2,68 @@
 //! @file String.h
 //! @brief Various string utilities
 //
-// File is under the MIT license; see LICENSE for details
+// SPDX-FileCopyrightText: Michael Popoloski
+// SPDX-License-Identifier: MIT
 //------------------------------------------------------------------------------
 #pragma once
 
+#include <filesystem>
+#include <optional>
+
 #include "slang/util/SmallVector.h"
 #include "slang/util/Util.h"
 
 namespace slang {
 
-/// Converts a span of characters into a string_view.
-inline string_view toStringView(span<char> text) {
-    return string_view(text.data(), text.size());
+/// Converts a span of characters into a std::string_view.
+inline std::string_view toStringView(std::span<char> text) {
+    return std::string_view(text.data(), text.size());
 }
 
 /// Converts the provided @a value into a string, storing it into @a buffer.
-void uintToStr(SmallVector<char>& buffer, uint32_t value);
+void uintToStr(SmallVectorBase<char>& buffer, uint32_t value);
 
 /// Converts the provided @a value into a string, storing it into @a buffer.
-void uintToStr(SmallVector<char>& buffer, uint64_t value);
+void uintToStr(SmallVectorBase<char>& buffer, uint64_t value);
 
 /// Parses the provided @a str in the specified @a base into a signed integer.
 /// If @a pos is non-null, it will be set to point to the first character in the
 /// source string that was not part of the conversion.
 ///
 /// @return the converted value, or nullopt if conversion fails (invalid string, overflow, etc).
-optional<int32_t> strToInt(string_view str, size_t* pos = nullptr, int base = 10);
+std::optional<int32_t> strToInt(std::string_view str, size_t* pos = nullptr, int base = 10);
 
 /// Parses the provided @a str in the specified @a base into an unsigned integer.
 /// If @a pos is non-null, it will be set to point to the first character in the
 /// source string that was not part of the conversion.
 ///
 /// @return the converted value, or nullopt if conversion fails (invalid string, overflow, etc).
-optional<uint32_t> strToUInt(string_view str, size_t* pos = nullptr, int base = 10);
+std::optional<uint32_t> strToUInt(std::string_view str, size_t* pos = nullptr, int base = 10);
 
 /// Parses the provided @a str into a floating point value.
 /// If @a pos is non-null, it will be set to point to the first character in the
 /// source string that was not part of the conversion.
 ///
 /// @return the converted value, or nullopt if conversion fails (invalid string, overflow, etc).
-optional<double> strToDouble(string_view str, size_t* pos = nullptr);
+std::optional<double> strToDouble(std::string_view str, size_t* pos = nullptr);
+
+/// Converts the provided string to all uppercase characters (assuming ASCII contents).
+/// The string is converted in place.
+void strToUpper(std::string& str);
+
+/// Converts the provided string to all lowercase characters (assuming ASCII contents).
+/// The string is converted in place.
+void strToLower(std::string& str);
+
+/// Converts a character to uppercase (assuming ASCII).
+inline char charToUpper(char c) {
+    return (char)::toupper(c);
+}
+
+/// Converts a character to lowercase (assuming ASCII).
+inline char charToLower(char c) {
+    return (char)::tolower(c);
+}
 
 /// Determines the number of edits to the left string that are required to
 /// change it into the right string.
@@ -50,36 +72,21 @@ optional<double> strToDouble(string_view str, size_t* pos = nullptr);
 ///
 /// If @a maxDistance is >0 and the computed distance is at least that much, give
 /// up and return maxDistance + 1.
-int editDistance(string_view left, string_view right, bool allowReplacements = true,
+int editDistance(std::string_view left, std::string_view right, bool allowReplacements = true,
                  int maxDistance = 0);
 
-// TODO: remove once we have C++20
-inline bool startsWith(string_view str, string_view prefix) {
-    return str.size() >= prefix.size() && str.compare(0, prefix.size(), prefix) == 0;
-};
-inline bool endsWith(string_view str, string_view suffix) {
-    return str.size() >= suffix.size() &&
-           str.compare(str.size() - suffix.size(), suffix.size(), suffix) == 0;
-};
+#if defined(_WIN32)
 
-#if defined(_MSC_VER)
-
-/// Widens the provided UTF8 string into UTF16 wchars.
-std::wstring widen(string_view str);
-
-/// Widens the provided UTF16 string into UTF8.
-std::string narrow(std::wstring_view str);
+/// Gets a string representation of the given path, in UTF-8 encoding.
+inline std::string getU8Str(const std::filesystem::path& path) {
+    return path.string();
+}
 
 #else
 
-/// Widens the provided UTF8 string into UTF16 wchars.
-inline string_view widen(string_view str) {
-    return str;
-}
-
-/// Widens the provided UTF16 string into UTF8.
-inline string_view narrow(string_view str) {
-    return str;
+/// Gets a string representation of the given path, in UTF-8 encoding.
+inline const std::string& getU8Str(const std::filesystem::path& path) {
+    return path.native();
 }
 
 #endif
