@@ -1,5 +1,9 @@
 #!/usr/bin/env python
 # This script generates C++ code for compiler diagnostics.
+#
+# SPDX-FileCopyrightText: Michael Popoloski
+# SPDX-License-Identifier: MIT
+
 import argparse
 import os
 import shlex
@@ -108,7 +112,8 @@ def createheader(path, subsys, diags):
 //! @file {}Diags.h
 //! @brief Generated diagnostic enums for the {} subsystem
 //
-// File is under the MIT license; see LICENSE for details
+// SPDX-FileCopyrightText: Michael Popoloski
+// SPDX-License-Identifier: MIT
 //------------------------------------------------------------------------------
 #pragma once
 
@@ -138,7 +143,8 @@ def createsource(path, diags, groups):
 // DiagCode.cpp
 // Generated diagnostic helpers
 //
-// File is under the MIT license; see LICENSE for details
+// SPDX-FileCopyrightText: Michael Popoloski
+// SPDX-License-Identifier: MIT
 //------------------------------------------------------------------------------
 #include "slang/diagnostics/AllDiags.h"
 
@@ -148,7 +154,7 @@ def createsource(path, diags, groups):
 
 namespace slang {
 
-static const flat_hash_map<DiagCode, std::tuple<string_view, string_view, DiagnosticSeverity, string_view>> data = {
+static const flat_hash_map<DiagCode, std::tuple<std::string_view, std::string_view, DiagnosticSeverity, std::string_view>> data = {
 """
 
     for _, v in sorted(diags.items()):
@@ -159,7 +165,7 @@ static const flat_hash_map<DiagCode, std::tuple<string_view, string_view, Diagno
 
     output += """};
 
-static const flat_hash_map<string_view, std::vector<DiagCode>> optionMap = {
+static const flat_hash_map<std::string_view, std::vector<DiagCode>> optionMap = {
 """
 
     optionMap = {}
@@ -181,7 +187,7 @@ static const flat_hash_map<string_view, std::vector<DiagCode>> optionMap = {
 
     output += """};
 
-static const flat_hash_map<string_view, DiagGroup> groupMap = {
+static const flat_hash_map<std::string_view, DiagGroup> groupMap = {
 """
 
     for g in sorted(groups):
@@ -201,13 +207,13 @@ std::ostream& operator<<(std::ostream& os, DiagCode code) {
     return os;
 }
 
-string_view toString(DiagCode code) {
+std::string_view toString(DiagCode code) {
     if (auto it = data.find(code); it != data.end())
         return std::get<0>(it->second);
     return "<user-diag>"sv;
 }
 
-string_view getDefaultMessage(DiagCode code) {
+std::string_view getDefaultMessage(DiagCode code) {
     if (auto it = data.find(code); it != data.end())
         return std::get<1>(it->second);
     return ""sv;
@@ -219,19 +225,19 @@ DiagnosticSeverity getDefaultSeverity(DiagCode code) {
     return DiagnosticSeverity::Ignored;
 }
 
-string_view getDefaultOptionName(DiagCode code) {
+std::string_view getDefaultOptionName(DiagCode code) {
     if (auto it = data.find(code); it != data.end())
         return std::get<3>(it->second);
     return ""sv;
 }
 
-span<const DiagCode> findDiagsFromOptionName(string_view name) {
+std::span<const DiagCode> findDiagsFromOptionName(std::string_view name) {
     if (auto it = optionMap.find(name); it != optionMap.end())
         return it->second;
     return {};
 }
 
-const DiagGroup* findDefaultDiagGroup(string_view name) {
+const DiagGroup* findDefaultDiagGroup(std::string_view name) {
     if (auto it = groupMap.find(name); it != groupMap.end())
         return &it->second;
     return nullptr;
@@ -258,7 +264,8 @@ def createallheader(path, diags):
 //! @file AllDiags.h
 //! @brief Combined header that includes all subsystem-specific diagnostic headers
 //
-// File is under the MIT license; see LICENSE for details
+// SPDX-FileCopyrightText: Michael Popoloski
+// SPDX-License-Identifier: MIT
 //------------------------------------------------------------------------------
 #pragma once
 
@@ -293,9 +300,11 @@ def createdocs(outDir, inpath, slangBin, diags, groups):
         elif line.startswith("-W"):
             if curropt:
                 exampleMap[curropt[0]] = curropt
-            curropt = [line[2:], "", "", ""]
+            curropt = [line[2:], "", "", "", ""]
         elif line.startswith("```"):
             inexample = True
+        elif line.startswith("@options "):
+            curropt[4] = line[9:]
         else:
             if curropt[1]:
                 curropt[1] += " "
@@ -321,6 +330,9 @@ def createdocs(outDir, inpath, slangBin, diags, groups):
             testPath,
         ]
 
+        if v[4]:
+            args.extend(v[4].split())
+
         result = subprocess.run(
             args, encoding="utf-8", stdout=subprocess.PIPE, stderr=subprocess.STDOUT
         )
