@@ -1,6 +1,11 @@
 #!/usr/bin/env python
 # This script generates C++ source for parse tree syntax nodes from a data file.
+#
+# SPDX-FileCopyrightText: Michael Popoloski
+# SPDX-License-Identifier: MIT
+
 import argparse
+import math
 import os
 
 
@@ -36,6 +41,9 @@ class TypeInfo:
         self.baseInitializers = baseInitializers
 
 
+vowels = ("a", "e", "i", "o", "u", "A", "E", "I", "O", "U")
+
+
 def main():
     parser = argparse.ArgumentParser(description="Diagnostic source generator")
     parser.add_argument("--dir", default=os.getcwd(), help="Output directory")
@@ -48,6 +56,7 @@ def main():
     if args.python_bindings:
         generatePyBindings(args.dir, alltypes)
     else:
+        generateSyntaxClone(args.dir, alltypes, kindmap)
         generateSyntax(args.dir, alltypes, kindmap)
         generateTokenKinds(ourdir, args.dir)
 
@@ -255,7 +264,8 @@ def generateSyntax(builddir, alltypes, kindmap):
 //! @file AllSyntax.h
 //! @brief All generated syntax node data structures
 //
-// File is under the MIT license; see LICENSE for details
+// SPDX-FileCopyrightText: Michael Popoloski
+// SPDX-License-Identifier: MIT
 //------------------------------------------------------------------------------
 #pragma once
 
@@ -266,20 +276,37 @@ def generateSyntax(builddir, alltypes, kindmap):
 // This file contains all parse tree syntax nodes.
 // It is auto-generated by the syntax_gen.py script under the scripts/ directory.
 
-namespace slang {
+namespace slang::syntax {
 
 """
     )
 
+    # Start a documentation header file.
+    docf = open(os.path.join(headerdir, "SyntaxDoc.dox"), "w")
+    docf.write("/** @file */\n\n")
+
     # Write all type definitions.
+    alltypesSaved = alltypes.copy()
     for name, currtype in alltypes.items():
         if name == "SyntaxNode":
             continue
 
-        outf.write("struct {} : public {} {{\n".format(name, currtype.base))
+        outf.write(
+            "struct SLANG_EXPORT {} : public {} {{\n".format(name, currtype.base)
+        )
+
+        article = "an" if name[0] in vowels else "a"
+        docf.write("/** @struct slang::syntax::{}\n".format(name))
+        docf.write(
+            "    @brief Concrete syntax definition for {} {}\n".format(
+                article, name[:-6]
+            )
+        )
 
         for m in currtype.members:
             outf.write("    {} {};\n".format(m[0], m[1]))
+            docf.write("    @var slang::syntax::{}::{}\n".format(name, m[1]))
+            docf.write("    @brief The {} member\n".format(m[1]))
 
         outf.write("\n")
         outf.write("    {}({}) :\n".format(name, currtype.constructorArgs))
@@ -292,6 +319,15 @@ namespace slang {
             )
         )
 
+        docf.write(
+            "    @fn slang::syntax::{}::{}({})\n".format(
+                name, name, currtype.constructorArgs
+            )
+        )
+        docf.write(
+            "    @brief Constructs a new instance of the {} struct\n".format(name)
+        )
+
         # Write constructor body.
         for m in currtype.members:
             if m[0] == "Token":
@@ -315,17 +351,66 @@ namespace slang {
         # Copy constructor (defaulted).
         outf.write("    explicit {}(const {}&) = default;\n\n".format(name, name))
 
+        docf.write(
+            "    @fn slang::syntax::{}::{}(const {}&)\n".format(name, name, name)
+        )
+        docf.write("    @brief Copy constructor\n")
+
+        docf.write("    @fn slang::syntax::{}::isKind\n".format(name))
+        docf.write(
+            "    @brief Returns true if the provided syntax kind is represented by this type\n"
+        )
+
         if not currtype.members and currtype.final == "":
             outf.write("    static bool isKind(SyntaxKind kind);\n")
         else:
             outf.write("    static bool isKind(SyntaxKind kind);\n\n")
 
+            outf.write("    static bool isChildOptional(size_t index);\n")
             outf.write("    TokenOrSyntax getChild(size_t index);\n")
             outf.write("    ConstTokenOrSyntax getChild(size_t index) const;\n")
+            outf.write("    PtrTokenOrSyntax getChildPtr(size_t index);\n")
             outf.write("    void setChild(size_t index, TokenOrSyntax child);\n\n")
-            outf.write("    {}* clone(BumpAllocator& alloc) const;\n".format(name))
+
+            docf.write(
+                "    @fn static bool slang::syntax::{}::isChildOptional(size_t index);\n".format(
+                    name
+                )
+            )
+            docf.write(
+                "    @brief Returns true if child member (token or syntax node) at the provided index within this struct is a nullable pointer\n"
+            )
+            docf.write(
+                "    @fn TokenOrSyntax slang::syntax::{}::getChild(size_t index)\n".format(
+                    name
+                )
+            )
+            docf.write(
+                "    @brief Gets the child member (token or syntax node) at the provided index within this struct\n"
+            )
+            docf.write(
+                "    @fn ConstTokenOrSyntax slang::syntax::{}::getChild(size_t index) const\n".format(
+                    name
+                )
+            )
+            docf.write(
+                "    @brief Gets the child member (token or syntax node) as a pointer at the provided index within this struct\n"
+            )
+            docf.write(
+                "    @fn PtrTokenOrSyntax slang::syntax::{}::getChildPtr(size_t index)\n".format(
+                    name
+                )
+            )
+            docf.write(
+                "    @brief Gets the child member (token or syntax node) at the provided index within this struct\n"
+            )
+            docf.write("    @fn slang::syntax::{}::setChild\n".format(name))
+            docf.write(
+                "    @brief Sets the child member (token or syntax node) at the provided index within this struct\n"
+            )
 
         outf.write("};\n\n")
+        docf.write("*/\n\n")
 
     # Start the source file.
     cppf = open(os.path.join(builddir, "AllSyntax.cpp"), "w")
@@ -334,14 +419,17 @@ namespace slang {
 // AllSyntax.cpp
 // All generated syntax node data structures
 //
-// File is under the MIT license; see LICENSE for details
+// SPDX-FileCopyrightText: Michael Popoloski
+// SPDX-License-Identifier: MIT
 //------------------------------------------------------------------------------
 #include "slang/syntax/AllSyntax.h"
 
+#include <type_traits>
+
 // This file contains all parse tree syntax node generated definitions.
 // It is auto-generated by the syntax_gen.py script under the scripts/ directory.
 
-namespace slang {
+namespace slang::syntax {
 
 size_t SyntaxNode::getChildCount() const {
     switch (kind) {
@@ -358,7 +446,7 @@ size_t SyntaxNode::getChildCount() const {
         cppf.write("        case SyntaxKind::{}: return {};\n".format(k, count))
 
     cppf.write("    }\n")
-    cppf.write("    THROW_UNREACHABLE;\n")
+    cppf.write("    SLANG_UNREACHABLE;\n")
     cppf.write("}\n\n")
 
     # Build a reverse mapping from class types to their syntax kinds.
@@ -404,19 +492,53 @@ size_t SyntaxNode::getChildCount() const {
         cppf.write("}\n\n")
 
         if v.members or v.final != "":
-            for returnType in ("TokenOrSyntax", "ConstTokenOrSyntax"):
+            cppf.write("bool {}::isChildOptional(size_t index) {{\n".format(k))
+            if v.optionalMembers:
+                cppf.write("    switch (index) {\n")
+
+                index = 0
+                for m in v.combinedMembers:
+                    if m[1] in v.optionalMembers:
+                        cppf.write("        case {}: return true;\n".format(index))
+                    index += 1
+
+                cppf.write("        default: return false;\n")
+                cppf.write("    }\n")
+            else:
+                cppf.write("    (void)index;\n")
+                cppf.write("    return false;\n")
+
+            cppf.write("}\n\n")
+
+            for returnType in (
+                "TokenOrSyntax",
+                "ConstTokenOrSyntax",
+                "PtrTokenOrSyntax",
+            ):
                 cppf.write(
-                    "{} {}::getChild(size_t index){} {{\n".format(
-                        returnType, k, "" if returnType == "TokenOrSyntax" else " const"
+                    "{} {}::getChild{}(size_t index){} {{\n".format(
+                        returnType,
+                        k,
+                        ("Ptr" if returnType.startswith("Ptr") else ""),
+                        "" if not returnType.startswith("Const") else " const",
                     )
                 )
 
+                returnPointer = returnType == "PtrTokenOrSyntax"
+
                 if v.combinedMembers:
                     cppf.write("    switch (index) {\n")
 
                     index = 0
                     for m in v.combinedMembers:
-                        addr = "&" if m[1] in v.pointerMembers else ""
+                        addr = ""
+                        if returnPointer:
+                            if m[0] == "Token" or (m[1] in v.pointerMembers):
+                                addr = "&"
+                        elif m[1] in v.pointerMembers:
+                            addr = "&"
+
+                        # addr = "&" if  != (returnPointer and not (m[1] in v.notNullMembers)) else ""
                         get = ".get()" if m[1] in v.notNullMembers else ""
                         cppf.write(
                             "        case {}: return {}{}{};\n".format(
@@ -452,10 +574,12 @@ size_t SyntaxNode::getChildCount() const {
                         )
                     else:
                         cppf.write(
-                            "{} = &child.node()->as<{}>(); return;\n".format(m[1], m[2])
+                            "{} = child.node() ? &child.node()->as<{}>() : nullptr; return;\n".format(
+                                m[1], m[2]
+                            )
                         )
 
-                cppf.write("        default: THROW_UNREACHABLE;\n")
+                cppf.write("        default: SLANG_UNREACHABLE;\n")
                 cppf.write("    }\n")
             else:
                 cppf.write("    (void)index;\n")
@@ -463,13 +587,10 @@ size_t SyntaxNode::getChildCount() const {
 
             cppf.write("}\n\n")
 
-            cppf.write("{0}* {0}::clone(BumpAllocator& alloc) const {{\n".format(k))
-            cppf.write("    return alloc.emplace<{}>(*this);\n".format(k))
-            cppf.write("}\n\n")
-
     # Write out syntax factory methods.
-    outf.write("class SyntaxFactory {\n")
+    outf.write("class SLANG_EXPORT SyntaxFactory {\n")
     outf.write("public:\n")
+    outf.write("    using Token = parsing::Token;\n\n")
     outf.write("    explicit SyntaxFactory(BumpAllocator& alloc) : alloc(alloc) {}\n")
     outf.write("\n")
 
@@ -498,7 +619,7 @@ std::ostream& operator<<(std::ostream& os, SyntaxKind kind) {
     return os;
 }
 
-string_view toString(SyntaxKind kind) {
+std::string_view toString(SyntaxKind kind) {
     switch (kind) {
         case SyntaxKind::Unknown: return "Unknown";
         case SyntaxKind::SyntaxList: return "SyntaxList";
@@ -532,6 +653,7 @@ string_view toString(SyntaxKind kind) {
     cppf.write(
         """};
 
+#ifdef SLANG_RTTI_ENABLED
 const std::type_info* typeFromSyntaxKind(SyntaxKind kind) {
     switch (kind) {
         case SyntaxKind::Unknown: break;
@@ -548,6 +670,7 @@ const std::type_info* typeFromSyntaxKind(SyntaxKind kind) {
         """    }
     return nullptr;
 }
+#endif
 
 }
 """
@@ -560,6 +683,17 @@ const std::type_info* typeFromSyntaxKind(SyntaxKind kind) {
 
     # Write out a dispatch method to get from SyntaxKind to actual concrete type
     outf.write("namespace detail {\n\n")
+    outf.write("struct InvalidSyntaxNode : public SyntaxNode {\n")
+    outf.write(
+        "    static bool isKind(SyntaxKind kind) { return kind == SyntaxKind::Unknown; }\n"
+    )
+    outf.write("    static bool isChildOptional(size_t) { return true; }\n")
+    outf.write("    TokenOrSyntax getChild(size_t) { return nullptr; }\n")
+    outf.write("    ConstTokenOrSyntax getChild(size_t) const { return nullptr; }\n")
+    outf.write("    PtrTokenOrSyntax getChildPtr(size_t) { return nullptr; }\n")
+    outf.write("    void setChild(size_t, TokenOrSyntax) {}\n")
+    outf.write("};\n\n")
+
     outf.write("template<typename TNode, typename TVisitor, typename... Args>\n")
     outf.write(
         "decltype(auto) visitSyntaxNode(TNode* node, TVisitor& visitor, Args&&... args) {\n"
@@ -567,7 +701,7 @@ const std::type_info* typeFromSyntaxKind(SyntaxKind kind) {
     outf.write("    static constexpr bool isConst = std::is_const_v<TNode>;")
     outf.write("    switch (node->kind) {\n")
     outf.write(
-        "        case SyntaxKind::Unknown: return visitor.visitInvalid(*node, std::forward<Args>(args)...);\n"
+        "        case SyntaxKind::Unknown: return visitor.visit(*static_cast<std::conditional_t<isConst, const InvalidSyntaxNode*, InvalidSyntaxNode*>>(node), std::forward<Args>(args)...);\n"
     )
     outf.write("        case SyntaxKind::SyntaxList:\n")
     outf.write("        case SyntaxKind::TokenList:\n")
@@ -588,7 +722,7 @@ const std::type_info* typeFromSyntaxKind(SyntaxKind kind) {
         alltypes.pop(v, None)
 
     outf.write("    }\n")
-    outf.write("    THROW_UNREACHABLE;\n")
+    outf.write("    SLANG_UNREACHABLE;\n")
     outf.write("}\n\n")
     outf.write("}\n\n")
 
@@ -626,17 +760,16 @@ const std::type_info* typeFromSyntaxKind(SyntaxKind kind) {
 //! @file SyntaxKind.h
 //! @brief Generated SyntaxKind enum
 //
-// File is under the MIT license; see LICENSE for details
+// SPDX-FileCopyrightText: Michael Popoloski
+// SPDX-License-Identifier: MIT
 //------------------------------------------------------------------------------
 #pragma once
 
-#include "slang/util/Enum.h"
-
 namespace std { class type_info; }
 
-namespace slang {
+namespace slang::syntax {
 
-enum class SyntaxKind {
+enum class SLANG_EXPORT SyntaxKind {
     Unknown,
     SyntaxList,
     TokenList,
@@ -650,15 +783,15 @@ enum class SyntaxKind {
     outf.write(
         """}};
 
-std::ostream& operator<<(std::ostream& os, SyntaxKind kind);
-string_view toString(SyntaxKind kind);
+SLANG_EXPORT std::ostream& operator<<(std::ostream& os, SyntaxKind kind);
+SLANG_EXPORT std::string_view toString(SyntaxKind kind);
 
-class SyntaxKind_traits {{
+class SLANG_EXPORT SyntaxKind_traits {{
 public:
     static const std::array<SyntaxKind, {}> values;
 }};
 
-const std::type_info* typeFromSyntaxKind(SyntaxKind kind);
+SLANG_EXPORT const std::type_info* typeFromSyntaxKind(SyntaxKind kind);
 
 }}
 """.format(
@@ -666,6 +799,151 @@ const std::type_info* typeFromSyntaxKind(SyntaxKind kind);
         )
     )
 
+    # Write the forward declaration header file.
+    outf = open(os.path.join(headerdir, "SyntaxFwd.h"), "w")
+    outf.write(
+        """//------------------------------------------------------------------------------
+//! @file SyntaxFwd.h
+//! @brief Forward declarations for syntax node types
+//
+// SPDX-FileCopyrightText: Michael Popoloski
+// SPDX-License-Identifier: MIT
+//------------------------------------------------------------------------------
+#pragma once
+
+namespace slang::syntax {
+
+class SyntaxNode;
+"""
+    )
+
+    # Write all type names.
+    for name, _ in alltypesSaved.items():
+        if name == "SyntaxNode":
+            continue
+
+        outf.write("struct {};\n".format(name))
+    outf.write("\n}\n")
+
+
+def generateSyntaxClone(builddir, alltypes, kindmap):
+    # Start the clone source file.
+    clonef = open(os.path.join(builddir, "SyntaxClone.cpp"), "w")
+    clonef.write(
+        """//------------------------------------------------------------------------------
+// SyntaxClone.cpp
+// All generated syntax node clone functionality
+//
+// SPDX-FileCopyrightText: Michael Popoloski
+// SPDX-License-Identifier: MIT
+//------------------------------------------------------------------------------
+#include "slang/syntax/AllSyntax.h"
+
+// This file contains all syntax generated clone implementations.
+// It is auto-generated by the syntax_gen.py script under the scripts/ directory.
+
+namespace slang::syntax::shallow {
+
+template<typename T>
+SyntaxNode* clone(const T& node, BumpAllocator& alloc) {
+    return alloc.emplace<T>(node);
+}
+
+}
+
+"""
+    )
+    clonef.write(
+        """namespace slang::syntax::deep {
+
+template<typename T>
+SyntaxNode* clone(const T& node, BumpAllocator& alloc) {
+    return alloc.emplace<T>(node);
+}
+
+SyntaxNode* clone(const SyntaxListBase&, BumpAllocator&) {
+    return nullptr;
+}
+"""
+    )
+    # Write out deepClone methods for each derived type.
+    for k, v in sorted(alltypes.items()):
+        if not v.final:
+            continue
+        if v.final:
+            clonef.write(
+                "SyntaxNode* clone(const {0}& node, BumpAllocator& alloc) {{\n".format(
+                    k
+                )
+            )
+            clonef.write("    return alloc.emplace<{0}>(\n".format(k))
+            if "kind" in v.argNames:
+                clonef.write("        node.kind,\n")
+            for i, m in enumerate(v.combinedMembers):
+                if m[1] in v.notNullMembers:
+                    clonef.write(
+                        "        *deepClone<{0}>(*node.{1}, alloc)".format(
+                            m[0][9:-2], m[1]
+                        )
+                    )
+                elif m[1] in v.optionalMembers:
+                    clonef.write(
+                        "        node.{0} ? deepClone(*node.{0}, alloc) : nullptr".format(
+                            m[1]
+                        )
+                    )
+                elif (
+                    m[0].startswith("SyntaxList")
+                    or m[0].startswith("SeparatedSyntaxList")
+                    or m[0].startswith("TokenList")
+                ):
+                    clonef.write("        *deepClone(node.{0}, alloc)".format(m[1]))
+                elif m[0] == "Token":
+                    clonef.write("        node.{0}.deepClone(alloc)".format(m[1]))
+                else:
+                    clonef.write("        node.{0}".format(m[1]))
+                if i != len(v.combinedMembers) - 1:
+                    clonef.write(",\n")
+                else:
+                    clonef.write("\n")
+            clonef.write("    );\n")
+            clonef.write("}\n\n")
+    clonef.write("}\n\n")
+    clonef.write(
+        """namespace slang::syntax {
+
+struct CloneVisitor {
+    template<typename T>
+    SyntaxNode* visit(const T& node, BumpAllocator& alloc) {
+        if constexpr (requires { node.clone(alloc); }) {
+            return node.clone(alloc);
+        } else {
+            return shallow::clone(node, alloc);
+        }
+    }
+};
+
+struct DeepCloneVisitor {
+    template<typename T>
+    SyntaxNode* visit(const T& node, BumpAllocator& alloc) {
+        return deep::clone(node, alloc);
+    }
+};
+
+SyntaxNode* deepClone(const SyntaxNode& node, BumpAllocator& alloc) {
+    DeepCloneVisitor visitor;
+    return node.visit(visitor, alloc);
+}
+
+SyntaxNode* clone(const SyntaxNode& node, BumpAllocator& alloc) {
+    CloneVisitor visitor;
+    return node.visit(visitor, alloc);
+}
+
+}
+"""
+    )
+
 
 def loadkinds(ourdir, filename):
     kinds = []
@@ -680,17 +958,17 @@ def loadkinds(ourdir, filename):
 
 
 def writekinddecl(outf, name, basetype, kinds):
-    outf.write("enum class {} : {} {{\n".format(name, basetype))
+    outf.write("enum class SLANG_EXPORT {} : {} {{\n".format(name, basetype))
     for k in kinds:
         outf.write("    {},\n".format(k))
 
     outf.write(
         """}};
 
-std::ostream& operator<<(std::ostream& os, {} kind);
-string_view toString({} kind);
+SLANG_EXPORT std::ostream& operator<<(std::ostream& os, {} kind);
+SLANG_EXPORT std::string_view toString({} kind);
 
-class {}_traits {{
+class SLANG_EXPORT {}_traits {{
 public:
     static const std::array<{}, {}> values;
 }};
@@ -708,7 +986,7 @@ def writekindimpls(outf, name, kinds):
     return os;
 }}
 
-string_view toString({} kind) {{
+std::string_view toString({} kind) {{
     switch (kind) {{
 """.format(
             name, name
@@ -757,15 +1035,17 @@ def generateTokenKinds(ourdir, builddir):
 //! @file TokenKind.h
 //! @brief Generated TokenKind and TriviaKind enums
 //
-// File is under the MIT license; see LICENSE for details
+// SPDX-FileCopyrightText: Michael Popoloski
+// SPDX-License-Identifier: MIT
 //------------------------------------------------------------------------------
 #pragma once
 
+#include <array>
 #include <ostream>
 
 #include "slang/util/Util.h"
 
-namespace slang {
+namespace slang::parsing {
 
 """
     )
@@ -780,11 +1060,12 @@ namespace slang {
 // TokenKind.cpp
 // Generated TokenKind and TriviaKind enums
 //
-// File is under the MIT license; see LICENSE for details
+// SPDX-FileCopyrightText: Michael Popoloski
+// SPDX-License-Identifier: MIT
 //------------------------------------------------------------------------------
 #include "slang/parsing/TokenKind.h"
 
-namespace slang {
+namespace slang::parsing {
 
 """
     )
@@ -795,32 +1076,43 @@ namespace slang {
 
 
 def generatePyBindings(builddir, alltypes):
-    outf = open(os.path.join(builddir, "PySyntaxBindings.cpp"), "w")
-    outf.write(
-        """//------------------------------------------------------------------------------
-// PySyntaxBindings.cpp
+    numfiles = 4
+    items = list(alltypes.items())
+    perfile = math.ceil(len(items) / numfiles)
+
+    for i in range(numfiles):
+        outf = open(os.path.join(builddir, f"PySyntaxBindings{i}.cpp"), "w")
+        outf.write(
+            """//------------------------------------------------------------------------------
+// PySyntaxBindings{0}.cpp
 // Generated Python bindings for syntax types
 //
-// File is under the MIT license; see LICENSE for details
+// SPDX-FileCopyrightText: Michael Popoloski
+// SPDX-License-Identifier: MIT
 //------------------------------------------------------------------------------
 #include "pyslang.h"
 
 #include "slang/syntax/AllSyntax.h"
 
-void registerSyntaxNodes(py::module_& m) {
-"""
-    )
+void registerSyntaxNodes{0}(py::module_& m) {{
+""".format(
+                i
+            )
+        )
 
-    for k, v in alltypes.items():
-        if k == "SyntaxNode":
-            continue
+        idx = i * perfile
+        for k, v in items[idx : idx + perfile]:
+            if k == "SyntaxNode":
+                continue
 
-        outf.write('    py::class_<{}, {}>(m, "{}")'.format(k, v.base, k))
-        for m in v.members:
-            outf.write('\n        .def_readwrite("{}", &{}::{})'.format(m[1], k, m[1]))
-        outf.write(";\n\n")
+            outf.write('    py::class_<{}, {}>(m, "{}")'.format(k, v.base, k))
+            for m in v.members:
+                outf.write(
+                    '\n        .def_readwrite("{}", &{}::{})'.format(m[1], k, m[1])
+                )
+            outf.write(";\n\n")
 
-    outf.write("}\n")
+        outf.write("}\n")
 
 
 if __name__ == "__main__":
