@@ -2,26 +2,31 @@
 // DiagnosticEngine.cpp
 // Primary interface for managing diagnostic reporting
 //
-// File is under the MIT license; see LICENSE for details
+// SPDX-FileCopyrightText: Michael Popoloski
+// SPDX-License-Identifier: MIT
 //------------------------------------------------------------------------------
 #include "slang/diagnostics/DiagnosticEngine.h"
 
+#include <algorithm>
+#include <fmt/args.h>
+
 #include "slang/diagnostics/DiagArgFormatter.h"
 #include "slang/diagnostics/DiagnosticClient.h"
 #include "slang/diagnostics/MetaDiags.h"
 #include "slang/diagnostics/TextDiagnosticClient.h"
+#include "slang/text/Glob.h"
 #include "slang/text/SourceManager.h"
-#include "slang/util/StackContainer.h"
-#include "slang/util/String.h"
+
+namespace fs = std::filesystem;
 
 namespace slang {
 
 // These are defined in the generated DiagCode.cpp file.
-string_view getDefaultMessage(DiagCode code);
+std::string_view getDefaultMessage(DiagCode code);
 DiagnosticSeverity getDefaultSeverity(DiagCode code);
-string_view getDefaultOptionName(DiagCode code);
-span<const DiagCode> findDiagsFromOptionName(string_view name);
-const DiagGroup* findDefaultDiagGroup(string_view name);
+std::string_view getDefaultOptionName(DiagCode code);
+std::span<const DiagCode> findDiagsFromOptionName(std::string_view name);
+const DiagGroup* findDefaultDiagGroup(std::string_view name);
 
 DiagnosticEngine::FormatterMap DiagnosticEngine::defaultFormatters;
 
@@ -93,21 +98,21 @@ void DiagnosticEngine::setMessage(DiagCode code, const std::string& message) {
     messageTable[code] = message;
 }
 
-string_view DiagnosticEngine::getMessage(DiagCode code) const {
+std::string_view DiagnosticEngine::getMessage(DiagCode code) const {
     if (auto it = messageTable.find(code); it != messageTable.end())
         return it->second;
     return getDefaultMessage(code);
 }
 
-string_view DiagnosticEngine::getOptionName(DiagCode code) const {
+std::string_view DiagnosticEngine::getOptionName(DiagCode code) const {
     return getDefaultOptionName(code);
 }
 
-span<const DiagCode> DiagnosticEngine::findFromOptionName(string_view optionName) const {
+std::span<const DiagCode> DiagnosticEngine::findFromOptionName(std::string_view optionName) const {
     return findDiagsFromOptionName(optionName);
 }
 
-const DiagGroup* DiagnosticEngine::findDiagGroup(string_view name) const {
+const DiagGroup* DiagnosticEngine::findDiagGroup(std::string_view name) const {
     return findDefaultDiagGroup(name);
 }
 
@@ -117,32 +122,45 @@ void DiagnosticEngine::clearMappings() {
 }
 
 void DiagnosticEngine::clearMappings(DiagnosticSeverity severity) {
-    for (auto it = severityTable.begin(); it != severityTable.end();) {
-        if (it->second == severity)
-            it = severityTable.erase(it);
-        else
-            it++;
-    }
+    erase_if(severityTable, [severity](auto& pair) { return pair.second == severity; });
+}
+
+std::error_code DiagnosticEngine::addIgnorePaths(std::string_view pattern) {
+    std::error_code ec;
+    auto p = fs::weakly_canonical(pattern, ec);
+    if (!ec)
+        ignoreWarnPatterns.emplace_back(std::move(p));
+
+    return ec;
+}
+
+std::error_code DiagnosticEngine::addIgnoreMacroPaths(std::string_view pattern) {
+    std::error_code ec;
+    auto p = fs::weakly_canonical(pattern, ec);
+    if (!ec)
+        ignoreMacroWarnPatterns.emplace_back(std::move(p));
+
+    return ec;
 }
 
 // Checks that all of the given ranges are in the same macro argument expansion as `loc`
 static bool checkMacroArgRanges(const DiagnosticEngine& engine, SourceLocation loc,
-                                span<const SourceRange> ranges) {
+                                std::span<const SourceRange> ranges) {
     const SourceManager& sm = engine.getSourceManager();
     if (!sm.isMacroArgLoc(loc))
         return false;
 
     loc = sm.getExpansionLoc(loc);
 
-    SmallVectorSized<SourceRange, 8> mappedRanges;
+    SmallVector<SourceRange, 8> mappedRanges;
     engine.mapSourceRanges(loc, ranges, mappedRanges, false);
 
     if (ranges.size() > mappedRanges.size())
         return false;
 
     for (auto& range : mappedRanges) {
-        ASSERT(range.start().buffer() == loc.buffer());
-        ASSERT(range.end().buffer() == loc.buffer());
+        SLANG_ASSERT(range.start().buffer() == loc.buffer());
+        SLANG_ASSERT(range.end().buffer() == loc.buffer());
 
         if (loc < range.start() || loc >= range.end())
             return false;
@@ -155,61 +173,83 @@ void DiagnosticEngine::issue(const Diagnostic& diagnostic) {
     if (issuedOverLimitErr)
         return;
 
-    DiagnosticSeverity severity = getSeverity(diagnostic.code, diagnostic.location);
-    switch (severity) {
-        case DiagnosticSeverity::Ignored:
-            return;
-        case DiagnosticSeverity::Note:
-            break;
-        case DiagnosticSeverity::Warning:
-            numWarnings++;
-            break;
-        case DiagnosticSeverity::Error:
-        case DiagnosticSeverity::Fatal:
-            if (errorLimit && numErrors >= errorLimit) {
-                Diagnostic diag(diag::TooManyErrors, SourceLocation::NoLocation);
-                issueImpl(diag, DiagnosticSeverity::Fatal);
-                issuedOverLimitErr = true;
-                return;
-            }
+    const DiagnosticSeverity severity = getSeverity(diagnostic.code, diagnostic.location);
+    if (severity == DiagnosticSeverity::Ignored)
+        return;
 
-            numErrors++;
-            break;
+    const bool isError = severity == DiagnosticSeverity::Error ||
+                         severity == DiagnosticSeverity::Fatal;
+    if (isError && errorLimit && numErrors >= errorLimit) {
+        Diagnostic diag(diag::TooManyErrors, SourceLocation::NoLocation);
+        issueImpl(diag, DiagnosticSeverity::Fatal);
+        issuedOverLimitErr = true;
+        return;
     }
 
-    issueImpl(diagnostic, severity);
+    if (!issueImpl(diagnostic, severity))
+        return;
+
+    if (severity == DiagnosticSeverity::Warning)
+        numWarnings++;
+    else if (isError)
+        numErrors++;
 }
 
-void DiagnosticEngine::issueImpl(const Diagnostic& diagnostic, DiagnosticSeverity severity) {
+bool DiagnosticEngine::issueImpl(const Diagnostic& diagnostic, DiagnosticSeverity severity) {
     // Walk out until we find a location for this diagnostic that isn't inside a macro.
-    SmallVectorSized<SourceLocation, 8> expansionLocs;
+    SmallVector<SourceLocation, 8> expansionLocs;
     SourceLocation loc = diagnostic.location;
-    size_t ignoreUntil = 0;
+    size_t ignoreExpansionsUntil = 0;
     bool showIncludeStack = false;
 
     if (loc != SourceLocation::NoLocation) {
         while (sourceManager.isMacroLoc(loc)) {
             SourceLocation prevLoc = loc;
             if (sourceManager.isMacroArgLoc(loc)) {
-                expansionLocs.append(sourceManager.getExpansionLoc(loc));
+                expansionLocs.push_back(sourceManager.getExpansionLoc(loc));
                 loc = sourceManager.getOriginalLoc(loc);
             }
             else {
-                expansionLocs.append(loc);
+                expansionLocs.push_back(loc);
                 loc = sourceManager.getExpansionLoc(loc);
             }
 
             if (checkMacroArgRanges(*this, prevLoc, diagnostic.ranges))
-                ignoreUntil = expansionLocs.size();
+                ignoreExpansionsUntil = expansionLocs.size();
         }
 
         showIncludeStack = reportedIncludeStack.emplace(loc.buffer()).second;
+
+        auto checkSuppressed = [&](const std::vector<fs::path>& patterns, SourceLocation loc) {
+            if (patterns.empty())
+                return false;
+
+            auto& path = sourceManager.getFullPath(loc.buffer());
+            for (auto& pattern : patterns) {
+                if (svGlobMatches(path, pattern))
+                    return true;
+            }
+            return false;
+        };
+
+        if (getDefaultSeverity(diagnostic.code) == DiagnosticSeverity::Warning) {
+            if (checkSuppressed(ignoreWarnPatterns, loc))
+                return false;
+
+            if (ignoreExpansionsUntil < expansionLocs.size() && !ignoreMacroWarnPatterns.empty()) {
+                auto originalLoc = sourceManager.getFullyOriginalLoc(
+                    expansionLocs[ignoreExpansionsUntil]);
+
+                if (checkSuppressed(ignoreMacroWarnPatterns, originalLoc))
+                    return false;
+            }
+        }
     }
 
     std::string message = formatMessage(diagnostic);
 
     ReportedDiagnostic report(diagnostic);
-    report.expansionLocs = span<SourceLocation>(expansionLocs).subspan(ignoreUntil);
+    report.expansionLocs = std::span<SourceLocation>(expansionLocs).subspan(ignoreExpansionsUntil);
     report.ranges = diagnostic.ranges;
     report.location = loc;
     report.severity = severity;
@@ -225,6 +265,8 @@ void DiagnosticEngine::issueImpl(const Diagnostic& diagnostic, DiagnosticSeverit
         if (note.location != SourceLocation::NoLocation || note.code == diag::NoteFromHere2)
             issue(note);
     }
+
+    return true;
 }
 
 std::string DiagnosticEngine::formatMessage(const Diagnostic& diag) const {
@@ -247,11 +289,11 @@ std::string DiagnosticEngine::formatMessage(const Diagnostic& diag) const {
             [&](auto&& t) {
                 // If the argument is a pointer, the fmtlib API needs it unwrapped into a reference.
                 using T = std::decay_t<decltype(t)>;
-                if constexpr (std::is_same_v<std::any, T>) {
-                    if (auto it = formatters.find(t.type()); it != formatters.end())
-                        it->second->format(args, t);
+                if constexpr (std::is_same_v<Diagnostic::CustomArgType, T>) {
+                    if (auto it = formatters.find(t.first); it != formatters.end())
+                        args.push_back(it->second->format(t.second));
                     else
-                        throw std::runtime_error("No diagnostic formatter for type");
+                        SLANG_THROW(std::runtime_error("No diagnostic formatter for type"));
                 }
                 else if constexpr (std::is_same_v<ConstantValue, T>) {
                     if (t.isReal())
@@ -273,10 +315,10 @@ std::string DiagnosticEngine::formatMessage(const Diagnostic& diag) const {
 
 // Walks up a chain of macro argument expansions and collects their buffer IDs.
 static void getMacroArgExpansions(const SourceManager& sm, SourceLocation loc, bool isStart,
-                                  SmallVector<BufferID>& results) {
+                                  SmallVectorBase<BufferID>& results) {
     while (sm.isMacroLoc(loc)) {
         if (sm.isMacroArgLoc(loc)) {
-            results.append(loc.buffer());
+            results.push_back(loc.buffer());
             loc = sm.getOriginalLoc(loc);
         }
         else {
@@ -289,23 +331,22 @@ static void getMacroArgExpansions(const SourceManager& sm, SourceLocation loc, b
 // Finds all macro argument expansions that are common in both start and end.
 static void getCommonMacroArgExpansions(const SourceManager& sm, SourceLocation start,
                                         SourceLocation end, std::vector<BufferID>& results) {
-    SmallVectorSized<BufferID, 8> startArgExpansions;
-    SmallVectorSized<BufferID, 8> endArgExpansions;
+    SmallVector<BufferID> startArgExpansions;
+    SmallVector<BufferID> endArgExpansions;
     getMacroArgExpansions(sm, start, true, startArgExpansions);
     getMacroArgExpansions(sm, end, false, endArgExpansions);
-    std::stable_sort(startArgExpansions.begin(), startArgExpansions.end());
-    std::stable_sort(endArgExpansions.begin(), endArgExpansions.end());
+    std::ranges::stable_sort(startArgExpansions);
+    std::ranges::stable_sort(endArgExpansions);
 
-    std::set_intersection(startArgExpansions.begin(), startArgExpansions.end(),
-                          endArgExpansions.begin(), endArgExpansions.end(),
-                          std::back_inserter(results));
+    std::ranges::set_intersection(startArgExpansions, endArgExpansions,
+                                  std::back_inserter(results));
 }
 
 // Recursively tries to find an expansion location of `loc` that is in the
 // same buffer as `contextLoc`, taking into account macros and macro arguments.
 static SourceLocation getMatchingMacroLoc(const SourceManager& sm, SourceLocation loc,
                                           SourceLocation contextLoc, bool isStart,
-                                          span<const BufferID> commonArgs) {
+                                          std::span<const BufferID> commonArgs) {
     if (loc.buffer() == contextLoc.buffer())
         return loc;
 
@@ -317,7 +358,7 @@ static SourceLocation getMatchingMacroLoc(const SourceManager& sm, SourceLocatio
     if (sm.isMacroArgLoc(loc)) {
         // Only look at the original location of this argument if the other location
         // in the range is also present in the expansion.
-        if (std::binary_search(commonArgs.begin(), commonArgs.end(), loc.buffer())) {
+        if (std::ranges::binary_search(commonArgs, loc.buffer())) {
             SourceLocation orig = sm.getOriginalLoc(loc);
             macroRange = SourceRange(orig, orig);
         }
@@ -341,8 +382,8 @@ static SourceLocation getMatchingMacroLoc(const SourceManager& sm, SourceLocatio
     return getMatchingMacroLoc(sm, argLoc, contextLoc, isStart, commonArgs);
 }
 
-void DiagnosticEngine::mapSourceRanges(SourceLocation loc, span<const SourceRange> ranges,
-                                       SmallVector<SourceRange>& mapped,
+void DiagnosticEngine::mapSourceRanges(SourceLocation loc, std::span<const SourceRange> ranges,
+                                       SmallVectorBase<SourceRange>& mapped,
                                        bool mapOriginalLocations) const {
     const SourceManager& sm = sourceManager;
 
@@ -354,7 +395,7 @@ void DiagnosticEngine::mapSourceRanges(SourceLocation loc, span<const SourceRang
         // Find a common parent for start and end. Start with `start` and
         // walk upwards until we find `end`s buffer or run out of expansions.
         SmallMap<BufferID, SourceLocation, 8> startMap;
-        while (sm.isMacroLoc(start) && start.buffer() != end.buffer()) {
+        while (start.buffer() != end.buffer() && sm.isMacroLoc(start)) {
             startMap[start.buffer()] = start;
             start = sm.getExpansionLoc(start);
         }
@@ -385,12 +426,12 @@ void DiagnosticEngine::mapSourceRanges(SourceLocation loc, span<const SourceRang
             start = sm.getFullyOriginalLoc(start);
             end = sm.getFullyOriginalLoc(end);
         }
-        mapped.emplace(start, end);
+        mapped.emplace_back(start, end);
     }
 }
 
 std::string DiagnosticEngine::reportAll(const SourceManager& sourceManager,
-                                        span<const Diagnostic> diags) {
+                                        std::span<const Diagnostic> diags) {
     DiagnosticEngine engine(sourceManager);
 
     auto client = std::make_shared<TextDiagnosticClient>();
@@ -407,9 +448,10 @@ void DiagnosticEngine::setDefaultWarnings() {
     setSeverity(*findDiagGroup("default"sv), DiagnosticSeverity::Warning);
 }
 
-Diagnostics DiagnosticEngine::setWarningOptions(span<const std::string> options) {
+Diagnostics DiagnosticEngine::setWarningOptions(std::span<const std::string> options) {
     Diagnostics diags;
-    auto findAndSet = [&](string_view name, DiagnosticSeverity severity, const char* errorPrefix) {
+    auto findAndSet = [&](std::string_view name, DiagnosticSeverity severity,
+                          const char* errorPrefix) {
         if (auto group = findDiagGroup(name)) {
             setSeverity(*group, severity);
         }
@@ -435,15 +477,19 @@ Diagnostics DiagnosticEngine::setWarningOptions(span<const std::string> options)
             setIgnoreAllWarnings(true);
         }
         else if (arg == "error") {
+            for (auto it = severityTable.begin(); it != severityTable.end(); it++) {
+                if (it->second == DiagnosticSeverity::Warning)
+                    it->second = DiagnosticSeverity::Error;
+            }
             setWarningsAsErrors(true);
         }
-        else if (startsWith(arg, "error=")) {
+        else if (arg.starts_with("error=")) {
             findAndSet(arg.substr(6), DiagnosticSeverity::Error, "-Werror=");
         }
-        else if (startsWith(arg, "no-error=")) {
-            findAndSet(arg.substr(9), DiagnosticSeverity::Error, "-Wno-error=");
+        else if (arg.starts_with("no-error=")) {
+            findAndSet(arg.substr(9), DiagnosticSeverity::Warning, "-Wno-error=");
         }
-        else if (startsWith(arg, "no-")) {
+        else if (arg.starts_with("no-")) {
             findAndSet(arg.substr(3), DiagnosticSeverity::Ignored, "-Wno-");
         }
         else {
@@ -456,7 +502,7 @@ Diagnostics DiagnosticEngine::setWarningOptions(span<const std::string> options)
 
 template<typename TDirective>
 void DiagnosticEngine::setMappingsFromPragmasImpl(BufferID buffer,
-                                                  span<const TDirective> directives,
+                                                  std::span<const TDirective> directives,
                                                   Diagnostics& diags) {
 
     // Store the state of diagnostics each time the user pushes,
@@ -492,7 +538,7 @@ void DiagnosticEngine::setMappingsFromPragmasImpl(BufferID buffer,
             mappingStack.pop_back();
         }
         else {
-            if (startsWith(name, "-W"sv))
+            if (name.starts_with("-W"sv))
                 name = name.substr(2);
 
             if (auto group = findDiagGroup(name)) {
@@ -504,8 +550,8 @@ void DiagnosticEngine::setMappingsFromPragmasImpl(BufferID buffer,
                     noteDiag(code, directive);
             }
             else {
-                auto& diag =
-                    diags.add(diag::UnknownWarningOption, SourceLocation(buffer, directive.offset));
+                auto& diag = diags.add(diag::UnknownWarningOption,
+                                       SourceLocation(buffer, directive.offset));
                 diag << name;
             }
         }
@@ -533,8 +579,8 @@ Diagnostics DiagnosticEngine::setMappingsFromPragmas(BufferID buffer) {
     return diags;
 }
 
-optional<DiagnosticSeverity> DiagnosticEngine::findMappedSeverity(DiagCode code,
-                                                                  SourceLocation location) const {
+std::optional<DiagnosticSeverity> DiagnosticEngine::findMappedSeverity(
+    DiagCode code, SourceLocation location) const {
     auto byCode = diagMappings.find(code);
     if (byCode == diagMappings.end())
         return std::nullopt;
@@ -545,10 +591,8 @@ optional<DiagnosticSeverity> DiagnosticEngine::findMappedSeverity(DiagCode code,
         return std::nullopt;
 
     const std::vector<DiagnosticMapping>& mappings = byBuffer->second;
-    auto byOffset = std::lower_bound(
-        mappings.begin(), mappings.end(), fileLoc.offset(),
-        [](const DiagnosticMapping& mapping, size_t off) { return mapping.offset < off; });
-
+    auto byOffset = std::ranges::lower_bound(mappings, fileLoc.offset(), {},
+                                             &DiagnosticMapping::offset);
     if (byOffset == mappings.begin())
         return std::nullopt;
 
