@@ -2,7 +2,8 @@
 // Diagnostics.cpp
 // Diagnostic tracking and reporting
 //
-// File is under the MIT license; see LICENSE for details
+// SPDX-FileCopyrightText: Michael Popoloski
+// SPDX-License-Identifier: MIT
 //------------------------------------------------------------------------------
 #include "slang/diagnostics/Diagnostics.h"
 
@@ -10,14 +11,14 @@
 
 namespace slang {
 
-// Defined in the generated DiagCode.cpp file.
-DiagnosticSeverity getDefaultSeverity(DiagCode code);
+Diagnostic::Diagnostic() noexcept : location(SourceLocation::NoLocation) {
+}
 
 Diagnostic::Diagnostic(DiagCode code, SourceLocation location) noexcept :
     code(code), location(location) {
 }
 
-Diagnostic::Diagnostic(const Symbol& source, DiagCode code, SourceLocation location) noexcept :
+Diagnostic::Diagnostic(const ast::Symbol& source, DiagCode code, SourceLocation location) noexcept :
     code(code), location(location), symbol(&source) {
 }
 
@@ -26,7 +27,7 @@ bool Diagnostic::isError() const {
 }
 
 Diagnostic& Diagnostic::addNote(DiagCode noteCode, SourceLocation noteLocation) {
-    ASSERT(noteLocation);
+    SLANG_ASSERT(noteLocation);
     notes.emplace_back(noteCode, noteLocation);
     return notes.back();
 }
@@ -46,20 +47,20 @@ Diagnostic& Diagnostic::addStringAllowEmpty(const std::string& arg) {
 }
 
 Diagnostic& Diagnostic::operator<<(const std::string& arg) {
-    ASSERT(!arg.empty());
+    SLANG_ASSERT(!arg.empty());
     args.emplace_back(arg);
     return *this;
 }
 
-Diagnostic& Diagnostic::operator<<(string_view arg) {
-    ASSERT(!arg.empty());
+Diagnostic& Diagnostic::operator<<(std::string_view arg) {
+    SLANG_ASSERT(!arg.empty());
     args.emplace_back(std::string(arg));
     return *this;
 }
 
 Diagnostic& Diagnostic::operator<<(SourceRange range) {
-    ASSERT(range.start());
-    ASSERT(range.end());
+    SLANG_ASSERT(range.start());
+    SLANG_ASSERT(range.end());
     ranges.push_back(range);
     return *this;
 }
@@ -98,8 +99,8 @@ bool Diagnostic::operator==(const Diagnostic& rhs) const {
                 using LT = std::decay_t<decltype(l)>;
                 using RT = std::decay_t<decltype(r)>;
                 if constexpr (std::is_same_v<LT, RT>) {
-                    if constexpr (std::is_same_v<std::any, LT>) {
-                        return l.type() == r.type();
+                    if constexpr (std::is_same_v<Diagnostic::CustomArgType, LT>) {
+                        return l.first == r.first;
                     }
                     else {
                         return l == r;
@@ -119,8 +120,8 @@ bool Diagnostic::operator==(const Diagnostic& rhs) const {
 }
 
 Diagnostic& Diagnostics::add(DiagCode code, SourceLocation location) {
-    ASSERT(location);
-    emplace(code, location);
+    SLANG_ASSERT(location);
+    emplace_back(code, location);
     return back();
 }
 
@@ -128,13 +129,13 @@ Diagnostic& Diagnostics::add(DiagCode code, SourceRange range) {
     return add(code, range.start()) << range;
 }
 
-Diagnostic& Diagnostics::add(const Symbol& source, DiagCode code, SourceLocation location) {
-    ASSERT(location);
-    emplace(source, code, location);
+Diagnostic& Diagnostics::add(const ast::Symbol& source, DiagCode code, SourceLocation location) {
+    SLANG_ASSERT(location);
+    emplace_back(source, code, location);
     return back();
 }
 
-Diagnostic& Diagnostics::add(const Symbol& source, DiagCode code, SourceRange range) {
+Diagnostic& Diagnostics::add(const ast::Symbol& source, DiagCode code, SourceRange range) {
     return add(source, code, range.start()) << range;
 }
 
@@ -142,14 +143,33 @@ void Diagnostics::sort(const SourceManager& sourceManager) {
     auto compare = [&sourceManager](auto& x, auto& y) {
         SourceLocation xl = sourceManager.getFullyExpandedLoc(x.location);
         SourceLocation yl = sourceManager.getFullyExpandedLoc(y.location);
-        if (xl < yl)
+
+        auto xb = sourceManager.getSortKey(xl.buffer());
+        auto yb = sourceManager.getSortKey(yl.buffer());
+        if (xb < yb)
             return true;
-        if (xl == yl)
-            return x.code < y.code;
+        if (xb == yb) {
+            if (xl.offset() < yl.offset())
+                return true;
+            if (xl == yl)
+                return x.code < y.code;
+        }
         return false;
     };
 
-    std::stable_sort(begin(), end(), compare);
+    std::ranges::stable_sort(*this, compare);
+}
+
+Diagnostics Diagnostics::filter(std::initializer_list<DiagCode> list) const {
+    Diagnostics result;
+    result.reserve(size());
+
+    for (auto& diag : *this) {
+        if (std::ranges::none_of(list, [&](DiagCode code) { return code == diag.code; }))
+            result.push_back(diag);
+    }
+
+    return result;
 }
 
 } // namespace slang
