@@ -2,13 +2,15 @@
 // TextDiagnosticClient.cpp
 // Diagnostic client that formats to a text string
 //
-// File is under the MIT license; see LICENSE for details
+// SPDX-FileCopyrightText: Michael Popoloski
+// SPDX-License-Identifier: MIT
 //------------------------------------------------------------------------------
 #include "slang/diagnostics/TextDiagnosticClient.h"
 
-#include "../text/FormatBuffer.h"
+#include <ranges>
 
 #include "slang/text/CharInfo.h"
+#include "slang/text/FormatBuffer.h"
 #include "slang/text/SourceManager.h"
 
 namespace slang {
@@ -50,7 +52,7 @@ fmt::terminal_color TextDiagnosticClient::getSeverityColor(DiagnosticSeverity se
 
 void TextDiagnosticClient::report(const ReportedDiagnostic& diag) {
     if (diag.shouldShowIncludeStack && includeFileStack) {
-        SmallVectorSized<SourceLocation, 8> includeStack;
+        SmallVector<SourceLocation> includeStack;
         getIncludeStack(diag.location.buffer(), includeStack);
 
         // Show the stack in reverse.
@@ -63,8 +65,10 @@ void TextDiagnosticClient::report(const ReportedDiagnostic& diag) {
 
     // Print out the hierarchy where the diagnostic occurred, if we know it.
     auto& od = diag.originalDiagnostic;
-    if (od.coalesceCount && od.symbol && symbolPathCB && includeHierarchy) {
-        if (od.coalesceCount == 1)
+    if (od.symbol && symbolPathCB &&
+        (includeHierarchy == ShowHierarchyPathOption::Always ||
+         (includeHierarchy == ShowHierarchyPathOption::Auto && od.coalesceCount))) {
+        if (!od.coalesceCount || od.coalesceCount == 1u)
             buffer->append("  in instance: "sv);
         else
             buffer->format("  in {} instances, e.g. ", *od.coalesceCount);
@@ -74,7 +78,7 @@ void TextDiagnosticClient::report(const ReportedDiagnostic& diag) {
     }
 
     // Get all highlight ranges mapped into the reported location of the diagnostic.
-    SmallVectorSized<SourceRange, 8> mappedRanges;
+    SmallVector<SourceRange> mappedRanges;
     engine->mapSourceRanges(diag.location, diag.ranges, mappedRanges);
 
     // Write the diagnostic.
@@ -91,7 +95,7 @@ void TextDiagnosticClient::report(const ReportedDiagnostic& diag) {
             else
                 name = fmt::format("expanded from macro '{}'", name);
 
-            SmallVectorSized<SourceRange, 8> macroRanges;
+            SmallVector<SourceRange> macroRanges;
             engine->mapSourceRanges(loc, diag.ranges, macroRanges);
             formatDiag(sourceManager->getFullyOriginalLoc(loc), macroRanges,
                        DiagnosticSeverity::Note, name, "");
@@ -107,9 +111,9 @@ std::string TextDiagnosticClient::getString() const {
     return buffer->str();
 }
 
-static bool printableTextForNextChar(string_view sourceLine, size_t& index, uint32_t tabStop,
-                                     SmallVector<char>& out, size_t& columnWidth) {
-    ASSERT(index < sourceLine.size());
+static bool printableTextForNextChar(std::string_view sourceLine, size_t& index, uint32_t tabStop,
+                                     SmallVectorBase<char>& out, size_t& columnWidth) {
+    SLANG_ASSERT(index < sourceLine.size());
 
     // Expand tabs based on tabStop setting.
     if (sourceLine[index] == '\t') {
@@ -123,11 +127,11 @@ static bool printableTextForNextChar(string_view sourceLine, size_t& index, uint
         }
 
         uint32_t numSpaces = tabStop - col % tabStop;
-        ASSERT(numSpaces > 0 && numSpaces <= tabStop);
+        SLANG_ASSERT(numSpaces > 0 && numSpaces <= tabStop);
         index++;
 
         for (uint32_t j = 0; j < numSpaces; j++)
-            out.append(' ');
+            out.push_back(' ');
 
         columnWidth = out.size();
         return true;
@@ -139,21 +143,23 @@ static bool printableTextForNextChar(string_view sourceLine, size_t& index, uint
     // Try to decode the next UTF-8 character we see.
     int error;
     uint32_t c;
+    int unused;
     if (index + 4 <= sourceLine.size()) {
-        data = utf8Decode(data, &c, &error);
+        data = utf8Decode(data, &c, &error, unused);
     }
     else {
         char buf[4] = {};
-        memcpy(buf, data, sourceLine.size() - index);
+        auto spaceLeft = sourceLine.size() - index;
+        memcpy(buf, data, spaceLeft);
 
-        auto next = utf8Decode(buf, &c, &error);
-        data += next - buf;
+        auto next = utf8Decode(buf, &c, &error, unused);
+        data += std::min(size_t(next - buf), spaceLeft);
     }
 
     if (error) {
         // Not valid UTF-8, so print a placeholder instead.
         unsigned char invalid = (unsigned char)sourceLine[index++];
-        out.appendRange("<XX>"sv);
+        out.append_range("<XX>"sv);
         out[1] = getHexForDigit(invalid / 16);
         out[2] = getHexForDigit(invalid % 16);
         columnWidth = out.size();
@@ -163,28 +169,28 @@ static bool printableTextForNextChar(string_view sourceLine, size_t& index, uint
     index = size_t(data - sourceLine.data());
 
     if (!isPrintableUnicode(c)) {
-        SmallVectorSized<char, 8> buf;
-        while (c) {
-            buf.append(getHexForDigit(c % 16));
+        SmallVector<char, 8> buf;
+        do {
+            buf.push_back(getHexForDigit(c % 16));
             c /= 16;
-        }
+        } while (c);
 
-        out.appendRange("<U+"sv);
-        out.appendRange(make_reverse_range(buf));
-        out.append('>');
+        out.append_range("<U+"sv);
+        out.append_range(std::views::reverse(buf));
+        out.push_back('>');
         columnWidth = out.size();
         return false;
     }
 
     // Otherwise this is a normal printable character.
-    out.appendIterator(originalData, data);
+    out.append(originalData, data);
     columnWidth = (size_t)charWidthUnicode(c);
     return true;
 }
 
 struct SourceSnippet {
-    SourceSnippet(string_view sourceLine, uint32_t tabStop) {
-        ASSERT(!sourceLine.empty());
+    SourceSnippet(std::string_view sourceLine, uint32_t tabStop) {
+        SLANG_ASSERT(!sourceLine.empty());
 
         byteToColumn.resize(sourceLine.size() + 1);
         for (size_t i = 0; i < byteToColumn.size(); i++)
@@ -192,7 +198,7 @@ struct SourceSnippet {
 
         snippetLine.reserve(sourceLine.size());
 
-        SmallVectorSized<char, 16> buffer;
+        SmallVector<char> buffer;
         size_t column = 0;
         size_t i = 0;
         while (i < sourceLine.size()) {
@@ -201,7 +207,7 @@ struct SourceSnippet {
             size_t columnWidth;
             buffer.clear();
             if (!printableTextForNextChar(sourceLine, i, tabStop, buffer, columnWidth))
-                invalidRanges.append({ snippetLine.size(), buffer.size() });
+                invalidRanges.push_back({snippetLine.size(), buffer.size()});
 
             snippetLine.append(buffer.data(), buffer.size());
             column += columnWidth;
@@ -218,7 +224,7 @@ struct SourceSnippet {
     }
 
     void highlightRange(SourceRange range, SourceLocation caretLoc, size_t col,
-                        string_view sourceLine) {
+                        std::string_view sourceLine) {
         // Trim the range so that it only falls on the same line as the cursor
         size_t start = range.start().offset();
         size_t end = range.end().offset();
@@ -248,13 +254,13 @@ struct SourceSnippet {
 
         size_t startCol = getColumnForByte(start);
         size_t endCol = getColumnForByte(end);
-        ASSERT(startCol <= endCol);
+        SLANG_ASSERT(startCol <= endCol);
 
         if (highlightLine.size() < endCol)
             highlightLine.resize(endCol, ' ');
 
-        std::fill(highlightLine.begin() + ptrdiff_t(startCol),
-                  highlightLine.begin() + ptrdiff_t(endCol), '~');
+        std::ranges::fill(highlightLine.begin() + ptrdiff_t(startCol),
+                          highlightLine.begin() + ptrdiff_t(endCol), '~');
     }
 
     void insertCaret(size_t offset) {
@@ -266,7 +272,7 @@ struct SourceSnippet {
 
     void trimHighlight() { highlightLine.erase(highlightLine.find_last_not_of(' ') + 1); }
 
-    void printTo(FormatBuffer& out, fmt::v9::terminal_color highlightColor) {
+    void printTo(FormatBuffer& out, fmt::terminal_color highlightColor) {
         out.append("\n");
 
         if (invalidRanges.empty()) {
@@ -274,9 +280,9 @@ struct SourceSnippet {
         }
         else {
             size_t index = 0;
-            string_view view = snippetLine;
+            std::string_view view = snippetLine;
             for (auto [start, count] : invalidRanges) {
-                ASSERT(start >= index);
+                SLANG_ASSERT(start >= index);
                 out.append(view.substr(index, start - index));
 
                 out.append(fmt::emphasis::reverse, view.substr(start, count));
@@ -290,15 +296,15 @@ struct SourceSnippet {
         out.append(fg(highlightColor), highlightLine);
     }
 
-    SmallVectorSized<int, 256> byteToColumn;
-    SmallVectorSized<std::pair<size_t, size_t>, 4> invalidRanges;
+    SmallVector<int> byteToColumn;
+    SmallVector<std::pair<size_t, size_t>, 4> invalidRanges;
     std::string snippetLine;
     std::string highlightLine;
 };
 
-void TextDiagnosticClient::formatDiag(SourceLocation loc, span<const SourceRange> ranges,
-                                      DiagnosticSeverity severity, string_view message,
-                                      string_view optionName) {
+void TextDiagnosticClient::formatDiag(SourceLocation loc, std::span<const SourceRange> ranges,
+                                      DiagnosticSeverity severity, std::string_view message,
+                                      std::string_view optionName) {
     constexpr size_t MaxLineLengthToPrint = 4096;
 
     size_t col = 0;
@@ -330,7 +336,7 @@ void TextDiagnosticClient::formatDiag(SourceLocation loc, span<const SourceRange
         buffer->format(" [-W{}]", optionName);
 
     if (hasLocation && includeSource) {
-        string_view line = getSourceLine(loc, col);
+        std::string_view line = getSourceLine(loc, col);
         if (!line.empty() && line.length() < MaxLineLengthToPrint) {
             // We might want to make the tab width configurable at some point,
             // but for now hardcode it to 8 to match the default on basically
