@@ -2,13 +2,15 @@
 // ConstantValue.cpp
 // Compile-time constant representation
 //
-// File is under the MIT license; see LICENSE for details
+// SPDX-FileCopyrightText: Michael Popoloski
+// SPDX-License-Identifier: MIT
 //------------------------------------------------------------------------------
 #include "slang/numeric/ConstantValue.h"
 
-#include "../text/FormatBuffer.h"
-#include <fmt/format.h>
+#include <ostream>
 
+#include "slang/numeric/MathUtils.h"
+#include "slang/text/FormatBuffer.h"
 #include "slang/util/Hash.h"
 
 namespace slang {
@@ -18,14 +20,15 @@ struct always_false : std::false_type {};
 
 const ConstantValue ConstantValue::Invalid;
 
-std::string ConstantValue::toString() const {
+std::string ConstantValue::toString(bitwidth_t abbreviateThresholdBits, bool exactUnknowns,
+                                    bool useAssignmentPatterns) const {
     return std::visit(
-        [](auto&& arg) {
+        [abbreviateThresholdBits, exactUnknowns, useAssignmentPatterns](auto&& arg) {
             using T = std::decay_t<decltype(arg)>;
             if constexpr (std::is_same_v<T, std::monostate>)
                 return "<unset>"s;
             else if constexpr (std::is_same_v<T, SVInt>)
-                return arg.toString();
+                return arg.toString(abbreviateThresholdBits, exactUnknowns);
             else if constexpr (std::is_same_v<T, real_t>)
                 return fmt::format("{}", double(arg));
             else if constexpr (std::is_same_v<T, shortreal_t>)
@@ -36,51 +39,61 @@ std::string ConstantValue::toString() const {
                 return "$"s;
             else if constexpr (std::is_same_v<T, Elements>) {
                 FormatBuffer buffer;
-                buffer.append("[");
+                buffer.append(useAssignmentPatterns ? "'{"sv : "["sv);
                 for (auto& element : arg) {
-                    buffer.append(element.toString());
+                    buffer.append(element.toString(abbreviateThresholdBits, exactUnknowns,
+                                                   useAssignmentPatterns));
                     buffer.append(",");
                 }
 
                 if (!arg.empty())
                     buffer.pop_back();
-                buffer.append("]");
+                buffer.append(useAssignmentPatterns ? "}"sv : "]"sv);
                 return buffer.str();
             }
             else if constexpr (std::is_same_v<T, std::string>)
                 return fmt::format("\"{}\"", arg);
             else if constexpr (std::is_same_v<T, Map>) {
                 FormatBuffer buffer;
-                buffer.append("[");
+                buffer.append(useAssignmentPatterns ? "'{"sv : "["sv);
                 for (auto& [key, val] : *arg)
-                    buffer.format("{}:{},", key.toString(), val.toString());
+                    buffer.format("{}:{},",
+                                  key.toString(abbreviateThresholdBits, exactUnknowns,
+                                               useAssignmentPatterns),
+                                  val.toString(abbreviateThresholdBits, exactUnknowns,
+                                               useAssignmentPatterns));
 
                 if (arg->defaultValue)
-                    buffer.format("default:{}", arg->defaultValue.toString());
+                    buffer.format("default:{}",
+                                  arg->defaultValue.toString(abbreviateThresholdBits, exactUnknowns,
+                                                             useAssignmentPatterns));
                 else if (!arg->empty())
                     buffer.pop_back();
 
-                buffer.append("]");
+                buffer.append(useAssignmentPatterns ? "}"sv : "]"sv);
                 return buffer.str();
             }
             else if constexpr (std::is_same_v<T, Queue>) {
                 FormatBuffer buffer;
-                buffer.append("[");
+                buffer.append(useAssignmentPatterns ? "'{"sv : "["sv);
                 for (auto& element : *arg) {
-                    buffer.append(element.toString());
+                    buffer.append(element.toString(abbreviateThresholdBits, exactUnknowns,
+                                                   useAssignmentPatterns));
                     buffer.append(",");
                 }
 
                 if (!arg->empty())
                     buffer.pop_back();
-                buffer.append("]");
+                buffer.append(useAssignmentPatterns ? "}"sv : "]"sv);
                 return buffer.str();
             }
             else if constexpr (std::is_same_v<T, Union>) {
                 if (!arg->activeMember)
                     return "(unset)"s;
 
-                return fmt::format("({}) {}", *arg->activeMember, arg->value.toString());
+                return fmt::format("({}) {}", *arg->activeMember,
+                                   arg->value.toString(abbreviateThresholdBits, exactUnknowns,
+                                                       useAssignmentPatterns));
             }
             else {
                 static_assert(always_false<T>::value, "Missing case");
@@ -99,9 +112,9 @@ size_t ConstantValue::hash() const {
             else if constexpr (std::is_same_v<T, SVInt>)
                 hash_combine(h, arg.hash());
             else if constexpr (std::is_same_v<T, real_t>)
-                hash_combine(h, std::hash<double>()(arg));
+                hash_combine(h, slang::hash<double>()(arg));
             else if constexpr (std::is_same_v<T, shortreal_t>)
-                hash_combine(h, std::hash<float>()(arg));
+                hash_combine(h, slang::hash<float>()(arg));
             else if constexpr (std::is_same_v<T, ConstantValue::NullPlaceholder>)
                 hash_combine(h, 0);
             else if constexpr (std::is_same_v<T, ConstantValue::UnboundedPlaceholder>)
@@ -111,7 +124,7 @@ size_t ConstantValue::hash() const {
                     hash_combine(h, element.hash());
             }
             else if constexpr (std::is_same_v<T, std::string>)
-                hash_combine(h, std::hash<std::string>()(arg));
+                hash_combine(h, slang::hash<std::string>()(arg));
             else if constexpr (std::is_same_v<T, Map>) {
                 for (auto& [key, val] : *arg) {
                     hash_combine(h, key.hash());
@@ -152,6 +165,8 @@ size_t ConstantValue::size() const {
                 return arg->size();
             else if constexpr (std::is_same_v<T, Queue>)
                 return arg->size();
+            else if constexpr (std::is_same_v<T, std::string>)
+                return arg.size();
             else
                 return size_t(0);
         },
@@ -167,7 +182,7 @@ ConstantValue& ConstantValue::at(size_t index) {
             else if constexpr (std::is_same_v<T, Queue>)
                 return arg->at(index);
             else
-                THROW_UNREACHABLE;
+                SLANG_UNREACHABLE;
         },
         value);
 }
@@ -181,7 +196,7 @@ const ConstantValue& ConstantValue::at(size_t index) const {
             else if constexpr (std::is_same_v<T, Queue>)
                 return arg->at(index);
             else
-                THROW_UNREACHABLE;
+                SLANG_UNREACHABLE;
         },
         value);
 }
@@ -192,8 +207,8 @@ ConstantValue ConstantValue::getSlice(int32_t upper, int32_t lower,
         return integer().slice(upper, lower);
 
     if (isUnpacked()) {
-        span<const ConstantValue> elems = elements();
-        std::vector<ConstantValue> result{ size_t(upper - lower + 1) };
+        std::span<const ConstantValue> elems = elements();
+        std::vector<ConstantValue> result{size_t(upper - lower + 1)};
         ConstantValue* dest = result.data();
 
         for (int32_t i = lower; i <= upper; i++) {
@@ -218,8 +233,8 @@ ConstantValue ConstantValue::getSlice(int32_t upper, int32_t lower,
     }
 
     if (isString()) {
-        ASSERT(upper == lower);
-        ASSERT(upper >= 0);
+        SLANG_ASSERT(upper == lower);
+        SLANG_ASSERT(upper >= 0);
         return SVInt(8, (uint64_t)str()[size_t(upper)], false);
     }
 
@@ -305,7 +320,6 @@ ConstantValue ConstantValue::convertToInt() const {
     if (isShortReal())
         return convertToInt(32, true, false);
     if (isString()) {
-        // TODO: overflow
         bitwidth_t bits = bitwidth_t(str().length() * 8);
         if (!bits)
             bits = 8;
@@ -326,10 +340,10 @@ ConstantValue ConstantValue::convertToInt(bitwidth_t width, bool isSigned, bool
         return SVInt::fromFloat(width, shortReal(), isSigned);
 
     if (isString()) {
-        SmallVectorSized<byte, 8> buffer;
+        SmallVector<byte, 8> buffer;
         auto& s = str();
         for (auto it = s.rbegin(); it != s.rend(); it++)
-            buffer.append(static_cast<byte>(*it));
+            buffer.push_back(static_cast<byte>(*it));
 
         SVInt result(width, buffer, isSigned);
         if (!isFourState)
@@ -464,29 +478,32 @@ ConstantValue ConstantValue::convertToByteQueue(bool isSigned) const {
     return queue;
 }
 
-size_t ConstantValue::bitstreamWidth() const {
+uint64_t ConstantValue::getBitstreamWidth() const {
+    // Note that we don't have to worry about overflow in this
+    // method because we have an artificial limit on how
+    // large constant values are allowed to be.
+    // TODO: actually implement the mentioned limit
     if (isInteger())
         return integer().getBitWidth();
 
-    // TODO: check for overflow
     if (isString())
         return str().length() * CHAR_BIT;
 
-    size_t width = 0;
+    uint64_t width = 0;
     if (isUnpacked()) {
         for (const auto& cv : elements())
-            width += cv.bitstreamWidth();
+            width += cv.getBitstreamWidth();
     }
     else if (isMap()) {
         for (const auto& kv : *map())
-            width += kv.second.bitstreamWidth();
+            width += kv.second.getBitstreamWidth();
     }
     else if (isQueue()) {
         for (const auto& cv : *queue())
-            width += cv.bitstreamWidth();
+            width += cv.getBitstreamWidth();
     }
     else if (isUnion()) {
-        width = unionVal()->value.bitstreamWidth();
+        width = unionVal()->value.getBitstreamWidth();
     }
 
     return width;
@@ -546,52 +563,62 @@ bool operator==(const ConstantValue& lhs, const ConstantValue& rhs) {
         lhs.value);
 }
 
-bool operator!=(const ConstantValue& lhs, const ConstantValue& rhs) {
-    return !(lhs == rhs);
-}
-
-bool operator<(const ConstantValue& lhs, const ConstantValue& rhs) {
+std::partial_ordering operator<=>(const ConstantValue& lhs, const ConstantValue& rhs) {
     return std::visit(
-        [&](auto&& arg) {
+        [&](auto&& arg) -> std::partial_ordering {
+            constexpr auto unordered = std::partial_ordering::unordered;
             using T = std::decay_t<decltype(arg)>;
             if constexpr (std::is_same_v<T, std::monostate>)
-                return false;
-            else if constexpr (std::is_same_v<T, SVInt>)
-                return rhs.isInteger() && bool(arg < rhs.integer());
+                return unordered;
+            else if constexpr (std::is_same_v<T, SVInt>) {
+                if (!rhs.isInteger())
+                    return unordered;
+
+                auto& rhi = rhs.integer();
+                if (arg < rhi)
+                    return std::partial_ordering::less;
+                else if (arg == rhi)
+                    return std::partial_ordering::equivalent;
+                else
+                    return std::partial_ordering::greater;
+            }
             else if constexpr (std::is_same_v<T, real_t>)
-                return rhs.isReal() && arg < double(rhs.real());
+                return rhs.isReal() ? arg <=> double(rhs.real()) : unordered;
             else if constexpr (std::is_same_v<T, shortreal_t>)
-                return rhs.isShortReal() && arg < float(rhs.shortReal());
+                return rhs.isShortReal() ? arg <=> float(rhs.shortReal()) : unordered;
             else if constexpr (std::is_same_v<T, ConstantValue::NullPlaceholder>)
-                return false;
+                return unordered;
             else if constexpr (std::is_same_v<T, ConstantValue::UnboundedPlaceholder>)
-                return false;
+                return unordered;
             else if constexpr (std::is_same_v<T, ConstantValue::Elements>) {
                 if (!rhs.isUnpacked())
-                    return false;
+                    return unordered;
 
-                return arg < std::get<ConstantValue::Elements>(rhs.value);
+                return arg <=> std::get<ConstantValue::Elements>(rhs.value);
+            }
+            else if constexpr (std::is_same_v<T, std::string>) {
+                if (!rhs.isString())
+                    return unordered;
+
+                return arg <=> rhs.str();
             }
-            else if constexpr (std::is_same_v<T, std::string>)
-                return rhs.isString() && arg < rhs.str();
             else if constexpr (std::is_same_v<T, ConstantValue::Map>) {
                 if (!rhs.isMap())
-                    return false;
+                    return unordered;
 
-                return *arg < *rhs.map();
+                return *arg <=> *rhs.map();
             }
             else if constexpr (std::is_same_v<T, ConstantValue::Queue>) {
                 if (!rhs.isQueue())
-                    return false;
+                    return unordered;
 
-                return *arg < *rhs.queue();
+                return *arg <=> *rhs.queue();
             }
             else if constexpr (std::is_same_v<T, ConstantValue::Union>) {
                 if (!rhs.isUnion())
-                    return false;
+                    return unordered;
 
-                auto& ru = rhs.unionVal();
-                return arg->activeMember < ru->activeMember && arg->value < ru->value;
+                return *arg <=> *rhs.unionVal();
             }
             else {
                 static_assert(always_false<T>::value, "Missing case");
@@ -600,171 +627,29 @@ bool operator<(const ConstantValue& lhs, const ConstantValue& rhs) {
         lhs.value);
 }
 
-const ConstantValue& CVIterator::operator*() const {
-    return std::visit(
-        [](auto&& arg) -> const ConstantValue& {
-            using T = std::decay_t<decltype(arg)>;
-            if constexpr (std::is_same_v<T, AssociativeArray::iterator>)
-                return arg->second;
-            else
-                return *arg;
-        },
-        current);
-}
-
-ConstantValue& CVIterator::operator*() {
-    return std::visit(
-        [](auto&& arg) -> ConstantValue& {
-            using T = std::decay_t<decltype(arg)>;
-            if constexpr (std::is_same_v<T, AssociativeArray::iterator>)
-                return arg->second;
-            else
-                return *arg;
-        },
-        current);
-}
-
-const ConstantValue& CVIterator::key() const {
-    return std::visit(
-        [](auto&& arg) -> const ConstantValue& {
-            using T = std::decay_t<decltype(arg)>;
-            if constexpr (std::is_same_v<T, AssociativeArray::iterator>)
-                return arg->first;
-            else
-                return *arg;
-        },
-        current);
-}
-
-CVIterator& CVIterator::operator++() {
-    std::visit([](auto&& arg) { ++arg; }, current);
-    return *this;
-}
-
-CVIterator& CVIterator::operator--() {
-    std::visit([](auto&& arg) { --arg; }, current);
-    return *this;
-}
-
-const ConstantValue& CVConstIterator::operator*() const {
-    return std::visit(
-        [](auto&& arg) -> const ConstantValue& {
-            using T = std::decay_t<decltype(arg)>;
-            if constexpr (std::is_same_v<T, AssociativeArray::const_iterator>)
-                return arg->second;
-            else
-                return *arg;
-        },
-        current);
-}
-
-const ConstantValue& CVConstIterator::key() const {
-    return std::visit(
-        [](auto&& arg) -> const ConstantValue& {
-            using T = std::decay_t<decltype(arg)>;
-            if constexpr (std::is_same_v<T, AssociativeArray::const_iterator>)
-                return arg->first;
-            else
-                return *arg;
-        },
-        current);
-}
-
-CVConstIterator& CVConstIterator::operator++() {
-    std::visit([](auto&& arg) { ++arg; }, current);
-    return *this;
-}
-
-CVConstIterator& CVConstIterator::operator--() {
-    std::visit([](auto&& arg) { --arg; }, current);
-    return *this;
-}
-
-CVIterator begin(ConstantValue& cv) {
-    return std::visit(
-        [](auto&& arg) -> CVIterator {
-            using T = std::decay_t<decltype(arg)>;
-            if constexpr (std::is_same_v<T, ConstantValue::Elements>) {
-                return arg.begin();
-            }
-            else if constexpr (std::is_same_v<T, ConstantValue::Map> ||
-                               std::is_same_v<T, ConstantValue::Queue>) {
-                return arg->begin();
-            }
-            else {
-                THROW_UNREACHABLE;
-            }
-        },
-        cv.getVariant());
-}
-
-CVIterator end(ConstantValue& cv) {
-    return std::visit(
-        [](auto&& arg) -> CVIterator {
-            using T = std::decay_t<decltype(arg)>;
-            if constexpr (std::is_same_v<T, ConstantValue::Elements>) {
-                return arg.end();
-            }
-            else if constexpr (std::is_same_v<T, ConstantValue::Map> ||
-                               std::is_same_v<T, ConstantValue::Queue>) {
-                return arg->end();
-            }
-            else {
-                THROW_UNREACHABLE;
-            }
-        },
-        cv.getVariant());
-}
-
-CVConstIterator begin(const ConstantValue& cv) {
-    return std::visit(
-        [](auto&& arg) -> CVConstIterator {
-            using T = std::decay_t<decltype(arg)>;
-            if constexpr (std::is_same_v<T, ConstantValue::Elements>) {
-                return arg.begin();
-            }
-            else if constexpr (std::is_same_v<T, ConstantValue::Map> ||
-                               std::is_same_v<T, ConstantValue::Queue>) {
-                return arg->begin();
-            }
-            else {
-                THROW_UNREACHABLE;
-            }
-        },
-        cv.getVariant());
-}
-
-CVConstIterator end(const ConstantValue& cv) {
-    return std::visit(
-        [](auto&& arg) -> CVConstIterator {
-            using T = std::decay_t<decltype(arg)>;
-            if constexpr (std::is_same_v<T, ConstantValue::Elements>) {
-                return arg.end();
-            }
-            else if constexpr (std::is_same_v<T, ConstantValue::Map> ||
-                               std::is_same_v<T, ConstantValue::Queue>) {
-                return arg->end();
-            }
-            else {
-                THROW_UNREACHABLE;
-            }
-        },
-        cv.getVariant());
-}
-
 ConstantRange ConstantRange::subrange(ConstantRange select) const {
     int32_t l = lower();
     ConstantRange result;
     result.left = select.lower() + l;
     result.right = select.upper() + l;
 
-    ASSERT(result.right <= upper());
+    SLANG_ASSERT(result.right <= upper());
     if (isLittleEndian())
         return result;
     else
         return result.reverse();
 }
 
+ConstantRange ConstantRange::intersect(ConstantRange other) const {
+    if (overlaps(other)) {
+        ConstantRange result;
+        result.left = std::max(lower(), other.lower());
+        result.right = std::min(upper(), other.upper());
+        return result;
+    }
+    return ConstantRange();
+}
+
 int32_t ConstantRange::translateIndex(int32_t index) const {
     if (!isLittleEndian())
         return upper() - index;
@@ -776,18 +661,22 @@ bool ConstantRange::containsPoint(int32_t index) const {
     return index >= lower() && index <= upper();
 }
 
+bool ConstantRange::contains(ConstantRange other) const {
+    return other.lower() >= lower() && other.upper() <= upper();
+}
+
 bool ConstantRange::overlaps(ConstantRange other) const {
     return lower() <= other.upper() && upper() >= other.lower();
 }
 
-ConstantRange ConstantRange::getIndexedRange(int32_t l, int32_t r, bool littleEndian,
-                                             bool indexedUp) {
+std::optional<ConstantRange> ConstantRange::getIndexedRange(int32_t l, int32_t r, bool littleEndian,
+                                                            bool indexedUp) {
     ConstantRange result;
     int32_t count = r - 1;
     if (indexedUp) {
         auto upper = checkedAddS32(l, count);
         if (!upper)
-            upper = INT32_MAX;
+            return std::nullopt;
 
         result.left = *upper;
         result.right = l;
@@ -795,7 +684,7 @@ ConstantRange ConstantRange::getIndexedRange(int32_t l, int32_t r, bool littleEn
     else {
         auto lower = checkedSubS32(l, count);
         if (!lower)
-            lower = INT32_MIN;
+            return std::nullopt;
 
         result.left = l;
         result.right = *lower;
