@@ -2,12 +2,15 @@
 // SVInt.cpp
 // Arbitrary precision integer support
 //
-// File is under the MIT license; see LICENSE for details
+// SPDX-FileCopyrightText: Michael Popoloski
+// SPDX-License-Identifier: MIT
 //------------------------------------------------------------------------------
 #include "slang/numeric/SVInt.h"
 
 #include "SVIntHelpers.h"
-#include <fmt/format.h>
+#include <cmath>
+#include <fmt/core.h>
+#include <ostream>
 #include <stdexcept>
 
 #include "slang/text/CharInfo.h"
@@ -18,8 +21,8 @@ static const double log2_10 = log2(10.0);
 
 namespace slang {
 
-const logic_t logic_t::x{ logic_t::X_VALUE };
-const logic_t logic_t::z{ logic_t::Z_VALUE };
+const logic_t logic_t::x{logic_t::X_VALUE};
+const logic_t logic_t::z{logic_t::Z_VALUE};
 
 const SVInt SVInt::Zero = 0u;
 const SVInt SVInt::One = 1u;
@@ -67,9 +70,9 @@ std::ostream& operator<<(std::ostream& os, const logic_t& rhs) {
     return os;
 }
 
-SVInt SVInt::fromString(string_view str) {
+SVInt SVInt::fromString(std::string_view str) {
     if (str.empty())
-        throw std::invalid_argument("String is empty");
+        SLANG_THROW(std::invalid_argument("String is empty"));
 
     const char* c = str.data();
     const char* end = c + str.length();
@@ -77,7 +80,7 @@ SVInt SVInt::fromString(string_view str) {
     if (*c == '-' || *c == '+') {
         c++; // heh
         if (c == end)
-            throw std::invalid_argument("String only has a sign?");
+            SLANG_THROW(std::invalid_argument("String only has a sign?"));
     }
 
     // look for a base specifier (optional)
@@ -116,38 +119,38 @@ SVInt SVInt::fromString(string_view str) {
 
     if (apostrophe) {
         if (sizeBad || sizeOverflow || possibleSize == 0)
-            throw std::invalid_argument("Size is invalid (bad chars or overflow 16 bits)");
+            SLANG_THROW(std::invalid_argument("Size is invalid (bad chars or overflow 16 bits)"));
         bits = possibleSize;
 
         c = apostrophe + 1;
         if (c == end)
-            throw std::invalid_argument("Nothing after size specifier");
+            SLANG_THROW(std::invalid_argument("Nothing after size specifier"));
 
         if (*c == 's' || *c == 'S') {
             isSigned = true;
             c++;
             if (c == end)
-                throw std::invalid_argument("Nothing after sign specifier");
+                SLANG_THROW(std::invalid_argument("Nothing after sign specifier"));
         }
         else {
             isSigned = false;
         }
 
         if (!literalBaseFromChar(*c, base))
-            throw std::invalid_argument(fmt::format("Unknown base specifier '{}'", *c));
+            SLANG_THROW(std::invalid_argument(fmt::format("Unknown base specifier '{}'", *c)));
 
         c++;
         if (c == end)
-            throw std::invalid_argument("Nothing after base specifier");
+            SLANG_THROW(std::invalid_argument("Nothing after base specifier"));
     }
     else if (sizeBad) {
-        throw std::invalid_argument("Not an integer or sized literal");
+        SLANG_THROW(std::invalid_argument("Not an integer or sized literal"));
     }
 
     // convert the remaining chars to an array of digits to pass to the other
     // constructor
     bool isUnknown = false;
-    SmallVectorSized<logic_t, 16> digits;
+    SmallVector<logic_t> digits;
     for (; c != end; ++c) {
         char d = *c;
         if (d == '_')
@@ -170,7 +173,7 @@ SVInt SVInt::fromString(string_view str) {
                 value = logic_t(getHexDigitValue(d));
                 break;
         }
-        digits.append(value);
+        digits.push_back(value);
     }
 
     SVInt result = fromDigits(bits, base, isSigned, isUnknown, digits);
@@ -183,9 +186,9 @@ SVInt SVInt::fromString(string_view str) {
 }
 
 SVInt SVInt::fromDigits(bitwidth_t bits, LiteralBase base, bool isSigned, bool anyUnknown,
-                        span<logic_t const> digits) {
+                        std::span<logic_t const> digits) {
     if (digits.empty())
-        throw std::invalid_argument("No digits provided");
+        SLANG_THROW(std::invalid_argument("No digits provided"));
 
     // Note: If the user specified a number too large to fit in the number of bits specified,
     // the spec says to truncate from the left, which this method will successfully do.
@@ -224,8 +227,8 @@ SVInt SVInt::fromDigits(bitwidth_t bits, LiteralBase base, bool isSigned, bool a
 
             val += d.value;
             if (d.value >= radix) {
-                throw std::invalid_argument(
-                    fmt::format("Digit {} too large for radix {}", d.value, radix));
+                SLANG_THROW(std::invalid_argument(
+                    fmt::format("Digit {} too large for radix {}", d.value, radix)));
             }
         }
         return SVInt(bits, val, isSigned);
@@ -235,8 +238,8 @@ SVInt SVInt::fromDigits(bitwidth_t bits, LiteralBase base, bool isSigned, bool a
         // In base ten we can't have individual bits be X or Z, it's all or nothing
         if (anyUnknown) {
             if (digits.size() != 1) {
-                throw std::invalid_argument(
-                    "If a decimal number is unknown, it must have exactly one digit.");
+                SLANG_THROW(std::invalid_argument(
+                    "If a decimal number is unknown, it must have exactly one digit."));
             }
 
             if (exactlyEqual(digits[0], logic_t::z))
@@ -251,7 +254,7 @@ SVInt SVInt::fromDigits(bitwidth_t bits, LiteralBase base, bool isSigned, bool a
     return fromPow2Digits(bits, isSigned, anyUnknown, radix, shift, digits);
 }
 
-SVInt SVInt::fromDecimalDigits(bitwidth_t bits, bool isSigned, span<logic_t const> digits) {
+SVInt SVInt::fromDecimalDigits(bitwidth_t bits, bool isSigned, std::span<logic_t const> digits) {
     SVInt result = allocZeroed(bits, isSigned, false);
 
     constexpr int charsPerWord = 18; // 18 decimal digits can fit in a 64-bit word
@@ -263,7 +266,8 @@ SVInt SVInt::fromDecimalDigits(bitwidth_t bits, bool isSigned, span<logic_t cons
     auto nextDigit = [&]() {
         uint8_t v = d->value;
         if (v >= 10) {
-            throw std::invalid_argument(fmt::format("Digit {} too large for radix {}", v, 10));
+            SLANG_THROW(
+                std::invalid_argument(fmt::format("Digit {} too large for radix {}", v, 10)));
         }
         d++;
         return v;
@@ -305,7 +309,7 @@ SVInt SVInt::fromDecimalDigits(bitwidth_t bits, bool isSigned, span<logic_t cons
 }
 
 SVInt SVInt::fromPow2Digits(bitwidth_t bits, bool isSigned, bool anyUnknown, uint32_t radix,
-                            uint32_t shift, span<logic_t const> digits) {
+                            uint32_t shift, std::span<logic_t const> digits) {
 
     SVInt result = allocZeroed(bits, isSigned, anyUnknown);
 
@@ -331,8 +335,8 @@ SVInt SVInt::fromPow2Digits(bitwidth_t bits, bool isSigned, bool anyUnknown, uin
             unknown = ones;
         }
         else if (value >= radix) {
-            throw std::invalid_argument(
-                fmt::format("Digit {} too large for radix {}", value, radix));
+            SLANG_THROW(std::invalid_argument(
+                fmt::format("Digit {} too large for radix {}", value, radix)));
         }
 
         word |= value << bitPos;
@@ -398,12 +402,12 @@ SVInt SVInt::fromPow2Digits(bitwidth_t bits, bool isSigned, bool anyUnknown, uin
     return result;
 }
 
-SVInt SVInt::fromDouble(bitwidth_t bits, double value, bool isSigned) {
-    return fromIEEE754<double, 11, 52, 1023>(bits, value, isSigned);
+SVInt SVInt::fromDouble(bitwidth_t bits, double value, bool isSigned, bool round) {
+    return fromIEEE754<double, 11, 52, 1023>(bits, value, isSigned, round);
 }
 
-SVInt SVInt::fromFloat(bitwidth_t bits, float value, bool isSigned) {
-    return fromIEEE754<float, 8, 23, 127>(bits, value, isSigned);
+SVInt SVInt::fromFloat(bitwidth_t bits, float value, bool isSigned, bool round) {
+    return fromIEEE754<float, 8, 23, 127>(bits, value, isSigned, round);
 }
 
 double SVInt::toDouble() const {
@@ -620,14 +624,14 @@ SVInt SVInt::ashr(bitwidth_t amount) const {
 
 SVInt SVInt::replicate(const SVInt& times) const {
     uint32_t n = times.as<uint32_t>().value();
-    SmallVectorSized<SVInt, 8> buffer;
+    SmallVector<SVInt> buffer(n, UninitializedTag());
     for (size_t i = 0; i < n; ++i)
-        buffer.append(*this);
-    return concat(span<SVInt const>(buffer.begin(), buffer.end()));
+        buffer.push_back(*this);
+    return concat(std::span<SVInt const>(buffer.begin(), buffer.end()));
 }
 
 size_t SVInt::hash() const {
-    return xxhash(getRawData(), getNumWords() * WORD_SIZE);
+    return (size_t)detail::hashing::hash(getRawData(), getNumWords() * WORD_SIZE);
 }
 
 std::ostream& operator<<(std::ostream& os, const SVInt& rhs) {
@@ -635,48 +639,54 @@ std::ostream& operator<<(std::ostream& os, const SVInt& rhs) {
     return os;
 }
 
-std::string SVInt::toString() const {
+std::string SVInt::toString(bitwidth_t abbreviateThresholdBits, bool exactUnknowns) const {
     // guess the base to use
+    // unknown bits require binary base for lossless representation
     LiteralBase base;
-    if (bitWidth < 8 || (unknownFlag && bitWidth <= 64))
+    if ((bitWidth < 8 && !signFlag) || (unknownFlag && exactUnknowns) ||
+        (unknownFlag && bitWidth <= 64)) {
         base = LiteralBase::Binary;
-    else if (bitWidth <= 32 || signFlag)
+    }
+    else if (bitWidth <= 32 || signFlag) {
         base = LiteralBase::Decimal;
-    else
+    }
+    else {
         base = LiteralBase::Hex;
+    }
 
-    return toString(base);
+    return toString(base, abbreviateThresholdBits);
 }
 
-std::string SVInt::toString(LiteralBase base) const {
+std::string SVInt::toString(LiteralBase base, bitwidth_t abbreviateThresholdBits) const {
     // Append the base unless we're a signed 32-bit base 10 integer.
     bool includeBase = base != LiteralBase::Decimal || bitWidth != 32 || !signFlag || unknownFlag;
 
-    SmallVectorSized<char, 32> buffer;
-    writeTo(buffer, base, includeBase, DefaultStringAbbreviationThresholdBits);
+    SmallVector<char> buffer;
+    writeTo(buffer, base, includeBase, abbreviateThresholdBits);
     return std::string(buffer.begin(), buffer.end());
 }
 
 std::string SVInt::toString(LiteralBase base, bool includeBase,
                             bitwidth_t abbreviateThresholdBits) const {
-    SmallVectorSized<char, 32> buffer;
+    SmallVector<char> buffer;
     writeTo(buffer, base, includeBase, abbreviateThresholdBits);
     return std::string(buffer.begin(), buffer.end());
 }
 
-void SVInt::writeTo(SmallVector<char>& buffer, LiteralBase base) const {
+void SVInt::writeTo(SmallVectorBase<char>& buffer, LiteralBase base,
+                    bitwidth_t abbreviateThresholdBits) const {
     // Append the base unless we're a signed 32-bit base 10 integer.
     bool includeBase = base != LiteralBase::Decimal || bitWidth != 32 || !signFlag || unknownFlag;
-    writeTo(buffer, base, includeBase);
+    writeTo(buffer, base, includeBase, abbreviateThresholdBits);
 }
 
-void SVInt::writeTo(SmallVector<char>& buffer, LiteralBase base, bool includeBase,
+void SVInt::writeTo(SmallVectorBase<char>& buffer, LiteralBase base, bool includeBase,
                     bitwidth_t abbreviateThresholdBits) const {
     // negative sign if necessary
     SVInt tmp(*this);
     if (signFlag && !unknownFlag && isNegative()) {
         tmp = -tmp;
-        buffer.append('-');
+        buffer.push_back('-');
     }
 
     // If the number is larger than the threshold to abbreviate,
@@ -689,21 +699,21 @@ void SVInt::writeTo(SmallVector<char>& buffer, LiteralBase base, bool includeBas
 
     if (includeBase) {
         uintToStr(buffer, bitWidth);
-        buffer.append('\'');
+        buffer.push_back('\'');
         if (signFlag)
-            buffer.append('s');
+            buffer.push_back('s');
         switch (base) {
             case LiteralBase::Binary:
-                buffer.append('b');
+                buffer.push_back('b');
                 break;
             case LiteralBase::Octal:
-                buffer.append('o');
+                buffer.push_back('o');
                 break;
             case LiteralBase::Decimal:
-                buffer.append('d');
+                buffer.push_back('d');
                 break;
             case LiteralBase::Hex:
-                buffer.append('h');
+                buffer.push_back('h');
                 break;
         }
     }
@@ -744,13 +754,13 @@ void SVInt::writeTo(SmallVector<char>& buffer, LiteralBase base, bool includeBas
 
             bool upperOnes = all(words, UINT64_MAX);
             if (upperOnes && all(0, UINT64_MAX))
-                buffer.append('z');
+                buffer.push_back('z');
             else if (upperOnes && all(0, 0))
-                buffer.append('x');
+                buffer.push_back('x');
             else if (anyXs())
-                buffer.append('X');
+                buffer.push_back('X');
             else
-                buffer.append('Z');
+                buffer.push_back('Z');
         }
         else {
             // Limit the size of the output string based on the given threshold.
@@ -765,7 +775,6 @@ void SVInt::writeTo(SmallVector<char>& buffer, LiteralBase base, bool includeBas
                 divide(tmp, tmp.getNumWords(), divisor, divisor.getNumWords(), &quotient,
                        &remainder);
                 tmp = quotient;
-                activeBits = tmp.getActiveBits();
             }
 
             // repeatedly divide by 10 to get each digit
@@ -776,8 +785,8 @@ void SVInt::writeTo(SmallVector<char>& buffer, LiteralBase base, bool includeBas
                 divide(tmp, tmp.getNumWords(), divisor, divisor.getNumWords(), &quotient,
                        &remainder);
                 uint64_t digit = remainder.as<uint64_t>().value();
-                ASSERT(digit < 10);
-                buffer.append(Digits[digit]);
+                SLANG_ASSERT(digit < 10);
+                buffer.push_back(Digits[digit]);
                 tmp = quotient;
             }
         }
@@ -800,7 +809,7 @@ void SVInt::writeTo(SmallVector<char>& buffer, LiteralBase base, bool includeBas
                 maskAmount = 15;
                 break;
             case LiteralBase::Decimal:
-                THROW_UNREACHABLE;
+                SLANG_UNREACHABLE;
         }
 
         // if we have unknown values here, the comparison will return X
@@ -813,19 +822,19 @@ void SVInt::writeTo(SmallVector<char>& buffer, LiteralBase base, bool includeBas
 
             uint32_t digit = uint32_t(tmp.getRawData()[0]) & maskAmount;
             if (!tmp.unknownFlag)
-                buffer.append(Digits[digit]);
+                buffer.push_back(Digits[digit]);
             else {
                 uint32_t u = uint32_t(tmp.pVal[getNumWords(bitWidth, false)]) & maskAmount;
                 if (!u)
-                    buffer.append(Digits[digit]);
+                    buffer.push_back(Digits[digit]);
                 else if (u == maskAmount && (digit & maskAmount) == 0)
-                    buffer.append('x');
+                    buffer.push_back('x');
                 else if (u == maskAmount && digit == maskAmount)
-                    buffer.append('z');
+                    buffer.push_back('z');
                 else if (~digit & u)
-                    buffer.append('X');
+                    buffer.push_back('X');
                 else
-                    buffer.append('Z');
+                    buffer.push_back('Z');
             }
             // this shift might shift away the unknown digits, at which point
             // it converts back to being a normal 2-state value
@@ -833,16 +842,22 @@ void SVInt::writeTo(SmallVector<char>& buffer, LiteralBase base, bool includeBas
             bitsLeft -= int(shiftAmount);
             x = tmp != 0;
         }
+
+        // If there are bits left over and the last digit we pushed was
+        // an unknown we need to insert an extra 0 to indicate that the
+        // leading bits are actually zeroes and not extended unknowns.
+        if (bitsLeft > 0 && !buffer.empty() && (buffer.back() == 'x' || buffer.back() == 'z'))
+            buffer.push_back('0');
     }
 
     // no digits means this is zero
     if (startOffset == buffer.size())
-        buffer.append('0');
+        buffer.push_back('0');
     else {
         // reverse the digits
-        std::reverse(buffer.begin() + startOffset, buffer.end());
+        std::ranges::reverse(buffer.begin() + startOffset, buffer.end());
         if (base10Exponent > 0) {
-            buffer.appendRange("...e"sv);
+            buffer.append_range("...e"sv);
             uintToStr(buffer, base10Exponent);
         }
     }
@@ -895,6 +910,11 @@ SVInt SVInt::pow(const SVInt& rhs) const {
         else
             return modPow(-(*this), rhs, signFlag);
     }
+
+    // Optimization for the common case of 2**y
+    if (lhsBits == 2 && getRawData()[0] == 2)
+        return SVInt(bitWidth, 1, signFlag).shl(rhs);
+
     return modPow(*this, rhs, signFlag);
 }
 
@@ -1030,7 +1050,7 @@ SVInt& SVInt::operator+=(const SVInt& rhs) {
     return *this;
 }
 
-NO_SANITIZE("unsigned-integer-overflow")
+SLANG_NO_SANITIZE("unsigned-integer-overflow")
 SVInt& SVInt::operator-=(const SVInt& rhs) {
     if (bitWidth != rhs.bitWidth) {
         if (bitWidth < rhs.bitWidth)
@@ -1380,19 +1400,12 @@ logic_t SVInt::operator<(const SVInt& rhs) const {
         else
             return *this < rhs.extend(bitWidth, bothSigned);
     }
+    // handle signed negatives
+    if (bothSigned && isNegative() ^ rhs.isNegative())
+        return logic_t(isNegative());
 
-    if (bothSigned) {
-        // handle negatives
-        if (isNegative()) {
-            if (rhs.isNegative())
-                return -(*this) > -rhs;
-            else
-                return logic_t(true);
-        }
-        if (rhs.isNegative())
-            return logic_t(false);
-    }
-
+    // both are positive or both are negative
+    // or not both are signed
     if (isSingleWord())
         return logic_t(val < rhs.val);
 
@@ -1441,7 +1454,7 @@ logic_t SVInt::operator[](int32_t index) const {
 }
 
 SVInt SVInt::slice(int32_t msb, int32_t lsb) const {
-    ASSERT(msb >= lsb);
+    SLANG_ASSERT(msb >= lsb);
 
     // handle indexing out of bounds
     bitwidth_t selectWidth = bitwidth_t(msb - lsb + 1);
@@ -1490,10 +1503,10 @@ SVInt SVInt::slice(int32_t msb, int32_t lsb) const {
 }
 
 void SVInt::set(int32_t msb, int32_t lsb, const SVInt& value) {
-    ASSERT(msb >= lsb);
+    SLANG_ASSERT(msb >= lsb);
 
     bitwidth_t selectWidth = bitwidth_t(msb - lsb + 1);
-    ASSERT(value.getBitWidth() == selectWidth);
+    SLANG_ASSERT(value.getBitWidth() == selectWidth);
     if (msb < 0 || lsb >= int32_t(bitWidth))
         return;
 
@@ -1529,7 +1542,7 @@ void SVInt::set(int32_t msb, int32_t lsb, const SVInt& value) {
 }
 
 SVInt SVInt::sext(bitwidth_t bits) const {
-    ASSERT(bits > bitWidth);
+    SLANG_ASSERT(bits > bitWidth);
 
     if (bits <= SVInt::BITS_PER_WORD && !unknownFlag) {
         uint64_t newVal = val << (SVInt::BITS_PER_WORD - bitWidth); // NOLINT
@@ -1578,7 +1591,7 @@ bool SVInt::isSignExtendedFrom(bitwidth_t msb) const {
 }
 
 void SVInt::signExtendFrom(bitwidth_t msb) {
-    ASSERT(msb < bitWidth - 1);
+    SLANG_ASSERT(msb < bitWidth - 1);
 
     uint64_t maskMsw;
     bitwidth_t bitsInMsw;
@@ -1600,7 +1613,7 @@ void SVInt::signExtendFrom(bitwidth_t msb) {
 }
 
 SVInt SVInt::zext(bitwidth_t bits) const {
-    ASSERT(bits > bitWidth);
+    SLANG_ASSERT(bits > bitWidth);
 
     if (bits <= SVInt::BITS_PER_WORD && !unknownFlag)
         return SVInt(bits, val, signFlag);
@@ -1625,7 +1638,7 @@ SVInt SVInt::extend(bitwidth_t bits, bool useSign) const {
 }
 
 SVInt SVInt::trunc(bitwidth_t bits) const {
-    ASSERT(bits > 0 && bits <= bitWidth);
+    SLANG_ASSERT(bits > 0 && bits <= bitWidth);
 
     if (isSingleWord()) {
         uint64_t mask = bits == 64 ? UINT64_MAX : (1ull << bits) - 1;
@@ -1722,7 +1735,7 @@ SVInt SVInt::conditional(const SVInt& condition, const SVInt& lhs, const SVInt&
     return result;
 }
 
-SVInt SVInt::concat(span<SVInt const> operands) {
+SVInt SVInt::concat(std::span<SVInt const> operands) {
     // 0 operand concatenations can be valid inside of larger concatenations.
     if (operands.size() == 0)
         return SVInt::Zero;
@@ -1768,12 +1781,12 @@ SVInt SVInt::concat(span<SVInt const> operands) {
 }
 
 SVInt SVInt::allocUninitialized(bitwidth_t bits, bool signFlag, bool unknownFlag) {
-    ASSERT(bits && (bits > 64 || unknownFlag));
+    SLANG_ASSERT(bits && (bits > 64 || unknownFlag));
     return SVInt(new uint64_t[getNumWords(bits, unknownFlag)], bits, signFlag, unknownFlag);
 }
 
 SVInt SVInt::allocZeroed(bitwidth_t bits, bool signFlag, bool unknownFlag) {
-    ASSERT(bits && (bits > 64 || unknownFlag));
+    SLANG_ASSERT(bits && (bits > 64 || unknownFlag));
     return SVInt(new uint64_t[getNumWords(bits, unknownFlag)](), bits, signFlag, unknownFlag);
 }
 
@@ -1796,7 +1809,7 @@ void SVInt::initSlowCase(uint64_t value) {
     }
 }
 
-void SVInt::initSlowCase(span<const byte> bytes) {
+void SVInt::initSlowCase(std::span<const byte> bytes) {
     if (isSingleWord()) {
         val = 0;
         memcpy(&val, bytes.data(), std::min<size_t>(WORD_SIZE, bytes.size()));
@@ -1812,7 +1825,7 @@ void SVInt::initSlowCase(span<const byte> bytes) {
 void SVInt::initSlowCase(const SVIntStorage& other) {
     uint32_t words = getNumWords();
     pVal = new uint64_t[words];
-    std::copy(other.pVal, other.pVal + words, pVal);
+    std::ranges::copy(other.pVal, other.pVal + words, pVal);
 }
 
 SVInt& SVInt::assignSlowCase(const SVInt& rhs) {
@@ -1903,14 +1916,14 @@ bitwidth_t SVInt::countLeadingZerosSlowCase() const {
     uint32_t i = getNumWords();
     uint64_t part = pVal[i - 1] & mask;
     if (part)
-        return slang::countLeadingZeros64(part) - (BITS_PER_WORD - bitsInMsw);
+        return (bitwidth_t)std::countl_zero(part) - (BITS_PER_WORD - bitsInMsw);
 
     bitwidth_t count = bitsInMsw;
     for (--i; i > 0; --i) {
         if (pVal[i - 1] == 0)
             count += BITS_PER_WORD;
         else {
-            count += slang::countLeadingZeros64(pVal[i - 1]);
+            count += (bitwidth_t)std::countl_zero(pVal[i - 1]);
             break;
         }
     }
@@ -1926,13 +1939,13 @@ bitwidth_t SVInt::countLeadingOnesSlowCase() const {
         shift = BITS_PER_WORD - bitsInMsw;
 
     int i = int(getNumWords() - 1);
-    bitwidth_t count = slang::countLeadingOnes64(pVal[i] << shift);
+    bitwidth_t count = (bitwidth_t)std::countl_one(pVal[i] << shift);
     if (count == bitsInMsw) {
         for (i--; i >= 0; i--) {
             if (pVal[i] == UINT64_MAX)
                 count += BITS_PER_WORD;
             else {
-                count += slang::countLeadingOnes64(pVal[i]);
+                count += (bitwidth_t)std::countl_one(pVal[i]);
                 break;
             }
         }
@@ -1943,17 +1956,17 @@ bitwidth_t SVInt::countLeadingOnesSlowCase() const {
 
 bitwidth_t SVInt::countOnes() const {
     if (isSingleWord())
-        return slang::countPopulation64(val);
+        return (bitwidth_t)std::popcount(val);
 
     bitwidth_t count = 0;
     if (!unknownFlag) {
         for (uint32_t i = 0; i < getNumWords(); i++)
-            count += slang::countPopulation64(pVal[i]);
+            count += (bitwidth_t)std::popcount(pVal[i]);
     }
     else {
         uint32_t words = getNumWords(bitWidth, false);
         for (uint32_t i = 0; i < words; i++)
-            count += slang::countPopulation64(pVal[i] & ~pVal[i + words]);
+            count += (bitwidth_t)std::popcount(pVal[i] & ~pVal[i + words]);
     }
 
     return count;
@@ -1961,17 +1974,17 @@ bitwidth_t SVInt::countOnes() const {
 
 bitwidth_t SVInt::countZeros() const {
     if (isSingleWord())
-        return bitWidth - slang::countPopulation64(val);
+        return bitWidth - (bitwidth_t)std::popcount(val);
 
     bitwidth_t count = 0;
     if (!unknownFlag) {
         for (uint32_t i = 0; i < getNumWords(); i++)
-            count += slang::countPopulation64(~pVal[i]);
+            count += (bitwidth_t)std::popcount(~pVal[i]);
     }
     else {
         uint32_t words = getNumWords(bitWidth, false);
         for (uint32_t i = 0; i < words; i++)
-            count += slang::countPopulation64(~pVal[i] & ~pVal[i + words]);
+            count += (bitwidth_t)std::popcount(~pVal[i] & ~pVal[i + words]);
     }
 
     uint32_t wordBits = bitWidth % BITS_PER_WORD;
@@ -1988,7 +2001,7 @@ bitwidth_t SVInt::countXs() const {
     bitwidth_t count = 0;
     uint32_t words = getNumWords(bitWidth, false);
     for (uint32_t i = 0; i < words; i++)
-        count += slang::countPopulation64(~pVal[i] & pVal[i + words]);
+        count += (bitwidth_t)std::popcount(~pVal[i] & pVal[i + words]);
 
     return count;
 }
@@ -2000,7 +2013,7 @@ bitwidth_t SVInt::countZs() const {
     bitwidth_t count = 0;
     uint32_t words = getNumWords(bitWidth, false);
     for (uint32_t i = 0; i < words; i++)
-        count += slang::countPopulation64(pVal[i] & pVal[i + words]);
+        count += (bitwidth_t)std::popcount(pVal[i] & pVal[i + words]);
 
     return count;
 }
@@ -2094,15 +2107,15 @@ void SVInt::buildDivideResult(SVInt* result, uint32_t* value, bitwidth_t bitWidt
     else {
         *result = SVInt(bitWidth, 0, signFlag);
         for (uint32_t i = 0; i < numWords; i++)
-            result->pVal[i] =
-                uint64_t(value[i * 2]) | (uint64_t(value[i * 2 + 1]) << (BITS_PER_WORD / 2));
+            result->pVal[i] = uint64_t(value[i * 2]) |
+                              (uint64_t(value[i * 2 + 1]) << (BITS_PER_WORD / 2));
     }
 }
 
-NO_SANITIZE("unsigned-integer-overflow")
+SLANG_NO_SANITIZE("unsigned-integer-overflow")
 void SVInt::divide(const SVInt& lhs, uint32_t lhsWords, const SVInt& rhs, uint32_t rhsWords,
                    SVInt* quotient, SVInt* remainder) {
-    ASSERT(lhsWords >= rhsWords);
+    SLANG_ASSERT(lhsWords >= rhsWords);
 
     // The Knuth algorithm requires arrays of 32-bit words (because results of operations
     // need to fit natively into 64 bits). Allocate space for the backing memory, either on
