@@ -2,10 +2,12 @@
 // SVIntHelpers.h
 // Contains internal helper functions for SVInt implementations
 //
-// File is under the MIT license; see LICENSE for details
+// SPDX-FileCopyrightText: Michael Popoloski
+// SPDX-License-Identifier: MIT
 //------------------------------------------------------------------------------
 #pragma once
 
+#include <bit>
 #include <cstdint>
 #include <cstring>
 
@@ -38,6 +40,33 @@ static uint8_t subborrow64(uint8_t c, uint64_t a, uint64_t b, calc_out_t* out) {
 }
 #endif
 
+#if defined(_MSC_VER) && defined(_M_IX86)
+#    include <intrin.h>
+
+uint64_t _umul128(uint64_t multiplier, uint64_t multiplicand, uint64_t* product_hi) {
+    // multiplier   = ab = a * 2^32 + b
+    // multiplicand = cd = c * 2^32 + d
+    // ab * cd = a * c * 2^64 + (a * d + b * c) * 2^32 + b * d
+    uint32_t a = (uint32_t)(multiplier >> 32);
+    uint32_t b = (uint32_t)multiplier;
+    uint32_t c = (uint32_t)(multiplicand >> 32);
+    uint32_t d = (uint32_t)multiplicand;
+
+    uint64_t ad = __emulu(a, d);
+    uint64_t bd = __emulu(b, d);
+
+    uint64_t adbc = ad + __emulu(b, c);
+    uint64_t adbc_carry = (adbc < ad);
+
+    // multiplier * multiplicand = product_hi * 2^64 + product_lo
+    uint64_t product_lo = bd + (adbc << 32);
+    uint64_t product_lo_carry = (product_lo < bd);
+    *product_hi = __emulu(a, c) + (adbc >> 32) + (adbc_carry << 32) + product_lo_carry;
+
+    return product_lo;
+}
+#endif
+
 namespace slang {
 
 /// Provides a temporary storage region of dynamic size. If that size is less than
@@ -51,7 +80,7 @@ public:
             ptr = new T[size];
         else {
             ptr = reinterpret_cast<T*>(stackBase);
-            std::uninitialized_default_construct_n(ptr, size);
+            std::ranges::uninitialized_default_construct_n(ptr, ptrdiff_t(size));
         }
     }
 
@@ -61,7 +90,7 @@ public:
         if (size > StackCount)
             delete[] ptr;
         else
-            std::destroy_n(ptr, size);
+            std::ranges::destroy_n(ptr, ptrdiff_t(size));
     }
 
     T* get() const { return ptr; }
@@ -191,6 +220,8 @@ static bool addOne(uint64_t* dst, uint64_t* src, uint32_t len, uint64_t value) {
 
         if (!carry)
             break;
+
+        value = 0;
     }
     return carry;
 }
@@ -205,12 +236,14 @@ static bool subOne(uint64_t* dst, uint64_t* src, uint32_t len, uint64_t value) {
 
         if (!borrow)
             break;
+
+        value = 0;
     }
     return borrow;
 }
 
 // Generalized adder
-NO_SANITIZE("unsigned-integer-overflow")
+SLANG_NO_SANITIZE("unsigned-integer-overflow")
 static bool addGeneral(uint64_t* dst, const uint64_t* x, const uint64_t* y, uint32_t len) {
     uint8_t carry = 0;
     for (uint32_t i = 0; i < len; i++) {
@@ -222,7 +255,7 @@ static bool addGeneral(uint64_t* dst, const uint64_t* x, const uint64_t* y, uint
 }
 
 // Generalized subtractor (x - y)
-NO_SANITIZE("unsigned-integer-overflow")
+SLANG_NO_SANITIZE("unsigned-integer-overflow")
 static bool subGeneral(uint64_t* dst, const uint64_t* x, const uint64_t* y, uint32_t len) {
     uint8_t borrow = 0;
     for (uint32_t i = 0; i < len; i++) {
@@ -234,7 +267,7 @@ static bool subGeneral(uint64_t* dst, const uint64_t* x, const uint64_t* y, uint
 }
 
 // One term of a multiply operation
-NO_SANITIZE("unsigned-integer-overflow")
+SLANG_NO_SANITIZE("unsigned-integer-overflow")
 static uint64_t mulTerm(uint64_t x, uint64_t y, uint64_t& carry) {
 #ifdef _MSC_VER
     uint64_t high;
@@ -261,7 +294,7 @@ static void mulKaratsuba(uint64_t* dst, const uint64_t* x, uint32_t xlen, const
                          uint32_t ylen);
 
 // Generalized multiplier
-NO_SANITIZE("unsigned-integer-overflow")
+SLANG_NO_SANITIZE("unsigned-integer-overflow")
 static void mul(uint64_t* dst, const uint64_t* x, uint32_t xlen, const uint64_t* y, uint32_t ylen) {
     if (xlen > 7 && ylen > 7) {
         mulKaratsuba(dst, x, xlen, y, ylen);
@@ -343,7 +376,7 @@ static void mulKaratsuba(uint64_t* dst, const uint64_t* x, uint32_t xlen, const
     t2.fill(0);
     unevenAdd(t2.get(), yh, yhSize, yl, ylSize);
 
-    ASSERT(std::max(xlSize, xhSize) + 1 + std::max(ylSize, yhSize) + 1 < xlen + ylen);
+    SLANG_ASSERT(std::max(xlSize, xhSize) + 1 + std::max(ylSize, yhSize) + 1 < xlen + ylen);
 
     TempBuffer<uint64_t, 128> t3(xlen + ylen);
     t3.fill(0);
@@ -356,13 +389,13 @@ static void mulKaratsuba(uint64_t* dst, const uint64_t* x, uint32_t xlen, const
 // from "Art of Computer Programming, Volume 2", section 4.3.1, p. 272.
 // Note that this implementation is based on the APInt implementation from
 // the LLVM project.
-NO_SANITIZE("unsigned-integer-overflow")
+SLANG_NO_SANITIZE("unsigned-integer-overflow")
 static void knuthDiv(uint32_t* u, uint32_t* v, uint32_t* q, uint32_t* r, uint32_t m, uint32_t n) {
-    ASSERT(u);
-    ASSERT(v);
-    ASSERT(q);
-    ASSERT(u != v && u != q && v != q);
-    ASSERT(n > 1);
+    SLANG_ASSERT(u);
+    SLANG_ASSERT(v);
+    SLANG_ASSERT(q);
+    SLANG_ASSERT(u != v && u != q && v != q);
+    SLANG_ASSERT(n > 1);
 
     // b denotes the base of the number system. In our case b is 2^32.
     const uint64_t b = 1ULL << 32;
@@ -375,7 +408,7 @@ static void knuthDiv(uint32_t* u, uint32_t* v, uint32_t* q, uint32_t* r, uint32_
     // and v so that its high bits are shifted to the top of v's range without
     // overflow. Note that this can require an extra word in u so that u must
     // be of length m+n+1.
-    uint32_t shift = countLeadingZeros32(v[n - 1]);
+    uint32_t shift = (uint32_t)std::countl_zero(v[n - 1]);
     uint32_t v_carry = 0;
     uint32_t u_carry = 0;
     if (shift) {
@@ -594,7 +627,7 @@ static void clearBits(uint64_t* dest, uint32_t destOffset, uint32_t length) {
 }
 
 template<typename TVal, int ExpBits, int MantissaBits, int Bias>
-static SVInt fromIEEE754(bitwidth_t bits, TVal value, bool isSigned) {
+static SVInt fromIEEE754(bitwidth_t bits, TVal value, bool isSigned, bool round) {
     uint64_t ival = 0;
     memcpy(&ival, &value, sizeof(TVal));
 
@@ -608,7 +641,7 @@ static SVInt fromIEEE754(bitwidth_t bits, TVal value, bool isSigned) {
     // If exponent is negative the value is less than 1. The SystemVerilog
     // spec says that values of exactly 0.5 round away from zero, so check
     // for an exponent of -1 (which sets us at 0.5) and return 1 in that case.
-    if (exp == -1) {
+    if (exp == -1 && round) {
         SVInt result(bits, 1, isSigned);
         return neg ? -result : result;
     }
@@ -626,7 +659,7 @@ static SVInt fromIEEE754(bitwidth_t bits, TVal value, bool isSigned) {
         uint64_t halfway = 1ull << (shift - 1);
         mantissa >>= shift;
 
-        if (remainder >= halfway)
+        if (remainder >= halfway && round)
             mantissa++;
 
         SVInt result(bits, mantissa, isSigned);
@@ -647,12 +680,12 @@ static TVal toIEEE754(SVInt value) {
 
     // Optimize for single word case.
     if (value.isSigned()) {
-        optional<int64_t> i64 = value.as<int64_t>();
+        auto i64 = value.as<int64_t>();
         if (i64)
             return TVal(*i64);
     }
     else {
-        optional<uint64_t> u64 = value.as<uint64_t>();
+        auto u64 = value.as<uint64_t>();
         if (u64)
             return TVal(*u64);
     }
@@ -669,7 +702,7 @@ static TVal toIEEE754(SVInt value) {
     // 2. If they are less than half way then round down.
     // 3. If they are exactly equal to the half way point, round to even.
     bitwidth_t bwidth = value.getActiveBits();
-    ASSERT(bwidth);
+    SLANG_ASSERT(bwidth);
 
     uint64_t mantissa, remainder = 0, halfway = 1;
     uint32_t word = (bwidth - 1) / SVInt::BITS_PER_WORD;
