@@ -2,42 +2,51 @@
 // Lexer.cpp
 // Source file lexer
 //
-// File is under the MIT license; see LICENSE for details
+// SPDX-FileCopyrightText: Michael Popoloski
+// SPDX-License-Identifier: MIT
 //------------------------------------------------------------------------------
 #include "slang/parsing/Lexer.h"
 
-#include <algorithm>
 #include <cmath>
+#include <fmt/core.h>
 
 #include "slang/diagnostics/LexerDiags.h"
 #include "slang/diagnostics/NumericDiags.h"
-#include "slang/syntax/SyntaxNode.h"
+#include "slang/diagnostics/PreprocessorDiags.h"
+#include "slang/syntax/SyntaxKind.h"
 #include "slang/text/CharInfo.h"
 #include "slang/text/SourceManager.h"
 #include "slang/util/BumpAllocator.h"
+#include "slang/util/ScopeGuard.h"
 #include "slang/util/String.h"
 
 static_assert(std::numeric_limits<double>::is_iec559, "SystemVerilog requires IEEE 754");
 
 static const double BitsPerDecimal = log2(10.0);
 
-namespace slang {
+static constexpr std::string_view PragmaBeginProtected = "pragma protect begin_protected"sv;
+static constexpr std::string_view PragmaEndProtected = "pragma protect end_protected"sv;
+
+namespace slang::parsing {
+
+using namespace syntax;
 
 using LF = LexerFacts;
 
 Lexer::Lexer(SourceBuffer buffer, BumpAllocator& alloc, Diagnostics& diagnostics,
              LexerOptions options) :
     Lexer(buffer.id, buffer.data, buffer.data.data(), alloc, diagnostics, options) {
+    library = buffer.library;
 }
 
-Lexer::Lexer(BufferID bufferId, string_view source, const char* startPtr, BumpAllocator& alloc,
+Lexer::Lexer(BufferID bufferId, std::string_view source, const char* startPtr, BumpAllocator& alloc,
              Diagnostics& diagnostics, LexerOptions options) :
-    alloc(alloc),
-    diagnostics(diagnostics), options(options), bufferId(bufferId), originalBegin(source.data()),
-    sourceBuffer(startPtr), sourceEnd(source.data() + source.length()), marker(nullptr) {
+    alloc(alloc), diagnostics(diagnostics), options(options), bufferId(bufferId),
+    originalBegin(source.data()), sourceBuffer(startPtr),
+    sourceEnd(source.data() + source.length()), marker(nullptr) {
     ptrdiff_t count = sourceEnd - sourceBuffer;
-    ASSERT(count);
-    ASSERT(sourceEnd[-1] == '\0');
+    SLANG_ASSERT(count);
+    SLANG_ASSERT(sourceEnd[-1] == '\0');
 
     // detect BOMs so we can give nice errors for invalid encoding
     if (count >= 2) {
@@ -72,12 +81,11 @@ Token Lexer::concatenateTokens(BumpAllocator& alloc, Token left, Token right) {
     leftText.copy(mem, leftText.length());
     rightText.copy(mem + leftText.length(), rightText.length());
     mem[newLength - 1] = '\0';
-    string_view combined{ mem, newLength };
+    std::string_view combined{mem, newLength};
 
     Diagnostics unused;
     Lexer lexer{
-        BufferID::getPlaceholder(), combined, combined.data(), alloc, unused, LexerOptions{}
-    };
+        BufferID::getPlaceholder(), combined, combined.data(), alloc, unused, LexerOptions{}};
 
     auto token = lexer.lex();
     if (token.kind == TokenKind::Unknown || token.rawText().empty())
@@ -91,92 +99,117 @@ Token Lexer::concatenateTokens(BumpAllocator& alloc, Token left, Token right) {
     return token.clone(alloc, trivia, token.rawText(), location);
 }
 
-Token Lexer::stringify(BumpAllocator& alloc, SourceLocation location, span<Trivia const> trivia,
-                       Token* begin, Token* end) {
-    SmallVectorSized<char, 64> text;
-    text.append('"');
+Token Lexer::stringify(Lexer& parentLexer, Token startToken, std::span<Token> bodyTokens,
+                       Token endToken) {
+    SmallVector<char> text;
+    text.push_back('"');
 
-    while (begin != end) {
-        Token cur = *begin;
+    const bool tripleQuoted = startToken.kind == TokenKind::MacroTripleQuote;
+    if (tripleQuoted) {
+        text.push_back('"');
+        text.push_back('"');
+    }
 
+    auto addTrivia = [&](Token cur) {
         for (const Trivia& t : cur.trivia()) {
-            if (t.kind == TriviaKind::Whitespace)
-                text.appendRange(t.getRawText());
+            if (t.kind == TriviaKind::Whitespace ||
+                (tripleQuoted && t.kind == TriviaKind::EndOfLine)) {
+                text.append_range(t.getRawText());
+            }
         }
+    };
+
+    for (auto cur : bodyTokens) {
+        addTrivia(cur);
 
         if (cur.kind == TokenKind::MacroEscapedQuote) {
-            text.append('\\');
-            text.append('"');
+            text.push_back('\\');
+            text.push_back('"');
         }
         else if (cur.kind == TokenKind::StringLiteral) {
-            text.append('\\');
-            text.append('"');
-
             auto raw = cur.rawText();
-            if (raw.size() > 2)
-                text.appendRange(raw.substr(1, raw.size() - 2));
+            const bool nestedHasTriple = raw.starts_with("\"\"\""sv);
+            if (nestedHasTriple) {
+                text.append_range(R"(\"\"\")"sv);
+                raw = raw.substr(3, raw.size() - 6);
+            }
+            else {
+                text.append_range(R"(\")"sv);
+                raw = raw.substr(1, raw.size() - 2);
+            }
 
-            text.append('\\');
-            text.append('"');
+            text.append_range(raw);
+
+            if (nestedHasTriple)
+                text.append_range(R"(\"\"\")"sv);
+            else
+                text.append_range(R"(\")"sv);
         }
         else if (cur.kind != TokenKind::EmptyMacroArgument) {
-            text.appendRange(cur.rawText());
+            text.append_range(cur.rawText());
         }
-        begin++;
     }
-    text.append('"');
-    text.append('\0');
 
-    string_view raw = toStringView(text.copy(alloc));
+    if (endToken)
+        addTrivia(endToken);
+
+    if (tripleQuoted) {
+        text.push_back('"');
+        text.push_back('"');
+    }
+
+    text.push_back('"');
+    text.push_back('\0');
+
+    std::string_view raw = toStringView(text.copy(parentLexer.alloc));
 
     Diagnostics unused;
-    Lexer lexer{ BufferID::getPlaceholder(), raw, raw.data(), alloc, unused, LexerOptions{} };
+    Lexer lexer{BufferID::getPlaceholder(), raw,    raw.data(),
+                parentLexer.alloc,          unused, parentLexer.options};
 
     auto token = lexer.lex();
-    ASSERT(token.kind == TokenKind::StringLiteral);
-    ASSERT(lexer.lex().kind == TokenKind::EndOfFile);
+    SLANG_ASSERT(token.kind == TokenKind::StringLiteral);
+    SLANG_ASSERT(lexer.lex().kind == TokenKind::EndOfFile);
 
-    return token.clone(alloc, trivia, raw.substr(0, raw.length() - 1), location);
+    return token.clone(parentLexer.alloc, startToken.trivia(), raw.substr(0, raw.length() - 1),
+                       startToken.location());
 }
 
-Trivia Lexer::commentify(BumpAllocator& alloc, Token* begin, Token* end) {
-    SmallVectorSized<char, 64> text;
-    while (begin != end) {
-        Token cur = *begin;
+Trivia Lexer::commentify(BumpAllocator& alloc, std::span<Token> tokens) {
+    SmallVector<char> text;
+    for (auto cur : tokens) {
         for (const Trivia& t : cur.trivia())
-            text.appendRange(t.getRawText());
+            text.append_range(t.getRawText());
 
         if (cur.kind != TokenKind::EmptyMacroArgument)
-            text.appendRange(cur.rawText());
-
-        begin++;
+            text.append_range(cur.rawText());
     }
-    text.append('\0');
+    text.push_back('\0');
 
-    string_view raw = toStringView(text.copy(alloc));
+    std::string_view raw = toStringView(text.copy(alloc));
 
     Diagnostics unused;
-    Lexer lexer{ BufferID::getPlaceholder(), raw, raw.data(), alloc, unused, LexerOptions{} };
+    Lexer lexer{BufferID::getPlaceholder(), raw, raw.data(), alloc, unused, LexerOptions{}};
 
     auto token = lexer.lex();
-    ASSERT(token.kind == TokenKind::EndOfFile);
-    ASSERT(token.trivia().size() == 1);
+    SLANG_ASSERT(token.kind == TokenKind::EndOfFile);
+    SLANG_ASSERT(token.trivia().size() == 1);
 
     return token.trivia()[0];
 }
 
 void Lexer::splitTokens(BumpAllocator& alloc, Diagnostics& diagnostics,
                         const SourceManager& sourceManager, Token sourceToken, size_t offset,
-                        KeywordVersion keywordVersion, SmallVector<Token>& results) {
+                        KeywordVersion keywordVersion, SmallVectorBase<Token>& results) {
     auto loc = sourceToken.location();
     if (sourceManager.isMacroLoc(loc))
         loc = sourceManager.getOriginalLoc(loc);
 
     auto sourceText = sourceManager.getSourceText(loc.buffer());
-    ASSERT(!sourceText.empty());
+    SLANG_ASSERT(!sourceText.empty());
 
-    Lexer lexer{ loc.buffer(), sourceText,  sourceToken.rawText().substr(offset).data(),
-                 alloc,        diagnostics, LexerOptions{} };
+    Lexer lexer{loc.buffer(), sourceText,  sourceToken.rawText().substr(offset).data(),
+                alloc,        diagnostics, LexerOptions{}};
 
     size_t endOffset = loc.offset() + sourceToken.rawText().length();
     while (true) {
@@ -185,13 +218,17 @@ void Lexer::splitTokens(BumpAllocator& alloc, Diagnostics& diagnostics,
             token.location().offset() >= endOffset)
             break;
 
-        results.append(token);
+        results.push_back(token);
     }
 }
 
+Token Lexer::lex() {
+    return lex(LF::getDefaultKeywordVersion(options.languageVersion));
+}
+
 Token Lexer::lex(KeywordVersion keywordVersion) {
     triviaBuffer.clear();
-    lexTrivia();
+    lexTrivia<false>();
 
     // lex the next token
     mark();
@@ -202,7 +239,7 @@ Token Lexer::lex(KeywordVersion keywordVersion) {
         addDiag(diag::TooManyLexerErrors, currentOffset());
         sourceBuffer = sourceEnd - 1;
 
-        triviaBuffer.append(Trivia(TriviaKind::DisabledText, lexeme()));
+        triviaBuffer.push_back(Trivia(TriviaKind::DisabledText, lexeme()));
         return Token(alloc, TokenKind::EndOfFile, triviaBuffer.copy(alloc), token.rawText(),
                      token.location());
     }
@@ -210,6 +247,48 @@ Token Lexer::lex(KeywordVersion keywordVersion) {
     return token;
 }
 
+bool Lexer::isNextTokenOnSameLine() {
+    auto guard = ScopeGuard([this, currBuff = sourceBuffer] { sourceBuffer = currBuff; });
+
+    while (true) {
+        switch (peek()) {
+            case ' ':
+            case '\t':
+            case '\v':
+            case '\f':
+                advance();
+                break;
+            case '/':
+                switch (peek(1)) {
+                    case '/':
+                        return false;
+                    case '*':
+                        advance(2);
+                        scanBlockComment();
+                        break;
+                    default:
+                        return true;
+                }
+                break;
+            case '\0':
+            case '\r':
+            case '\n':
+                return false;
+            default:
+                return true;
+        }
+    }
+}
+
+Token Lexer::lexEncodedText(ProtectEncoding encoding, uint32_t expectedBytes, bool singleLine,
+                            bool legacyProtectedMode) {
+    triviaBuffer.clear();
+    lexTrivia<true>();
+    mark();
+    scanEncodedText(encoding, expectedBytes, singleLine, legacyProtectedMode);
+    return create(TokenKind::Unknown);
+}
+
 Token Lexer::lexToken(KeywordVersion keywordVersion) {
     char c = peek();
     advance();
@@ -290,10 +369,7 @@ Token Lexer::lexToken(KeywordVersion keywordVersion) {
             else
                 return lexApostrophe();
         case '(':
-            if (!consume('*'))
-                return create(TokenKind::OpenParenthesis);
-            else
-                return create(TokenKind::OpenParenthesisStar);
+            return create(TokenKind::OpenParenthesis);
         case ')':
             return create(TokenKind::CloseParenthesis);
         case '*':
@@ -307,9 +383,6 @@ Token Lexer::lexToken(KeywordVersion keywordVersion) {
                 case '>':
                     advance();
                     return create(TokenKind::StarArrow);
-                case ')':
-                    advance();
-                    return create(TokenKind::StarCloseParenthesis);
             }
             return create(TokenKind::Star);
         case '+':
@@ -323,6 +396,18 @@ Token Lexer::lexToken(KeywordVersion keywordVersion) {
                 case ':':
                     advance();
                     return create(TokenKind::PlusColon);
+                case '/':
+                    if (peek(1) == '-') {
+                        advance(2);
+                        return create(TokenKind::PlusDivMinus);
+                    }
+                    break;
+                case '%':
+                    if (peek(1) == '-') {
+                        advance(2);
+                        return create(TokenKind::PlusModMinus);
+                    }
+                    break;
             }
             return create(TokenKind::Plus);
         case ',':
@@ -347,10 +432,7 @@ Token Lexer::lexToken(KeywordVersion keywordVersion) {
             }
             return create(TokenKind::Minus);
         case '.':
-            if (consume('*'))
-                return create(TokenKind::DotStar);
-            else
-                return create(TokenKind::Dot);
+            return create(TokenKind::Dot);
         case '/':
             if (consume('='))
                 return create(TokenKind::SlashEqual);
@@ -491,9 +573,10 @@ Token Lexer::lexToken(KeywordVersion keywordVersion) {
             scanIdentifier();
 
             // might be a keyword
-            TokenKind kind;
-            if (LF::getKeywordTable(keywordVersion)->lookup(lexeme(), kind))
-                return create(kind);
+            auto table = LF::getKeywordTable(keywordVersion);
+            SLANG_ASSERT(table);
+            if (auto it = table->find(lexeme()); it != table->end())
+                return create(it->second);
 
             return create(TokenKind::Identifier);
         }
@@ -517,7 +600,13 @@ Token Lexer::lexToken(KeywordVersion keywordVersion) {
             switch (peek()) {
                 case '"':
                     advance();
-                    return create(TokenKind::MacroQuote);
+                    if (peek() == '"' && peek(1) == '"') {
+                        advance(2);
+                        return create(TokenKind::MacroTripleQuote);
+                    }
+                    else {
+                        return create(TokenKind::MacroQuote);
+                    }
                 case '`':
                     advance();
                     return create(TokenKind::MacroPaste);
@@ -579,8 +668,7 @@ Token Lexer::lexToken(KeywordVersion keywordVersion) {
 
                 bool sawUTF8Error = false;
                 do {
-                    uint32_t unused;
-                    sawUTF8Error |= !scanUTF8Char(sawUTF8Error, &unused);
+                    sawUTF8Error |= !scanUTF8Char(sawUTF8Error);
                 } while (!isASCII(peek()));
             }
             return create(TokenKind::Unknown);
@@ -588,8 +676,19 @@ Token Lexer::lexToken(KeywordVersion keywordVersion) {
 }
 
 Token Lexer::lexStringLiteral() {
+    bool tripleQuoted = false;
+    if (peek() == '"' && peek(1) == '"') {
+        // New in v1800-2023: triple quoted string literals
+        if (options.languageVersion < LanguageVersion::v1800_2023) {
+            addDiag(diag::WrongLanguageVersion, currentOffset() - 1)
+                << toString(options.languageVersion);
+        }
+        tripleQuoted = true;
+        advance(2);
+    }
+
+    stringBuffer.clear();
     bool sawUTF8Error = false;
-    SmallVectorSized<char, 128> stringBuffer;
     while (true) {
         size_t offset = currentOffset();
         char c = peek();
@@ -602,13 +701,13 @@ Token Lexer::lexStringLiteral() {
             uint32_t charCode;
             switch (c) {
                     // clang-format off
-                case 'n': stringBuffer.append('\n'); break;
-                case 't': stringBuffer.append('\t'); break;
-                case '\\': stringBuffer.append('\\'); break;
-                case '"': stringBuffer.append('"'); break;
-                case 'v': stringBuffer.append('\v'); break;
-                case 'f': stringBuffer.append('\f'); break;
-                case 'a': stringBuffer.append('\a'); break;
+                case 'n': stringBuffer.push_back('\n'); break;
+                case 't': stringBuffer.push_back('\t'); break;
+                case '\\': stringBuffer.push_back('\\'); break;
+                case '"': stringBuffer.push_back('"'); break;
+                case 'v': stringBuffer.push_back('\v'); break;
+                case 'f': stringBuffer.push_back('\f'); break;
+                case 'a': stringBuffer.push_back('\a'); break;
                 case '\n': break;
                 case '\r': consume('\n'); break;
                 case '0': case '1': case '2': case '3':
@@ -628,37 +727,37 @@ Token Lexer::lexStringLiteral() {
                             }
                         }
                     }
-                    stringBuffer.append((char)charCode);
+                    stringBuffer.push_back((char)charCode);
                     break;
                 case 'x':
                     c = peek();
-                    advance();
                     if (!isHexDigit(c)) {
                         addDiag(diag::InvalidHexEscapeCode, offset);
-                        stringBuffer.append(c);
+                        stringBuffer.push_back('x');
                     }
                     else {
+                        advance();
                         charCode = getHexDigitValue(c);
                         if (isHexDigit(c = peek())) {
                             advance();
                             charCode = (charCode * 16) + getHexDigitValue(c);
                         }
-                        stringBuffer.append((char)charCode);
+                        stringBuffer.push_back((char)charCode);
                     }
                     break;
                 default: {
                     auto curr = --sourceBuffer;
 
                     uint32_t unicodeChar;
-                    if (scanUTF8Char(sawUTF8Error, &unicodeChar)) {
+                    int unicodeLen;
+                    if (scanUTF8Char(sawUTF8Error, &unicodeChar, unicodeLen)) {
                         if (isPrintableUnicode(unicodeChar)) {
                             // '\%' is not an actual escape code but other tools silently allow it
                             // and major UVM headers use it, so we'll issue a (fairly quiet) warning
                             // about it. Otherwise issue a louder warning (on by default).
-                            DiagCode code =
-                                c == '%' ? diag::NonstandardEscapeCode : diag::UnknownEscapeCode;
-                            addDiag(code, offset)
-                                << string_view(curr, (size_t)utf8Len((unsigned char)c));
+                            DiagCode code = c == '%' ? diag::NonstandardEscapeCode
+                                                     : diag::UnknownEscapeCode;
+                            addDiag(code, offset) << std::string_view(curr, (size_t)unicodeLen);
                         }
                     }
                     else {
@@ -673,10 +772,23 @@ Token Lexer::lexStringLiteral() {
             }
         }
         else if (c == '"') {
-            advance();
-            break;
+            if (tripleQuoted) {
+                if (peek(1) == '"' && peek(2) == '"') {
+                    advance(3);
+                    break;
+                }
+                else {
+                    advance();
+                    stringBuffer.push_back(c);
+                    sawUTF8Error = false;
+                }
+            }
+            else {
+                advance();
+                break;
+            }
         }
-        else if (isNewline(c)) {
+        else if (isNewline(c) && !tripleQuoted) {
             addDiag(diag::ExpectedClosingQuote, offset);
             break;
         }
@@ -693,25 +805,22 @@ Token Lexer::lexStringLiteral() {
         }
         else if (isASCII(c)) {
             advance();
-            stringBuffer.append(c);
+            stringBuffer.push_back(c);
             sawUTF8Error = false;
         }
         else {
             auto curr = sourceBuffer;
 
             uint32_t unused;
-            sawUTF8Error |= !scanUTF8Char(sawUTF8Error, &unused);
+            int unicodeLen;
+            sawUTF8Error |= !scanUTF8Char(sawUTF8Error, &unused, unicodeLen);
 
             // Regardless of whether the character sequence was valid or not
             // we want to add the bytes to the string, to allow for cases where
             // the source is actually something like latin-1 encoded. Ignoring the
             // warning and carrying on will do the right thing for them.
-            int len = utf8Len((unsigned char)c);
-            if (len == 0)
-                len = 1;
-
-            for (int i = 0; i < len; i++)
-                stringBuffer.append(curr[i]);
+            for (int i = 0; i < unicodeLen; i++)
+                stringBuffer.push_back(curr[i]);
         }
     }
 
@@ -729,7 +838,7 @@ Token Lexer::lexEscapeSequence(bool isMacroName) {
             return create(TokenKind::LineContinuation);
         }
 
-        addDiag(diag::EscapedWhitespace, currentOffset());
+        // Error issued in the Preprocessor
         return create(TokenKind::Unknown);
     }
 
@@ -769,17 +878,15 @@ Token Lexer::lexDirective() {
         return lexEscapeSequence(true);
     }
 
-    // store the offset before scanning in order to easily report error locations
-    size_t startingOffset = currentOffset();
     scanIdentifier();
 
     // if length is 1, we just have a grave character on its own, which is an error
     if (lexemeLength() == 1) {
-        addDiag(diag::MisplacedDirectiveChar, startingOffset);
+        // Error issued in the Preprocessor
         return create(TokenKind::Unknown);
     }
 
-    SyntaxKind directive = LF::getDirectiveKind(lexeme().substr(1));
+    SyntaxKind directive = LF::getDirectiveKind(lexeme().substr(1), options.enableLegacyProtect);
     return create(TokenKind::Directive, directive);
 }
 
@@ -805,8 +912,8 @@ Token Lexer::lexNumericLiteral() {
     // instead. Since we expect many more ints than floats, it makes sense to
     // not waste time populating that array up front.
     size_t startOfNum = currentOffset();
-    SmallVectorSized<logic_t, 32> digits;
-    SmallVectorSized<char, 32> floatChars;
+    SmallVector<logic_t> digits;
+    SmallVector<char> floatChars;
 
     while (true) {
         char c = peek();
@@ -815,17 +922,17 @@ Token Lexer::lexNumericLiteral() {
         else if (!isDecimalDigit(c))
             break;
         else {
-            digits.append(logic_t(getDigitValue(c)));
+            digits.push_back(logic_t(getDigitValue(c)));
             advance();
         }
     }
 
     auto populateChars = [&]() {
         if (digits.empty())
-            floatChars.append('0');
+            floatChars.push_back('0');
         else {
             for (auto d : digits)
-                floatChars.append(char((char)d.value + '0'));
+                floatChars.push_back(char((char)d.value + '0'));
         }
     };
 
@@ -833,10 +940,7 @@ Token Lexer::lexNumericLiteral() {
     if (peek() == '.') {
         advance();
         populateChars();
-        floatChars.append('.');
-
-        if (peek() == '_')
-            addDiag(diag::DigitsLeadingUnderscore, currentOffset());
+        floatChars.push_back('.');
 
         bool any = false;
         while (true) {
@@ -847,15 +951,13 @@ Token Lexer::lexNumericLiteral() {
                 break;
             else {
                 any = true;
-                floatChars.append(c);
+                floatChars.push_back(c);
                 advance();
             }
         }
 
-        if (!any) {
-            addDiag(diag::MissingFractionalDigits, currentOffset());
-            floatChars.append('0');
-        }
+        if (!any)
+            floatChars.push_back('0');
     }
 
     // Check for an exponent. Note that this case can be indistinguishable from
@@ -867,26 +969,23 @@ Token Lexer::lexNumericLiteral() {
     // If `FOO is defined to be 'h this represents an expression: 62 + 2
     // Otherwise, this represents a real literal: 300.0
 
-    optional<TimeUnit> timeSuffix;
+    std::optional<TimeUnit> timeSuffix;
     char c = peek();
     if (c == 'e' || c == 'E') {
         bool hasDecimal = !floatChars.empty();
         if (!hasDecimal)
             populateChars();
 
-        floatChars.append('e');
+        floatChars.push_back('e');
 
         // skip over leading sign
         int index = 1;
         c = peek(index);
         if (c == '+' || c == '-') {
-            floatChars.append(c);
+            floatChars.push_back(c);
             c = peek(++index);
         }
 
-        if (c == '_' && hasDecimal)
-            addDiag(diag::DigitsLeadingUnderscore, currentOffset());
-
         bool any = false;
         while (true) {
             if (c == '_')
@@ -895,17 +994,15 @@ Token Lexer::lexNumericLiteral() {
                 break;
             else {
                 any = true;
-                floatChars.append(c);
+                floatChars.push_back(c);
                 c = peek(++index);
             }
         }
 
         if (any || hasDecimal) {
             advance(index);
-            if (!any) {
-                addDiag(diag::MissingExponentDigits, currentOffset());
-                floatChars.append('1');
-            }
+            if (!any)
+                floatChars.push_back('1');
         }
         else {
             // This isn't a float, it's probably a hex literal. Back up (by not calling advance)
@@ -925,12 +1022,12 @@ Token Lexer::lexNumericLiteral() {
         // We have a floating point result. Let the standard library do the heavy lifting of
         // converting and rounding correctly. Note that we depend on this code returning
         // 0 for underflow and +inf for overflow.
-        floatChars.append('\0');
+        floatChars.push_back('\0');
 
         char* end;
         errno = 0;
         double value = strtod(floatChars.data(), &end);
-        ASSERT(end == floatChars.end() - 1); // should never error
+        SLANG_ASSERT(end == floatChars.end() - 1); // should never error
 
         // If we had an overflow or underflow, errno is now ERANGE. We can't warn here in case
         // this turns out to actually be a hex literal. Have the token carry this info so someone
@@ -980,13 +1077,10 @@ Token Lexer::lexApostrophe() {
         case 'S': {
             advance();
             LiteralBase base;
-            if (!literalBaseFromChar(peek(), base)) {
-                addDiag(diag::ExpectedIntegerBaseAfterSigned, currentOffset());
+            if (!literalBaseFromChar(peek(), base))
                 base = LiteralBase::Decimal;
-            }
-            else {
+            else
                 advance();
-            }
             return create(TokenKind::IntegerBase, base, true);
         }
         default: {
@@ -1002,7 +1096,7 @@ Token Lexer::lexApostrophe() {
     }
 }
 
-optional<TimeUnit> Lexer::lexTimeLiteral() {
+std::optional<TimeUnit> Lexer::lexTimeLiteral() {
 #define CASE(c, flag)              \
     case c:                        \
         if (peek(1) == 's') {      \
@@ -1029,6 +1123,7 @@ optional<TimeUnit> Lexer::lexTimeLiteral() {
     return std::nullopt;
 }
 
+template<bool StopAfterNewline>
 void Lexer::lexTrivia() {
     while (true) {
         mark();
@@ -1060,10 +1155,14 @@ void Lexer::lexTrivia() {
                 advance();
                 consume('\n');
                 addTrivia(TriviaKind::EndOfLine);
+                if (StopAfterNewline)
+                    return;
                 break;
             case '\n':
                 advance();
                 addTrivia(TriviaKind::EndOfLine);
+                if (StopAfterNewline)
+                    return;
                 break;
             default:
                 return;
@@ -1100,6 +1199,27 @@ void Lexer::scanWhitespace() {
 }
 
 void Lexer::scanLineComment() {
+    if (options.enableLegacyProtect) {
+        // See if we're looking at a pragma protect comment and skip
+        // over it if so.
+        while (peek() == ' ')
+            advance();
+
+        bool found = true;
+        for (char c : PragmaBeginProtected) {
+            if (!consume(c)) {
+                found = false;
+                break;
+            }
+        }
+
+        if (found) {
+            scanProtectComment();
+            addTrivia(TriviaKind::DisabledText);
+            return;
+        }
+    }
+
     bool sawUTF8Error = false;
     while (true) {
         char c = peek();
@@ -1119,8 +1239,7 @@ void Lexer::scanLineComment() {
             advance();
         }
         else {
-            uint32_t unused;
-            sawUTF8Error |= !scanUTF8Char(sawUTF8Error, &unused);
+            sawUTF8Error |= !scanUTF8Char(sawUTF8Error);
         }
     }
     addTrivia(TriviaKind::LineComment);
@@ -1157,30 +1276,44 @@ void Lexer::scanBlockComment() {
             }
         }
         else {
-            uint32_t unused;
-            sawUTF8Error |= !scanUTF8Char(sawUTF8Error, &unused);
+            sawUTF8Error |= !scanUTF8Char(sawUTF8Error);
         }
     }
 
     addTrivia(TriviaKind::BlockComment);
 }
 
-bool Lexer::scanUTF8Char(bool alreadyErrored, uint32_t* code) {
+bool Lexer::scanUTF8Char(bool alreadyErrored) {
+    uint32_t unused1;
+    int unused2;
+    return scanUTF8Char(alreadyErrored, &unused1, unused2);
+}
+
+bool Lexer::scanUTF8Char(bool alreadyErrored, uint32_t* code, int& computedLen) {
     int error;
     auto curr = sourceBuffer;
     if (sourceBuffer + 4 < sourceEnd) {
-        sourceBuffer = utf8Decode(sourceBuffer, code, &error);
+        sourceBuffer = utf8Decode(sourceBuffer, code, &error, computedLen);
     }
     else {
         char buf[4] = {};
-        memcpy(buf, sourceBuffer, size_t(sourceEnd - sourceBuffer - 1));
+        auto spaceLeft = sourceEnd - sourceBuffer - 1;
+        memcpy(buf, sourceBuffer, size_t(spaceLeft));
 
-        auto next = utf8Decode(buf, code, &error);
-        sourceBuffer += next - buf;
+        auto next = utf8Decode(buf, code, &error, computedLen);
+        sourceBuffer += std::min(next - buf, spaceLeft);
+        computedLen = std::min(computedLen, (int)spaceLeft);
     }
 
     if (error) {
-        errorCount++;
+        // if error, trim next pointer so that control char is read as next char
+        if ((computedLen > 1) && (curr[1] < 0x20))
+            sourceBuffer = curr + 1;
+        else if ((computedLen > 2) && (curr[2] < 0x20))
+            sourceBuffer = curr + 2;
+        else if ((computedLen > 3) && (curr[3] < 0x20))
+            sourceBuffer = curr + 3;
+
         if (!alreadyErrored)
             addDiag(diag::InvalidUTF8Seq, (size_t)(curr - originalBegin));
         return false;
@@ -1189,6 +1322,204 @@ bool Lexer::scanUTF8Char(bool alreadyErrored, uint32_t* code) {
     return true;
 }
 
+void Lexer::scanEncodedText(ProtectEncoding encoding, uint32_t expectedBytes, bool singleLine,
+                            bool legacyProtectedMode) {
+    // Helper function that returns true if the current position in the buffer
+    // is looking at the string "pragma".
+    auto lookingAtPragma = [&] {
+        int index = 0;
+        auto endStr = legacyProtectedMode ? "endprotected"sv : "pragma"sv;
+        for (char c : endStr) {
+            if (peek(++index) != c)
+                return false;
+        }
+        return true;
+    };
+
+    auto invalidByte = [&](char invalidChar, std::string_view name) {
+        auto& diag = addDiag(diag::InvalidEncodingByte, currentOffset()) << name;
+        diag << (isPrintableASCII(invalidChar) ? std::string(1, invalidChar)
+                                               : fmt::format("{:#04x}", invalidChar));
+
+        // Try to skip ahead to the next `pragma directive to get us out of this region.
+        while (true) {
+            char c = peek();
+            if (c == '\0' && reallyAtEnd())
+                break;
+
+            if (c == '`' && lookingAtPragma())
+                break;
+
+            if (singleLine && (c == '\r' || c == '\n'))
+                break;
+
+            advance();
+        }
+    };
+
+    uint32_t byteCount = 0;
+    while (true) {
+        if (expectedBytes && byteCount >= expectedBytes && !singleLine) {
+            if (byteCount != expectedBytes) {
+                addDiag(diag::ProtectEncodingBytes, currentOffset() - 1)
+                    << byteCount << expectedBytes;
+            }
+            return;
+        }
+
+        char c = peek();
+        if (c == '\r' || c == '\n') {
+            // If this is a single line region then we're done here.
+            if (singleLine) {
+                if (expectedBytes && byteCount != expectedBytes) {
+                    addDiag(diag::ProtectEncodingBytes, currentOffset() - 1)
+                        << byteCount << expectedBytes;
+                }
+                return;
+            }
+
+            // Otherwise continue on. This newline doesn't count toward our expected byte limit,
+            // unless this is the quoted printable encoding.
+            advance();
+            if (c == '\r' && peek() == '\n')
+                advance();
+
+            if (encoding == ProtectEncoding::QuotedPrintable) {
+                byteCount++;
+                if (c == '\r' && peek(-1) == '\n')
+                    byteCount++;
+            }
+
+            continue;
+        }
+
+        if (!expectedBytes && !singleLine && c == '`') {
+            // Encoding tools probably shouldn't do this but if they do we
+            // should try to gracefully guess the end of the region by looking
+            // for another non-encoded pragma that ends it.
+            if (lookingAtPragma())
+                return;
+        }
+
+        switch (encoding) {
+            case ProtectEncoding::UUEncode: {
+                // uuencode tells us the length of the line up front, so use that
+                // to read the whole line in one go. The encoding is invalid if that
+                // doesn't match up with what we find in the data.
+                if (c == '`') {
+                    // The grave character is a special case meaning 0 characters on this line.
+                    advance();
+                    continue;
+                }
+
+                if (c < 0x20 || c > 0x20 + 45) {
+                    invalidByte(c, "uuencode"sv);
+                    return;
+                }
+
+                uint32_t count = uint32_t(c - 0x20);
+                byteCount += count;
+                advance();
+
+                uint32_t encodedCount = (uint32_t)ceil(count * 4 / 3.0);
+                for (uint32_t i = 0; i < encodedCount; i++) {
+                    c = peek();
+                    if (c < 0x20 || c > 0x60) {
+                        invalidByte(c, "uuencode"sv);
+                        return;
+                    }
+                    advance();
+                }
+                break;
+            }
+            case ProtectEncoding::Base64:
+                byteCount += 3;
+                for (int i = 0; i < 4; i++) {
+                    c = peek();
+                    if (i > 1 && c == '=') {
+                        byteCount--;
+                    }
+                    else if (!isBase64Char(c)) {
+                        invalidByte(c, "base64"sv);
+                        return;
+                    }
+
+                    advance();
+                }
+                break;
+            case ProtectEncoding::QuotedPrintable:
+                if (!isPrintableASCII(c) && c != '\t') {
+                    invalidByte(c, "quoted-printable"sv);
+                    return;
+                }
+
+                advance();
+                if (c == '=') {
+                    // If this is a soft line break then it doesn't count
+                    // towards our byte count. Otherwise this is an escaped
+                    // character that does count.
+                    c = peek();
+                    if (c == '\r' || c == '\n') {
+                        advance();
+                        if (c == '\r' && peek() == '\n')
+                            advance();
+                        continue;
+                    }
+
+                    if (!isHexDigit(c) || !isHexDigit(peek(1))) {
+                        invalidByte(c, "quoted-printable"sv);
+                        return;
+                    }
+
+                    advance(2);
+                }
+                byteCount++;
+                break;
+            case ProtectEncoding::Raw:
+                if (c == '\0' && reallyAtEnd()) {
+                    addDiag(diag::RawProtectEOF, currentOffset() - 1);
+                    return;
+                }
+
+                advance();
+                byteCount++;
+                break;
+            default:
+                SLANG_UNREACHABLE;
+        }
+    }
+}
+
+void Lexer::scanProtectComment() {
+    addDiag(diag::ProtectedEnvelope, currentOffset() - PragmaBeginProtected.size());
+
+    while (true) {
+        char c = peek();
+        if (c == '\0' && reallyAtEnd()) {
+            addDiag(diag::RawProtectEOF, currentOffset() - 1);
+            return;
+        }
+
+        advance();
+        if (c == '/' && peek() == '/') {
+            advance();
+            while (peek() == ' ')
+                advance();
+
+            bool found = true;
+            for (char d : PragmaEndProtected) {
+                if (!consume(d)) {
+                    found = false;
+                    break;
+                }
+            }
+
+            if (found)
+                return;
+        }
+    }
+}
+
 template<typename... Args>
 Token Lexer::create(TokenKind kind, Args&&... args) {
     SourceLocation location(bufferId, size_t(marker - originalBegin));
@@ -1197,15 +1528,15 @@ Token Lexer::create(TokenKind kind, Args&&... args) {
 }
 
 void Lexer::addTrivia(TriviaKind kind) {
-    triviaBuffer.emplace(kind, lexeme());
+    triviaBuffer.emplace_back(kind, lexeme());
 }
 
 Diagnostic& Lexer::addDiag(DiagCode code, size_t offset) {
     return diagnostics.add(code, SourceLocation(bufferId, offset));
 }
 
-size_t Lexer::currentOffset() {
+size_t Lexer::currentOffset() const {
     return (size_t)(sourceBuffer - originalBegin);
 }
 
-} // namespace slang
+} // namespace slang::parsing
