@@ -2,44 +2,65 @@
 // LexerFacts.cpp
 // Random lexer-related utility functions
 //
-// File is under the MIT license; see LICENSE for details
+// SPDX-FileCopyrightText: Michael Popoloski
+// SPDX-License-Identifier: MIT
 //------------------------------------------------------------------------------
 #include "slang/parsing/LexerFacts.h"
 
-#include "slang/syntax/SyntaxNode.h"
+#include "slang/parsing/TokenKind.h"
+#include "slang/syntax/SyntaxKind.h"
 
-namespace slang {
+namespace slang::parsing {
+
+using namespace syntax;
 
 // clang-format off
-const static StringTable<TokenKind> systemIdentifierKeywords = {
+const static flat_hash_map<std::string_view, TokenKind> systemIdentifierKeywords = {
     { "$root", TokenKind::RootSystemName },
     { "$unit", TokenKind::UnitSystemName }
 };
 
-const static StringTable<SyntaxKind> directiveTable = {
-    { "begin_keywords", SyntaxKind::BeginKeywordsDirective },
-    { "celldefine", SyntaxKind::CellDefineDirective },
-    { "default_nettype", SyntaxKind::DefaultNetTypeDirective },
-    { "define", SyntaxKind::DefineDirective },
-    { "else", SyntaxKind::ElseDirective },
-    { "elsif", SyntaxKind::ElsIfDirective },
-    { "end_keywords", SyntaxKind::EndKeywordsDirective },
-    { "endcelldefine", SyntaxKind::EndCellDefineDirective },
-    { "endif", SyntaxKind::EndIfDirective },
-    { "ifdef", SyntaxKind::IfDefDirective },
-    { "ifndef", SyntaxKind::IfNDefDirective },
-    { "include", SyntaxKind::IncludeDirective },
-    { "line", SyntaxKind::LineDirective },
-    { "nounconnected_drive", SyntaxKind::NoUnconnectedDriveDirective },
-    { "pragma", SyntaxKind::PragmaDirective },
-    { "resetall", SyntaxKind::ResetAllDirective },
-    { "timescale", SyntaxKind::TimeScaleDirective },
-    { "unconnected_drive", SyntaxKind::UnconnectedDriveDirective },
-    { "undef", SyntaxKind::UndefDirective },
-    { "undefineall", SyntaxKind::UndefineAllDirective }
+#define STANDARD_DIRECTIVES \
+    { "begin_keywords", SyntaxKind::BeginKeywordsDirective },                  \
+    { "celldefine", SyntaxKind::CellDefineDirective },                         \
+    { "default_nettype", SyntaxKind::DefaultNetTypeDirective },                \
+    { "define", SyntaxKind::DefineDirective },                                 \
+    { "else", SyntaxKind::ElseDirective },                                     \
+    { "elsif", SyntaxKind::ElsIfDirective },                                   \
+    { "end_keywords", SyntaxKind::EndKeywordsDirective },                      \
+    { "endcelldefine", SyntaxKind::EndCellDefineDirective },                   \
+    { "endif", SyntaxKind::EndIfDirective },                                   \
+    { "ifdef", SyntaxKind::IfDefDirective },                                   \
+    { "ifndef", SyntaxKind::IfNDefDirective },                                 \
+    { "include", SyntaxKind::IncludeDirective },                               \
+    { "line", SyntaxKind::LineDirective },                                     \
+    { "nounconnected_drive", SyntaxKind::NoUnconnectedDriveDirective },        \
+    { "pragma", SyntaxKind::PragmaDirective },                                 \
+    { "resetall", SyntaxKind::ResetAllDirective },                             \
+    { "timescale", SyntaxKind::TimeScaleDirective },                           \
+    { "unconnected_drive", SyntaxKind::UnconnectedDriveDirective },            \
+    { "undef", SyntaxKind::UndefDirective },                                   \
+    { "undefineall", SyntaxKind::UndefineAllDirective },                       \
+    { "default_decay_time", SyntaxKind::DefaultDecayTimeDirective },           \
+    { "default_trireg_strength", SyntaxKind::DefaultTriregStrengthDirective }, \
+    { "delay_mode_distributed", SyntaxKind::DelayModeDistributedDirective },   \
+    { "delay_mode_path", SyntaxKind::DelayModePathDirective },                 \
+    { "delay_mode_unit", SyntaxKind::DelayModeUnitDirective },                 \
+    { "delay_mode_zero", SyntaxKind::DelayModeZeroDirective },
+
+const static flat_hash_map<std::string_view, SyntaxKind> directiveTable = {
+    STANDARD_DIRECTIVES
 };
 
-const static StringTable<KeywordVersion> keywordVersionTable = {
+const static flat_hash_map<std::string_view, SyntaxKind> directivesWithLegacyProtect = {
+    STANDARD_DIRECTIVES
+    { "protect", SyntaxKind::ProtectDirective },
+    { "endprotect", SyntaxKind::EndProtectDirective },
+    { "protected", SyntaxKind::ProtectedDirective },
+    { "endprotected", SyntaxKind::EndProtectedDirective }
+};
+
+const static flat_hash_map<std::string_view, KeywordVersion> keywordVersionTable = {
     { "1364-1995", KeywordVersion::v1364_1995 },
     { "1364-2001-noconfig", KeywordVersion::v1364_2001_noconfig },
     { "1364-2001", KeywordVersion::v1364_2001 },
@@ -47,7 +68,8 @@ const static StringTable<KeywordVersion> keywordVersionTable = {
     { "1800-2005", KeywordVersion::v1800_2005 },
     { "1800-2009", KeywordVersion::v1800_2009 },
     { "1800-2012", KeywordVersion::v1800_2012 },
-    { "1800-2017", KeywordVersion::v1800_2017 }
+    { "1800-2017", KeywordVersion::v1800_2017 },
+    { "1800-2023", KeywordVersion::v1800_2023 }
 };
 
 // Lists of keywords, separated by the specification in which they were first introduced
@@ -316,7 +338,7 @@ const static StringTable<KeywordVersion> keywordVersionTable = {
 
 // We maintain a separate table of keywords for all the various specifications,
 // to allow for easy switching between them when requested
-const static StringTable<TokenKind> allKeywords[8] =
+const static flat_hash_map<std::string_view, TokenKind> allKeywords[9] =
 { { // IEEE 1364-1995
     KEYWORDS_1364_1995
 }, { // IEEE 1364-2001-noconfig
@@ -360,6 +382,14 @@ const static StringTable<TokenKind> allKeywords[8] =
     NEWKEYWORDS_1800_2005,
     NEWKEYWORDS_1800_2009,
     NEWKEYWORDS_1800_2012
+}, { // IEEE 1800-2023
+    KEYWORDS_1364_1995,
+    NEWKEYWORDS_1364_2001_noconfig,
+    NEWKEYWORDS_1364_2001,
+    NEWKEYWORDS_1364_2005,
+    NEWKEYWORDS_1800_2005,
+    NEWKEYWORDS_1800_2009,
+    NEWKEYWORDS_1800_2012
 } };
 
 // clang-format on
@@ -620,37 +650,43 @@ bool LexerFacts::isKeyword(TokenKind kind) {
     }
 }
 
-TokenKind LexerFacts::getSystemKeywordKind(string_view text) {
-    TokenKind kind;
-    if (systemIdentifierKeywords.lookup(text, kind))
-        return kind;
+TokenKind LexerFacts::getSystemKeywordKind(std::string_view text) {
+    if (auto it = systemIdentifierKeywords.find(text); it != systemIdentifierKeywords.end())
+        return it->second;
     return TokenKind::Unknown;
 }
 
-SyntaxKind LexerFacts::getDirectiveKind(string_view directive) {
-    SyntaxKind kind;
-    if (directiveTable.lookup(directive, kind))
-        return kind;
+SyntaxKind LexerFacts::getDirectiveKind(std::string_view directive, bool enableLegacyProtect) {
+    auto& table = enableLegacyProtect ? directivesWithLegacyProtect : directiveTable;
+    if (auto it = table.find(directive); it != table.end())
+        return it->second;
     return SyntaxKind::MacroUsage;
 }
 
-KeywordVersion LexerFacts::getDefaultKeywordVersion() {
-    return KeywordVersion::v1800_2017;
+KeywordVersion LexerFacts::getDefaultKeywordVersion(LanguageVersion languageVersion) {
+    switch (languageVersion) {
+        case LanguageVersion::v1800_2017:
+            return KeywordVersion::v1800_2017;
+        case LanguageVersion::v1800_2023:
+            return KeywordVersion::v1800_2023;
+        default:
+            SLANG_UNREACHABLE;
+    }
 }
 
-optional<KeywordVersion> LexerFacts::getKeywordVersion(string_view text) {
-    KeywordVersion version;
-    if (keywordVersionTable.lookup(text, version))
-        return version;
+std::optional<KeywordVersion> LexerFacts::getKeywordVersion(std::string_view text) {
+    if (auto it = keywordVersionTable.find(text); it != keywordVersionTable.end())
+        return it->second;
     return std::nullopt;
 }
 
-const StringTable<TokenKind>* LexerFacts::getKeywordTable(KeywordVersion version) {
+const flat_hash_map<std::string_view, TokenKind>* LexerFacts::getKeywordTable(
+    KeywordVersion version) {
     return &allKeywords[(uint8_t)version];
 }
 
 // clang-format off
-string_view LexerFacts::getDirectiveText(SyntaxKind kind) {
+std::string_view LexerFacts::getDirectiveText(SyntaxKind kind) {
     switch (kind) {
         case SyntaxKind::BeginKeywordsDirective: return "`begin_keywords";
         case SyntaxKind::CellDefineDirective: return "`celldefine";
@@ -676,7 +712,7 @@ string_view LexerFacts::getDirectiveText(SyntaxKind kind) {
     }
 }
 
-string_view LexerFacts::getTokenKindText(TokenKind kind) {
+std::string_view LexerFacts::getTokenKindText(TokenKind kind) {
     switch (kind) {
         // punctuation
         case TokenKind::Apostrophe: return "'";
@@ -686,16 +722,13 @@ string_view LexerFacts::getTokenKindText(TokenKind kind) {
         case TokenKind::OpenBracket: return "[";
         case TokenKind::CloseBracket: return "]";
         case TokenKind::OpenParenthesis: return "(";
-        case TokenKind::OpenParenthesisStar: return "(*";
         case TokenKind::CloseParenthesis: return ")";
-        case TokenKind::StarCloseParenthesis: return "*)";
         case TokenKind::Semicolon: return ";";
         case TokenKind::Colon: return ":";
         case TokenKind::ColonEquals: return ":=";
         case TokenKind::ColonSlash: return ":/";
         case TokenKind::DoubleColon: return "::";
         case TokenKind::Comma: return ",";
-        case TokenKind::DotStar: return ".*";
         case TokenKind::Dot: return ".";
         case TokenKind::Slash: return "/";
         case TokenKind::Star: return "*";
@@ -704,6 +737,8 @@ string_view LexerFacts::getTokenKindText(TokenKind kind) {
         case TokenKind::Plus: return "+";
         case TokenKind::DoublePlus: return "++";
         case TokenKind::PlusColon: return "+:";
+        case TokenKind::PlusDivMinus: return "+/-";
+        case TokenKind::PlusModMinus: return "+%-";
         case TokenKind::Minus: return "-";
         case TokenKind::DoubleMinus: return "--";
         case TokenKind::MinusColon: return "-:";
@@ -1019,6 +1054,7 @@ string_view LexerFacts::getTokenKindText(TokenKind kind) {
 
         // directives
         case TokenKind::MacroQuote: return "`\"";
+        case TokenKind::MacroTripleQuote: return "`\"\"\"";
         case TokenKind::MacroEscapedQuote: return "`\\`\"";
         case TokenKind::MacroPaste: return "``";
 
@@ -1027,4 +1063,4 @@ string_view LexerFacts::getTokenKindText(TokenKind kind) {
 }
 // clang-format on
 
-} // namespace slang
+} // namespace slang::parsing
