@@ -2,7 +2,8 @@
 // NumberParser.cpp
 // Helper type to construct numeric tokens
 //
-// File is under the MIT license; see LICENSE for details
+// SPDX-FileCopyrightText: Michael Popoloski
+// SPDX-License-Identifier: MIT
 //------------------------------------------------------------------------------
 #include "slang/parsing/NumberParser.h"
 
@@ -10,7 +11,7 @@
 #include "slang/text/CharInfo.h"
 #include "slang/util/String.h"
 
-namespace slang {
+namespace slang::parsing {
 
 static logic_t getLogicCharValue(char c) {
     switch (c) {
@@ -26,8 +27,9 @@ static logic_t getLogicCharValue(char c) {
     }
 }
 
-NumberParser::NumberParser(Diagnostics& diagnostics, BumpAllocator& alloc) :
-    diagnostics(diagnostics), alloc(alloc) {
+NumberParser::NumberParser(Diagnostics& diagnostics, BumpAllocator& alloc,
+                           LanguageVersion languageVersion) :
+    languageVersion(languageVersion), diagnostics(diagnostics), alloc(alloc) {
 }
 
 void NumberParser::startVector(Token baseToken, Token sizeToken) {
@@ -40,13 +42,17 @@ void NumberParser::startVector(Token baseToken, Token sizeToken) {
     literalBase = baseFlags.base();
     signFlag = baseFlags.isSigned();
 
+    // Diagnose errors where they only provided a sign and no base.
+    if (auto baseRaw = baseToken.rawText(); baseRaw.length() == 2 && ::tolower(baseRaw[1]) == 's')
+        addDiag(diag::ExpectedIntegerBaseAfterSigned, baseToken.location() + 1);
+
     sizeBits = 0;
     if (sizeToken) {
         const SVInt& sizeVal = sizeToken.intValue();
         if (sizeVal == 0) {
             addDiag(diag::LiteralSizeIsZero, sizeToken.location());
         }
-        else if (sizeVal > SVInt::MAX_BITS) {
+        else if (sizeVal > (int)SVInt::MAX_BITS) {
             sizeBits = SVInt::MAX_BITS;
             addDiag(diag::LiteralSizeTooLarge, sizeToken.location()) << (int)SVInt::MAX_BITS;
         }
@@ -57,7 +63,7 @@ void NumberParser::startVector(Token baseToken, Token sizeToken) {
 }
 
 int NumberParser::append(Token token, bool isFirst) {
-    text.appendRange(token.rawText());
+    text.append_range(token.rawText());
 
     // If we've had an error thus far, don't bother doing anything else that
     // might just add more errors on the pile.
@@ -69,7 +75,7 @@ int NumberParser::append(Token token, bool isFirst) {
     valid = false;
 
     // underscore as the first char is not allowed
-    string_view chars = token.rawText();
+    std::string_view chars = token.rawText();
     SourceLocation location = token.location();
     if (isFirst && chars.length() && chars[0] == '_') {
         addDiag(diag::DigitsLeadingUnderscore, location);
@@ -161,56 +167,69 @@ int NumberParser::append(Token token, bool isFirst) {
             }
             break;
         default:
-            THROW_UNREACHABLE;
+            SLANG_UNREACHABLE;
     }
 
     valid = true;
     return -1;
 }
 
-Token NumberParser::finishValue(Token firstToken, bool singleToken) {
+Token NumberParser::finishValue(Token firstToken, bool singleToken, bool isNegated) {
     auto createResult = [&](auto&& val) {
         return Token(alloc, TokenKind::IntegerLiteral, firstToken.trivia(),
                      singleToken ? firstToken.rawText() : toStringView(text.copy(alloc)),
                      firstLocation, std::forward<decltype(val)>(val));
     };
 
+    auto checkOverflow = [&](bitwidth_t computedWidth, const SVInt& value) {
+        // Special case to avoid the warning when we have a minus operator preceeding
+        // a min value signed negative integer, with no unknowns.
+        if (isNegated && signFlag && !hasUnknown && computedWidth == sizeBits + 1 &&
+            value.isNegative() && value.countOnes() == 1) {
+            return;
+        }
+
+        addDiag(diag::VectorLiteralOverflow, firstLocation) << computedWidth;
+    };
+
     if (!valid)
         return createResult(0);
 
-    if (literalBase == LiteralBase::Decimal) {
-        // If we added an x or z, fall through to the general handler below.
+    if (literalBase == LiteralBase::Decimal && !hasUnknown) {
+        // If we added an x or z, we will fall through to the general handler below.
         // Otherwise, optimize for this case by reusing the integer value already
         // computed by the token itself.
-        if (!hasUnknown) {
-            // If no size was specified, just return the value as-is. Otherwise,
-            // resize it to match the desired size. Warn if that will truncate.
-            bitwidth_t width = decimalValue.getBitWidth();
-            SVInt result;
-            if (!sizeBits) {
-                // Unsized numbers are required to be at least 32 bits by the spec.
-                if (width < 32)
-                    result = decimalValue.resize(32);
-                else
-                    result = std::move(decimalValue);
-            }
-            else if (width != sizeBits) {
-                if (width > sizeBits)
-                    addDiag(diag::VectorLiteralOverflow, firstLocation);
-
-                result = decimalValue.resize(sizeBits);
-            }
-            else {
-                result = std::move(decimalValue);
-            }
-
-            result.setSigned(signFlag);
-            return createResult(result);
+        bitwidth_t width = decimalValue.getBitWidth();
+        if (signFlag) {
+            width++;
+            decimalValue = decimalValue.resize(width);
         }
+
+        // If no size was specified, just return the value as-is. Otherwise,
+        // resize it to match the desired size. Warn if that will truncate.
+        SVInt result;
+        if (!sizeBits) {
+            // Unsized numbers are required to be at least 32 bits by the spec.
+            if (width < 32)
+                result = decimalValue.resize(32);
+            else
+                result = std::move(decimalValue);
+        }
+        else if (width != sizeBits) {
+            result = decimalValue.resize(sizeBits);
+            if (width > sizeBits)
+                checkOverflow(width, result);
+        }
+        else {
+            result = std::move(decimalValue);
+        }
+
+        result.setSigned(signFlag);
+        return createResult(std::move(result));
     }
 
     if (digits.empty()) {
-        digits.append(logic_t(0));
+        digits.push_back(logic_t(0));
     }
     else if (literalBase != LiteralBase::Decimal) {
         uint32_t multiplier = 0;
@@ -225,7 +244,7 @@ Token NumberParser::finishValue(Token firstToken, bool singleToken) {
                 multiplier = 4;
                 break;
             default:
-                THROW_UNREACHABLE;
+                SLANG_UNREACHABLE;
         }
 
         // All of the digits in the number require `multiplier` bits, except for
@@ -237,7 +256,11 @@ Token NumberParser::finishValue(Token firstToken, bool singleToken) {
 
         // If the leading digit is unknown however, allow any size.
         if (!digits[0].isUnknown())
-            bits += clog2(digits[0].value + 1);
+            bits += (bitwidth_t)std::bit_width(digits[0].value);
+
+        // Signed numbers need an extra bit for the sign.
+        if (signFlag)
+            bits++;
 
         if (bits > sizeBits) {
             if (sizeBits == 0) {
@@ -249,9 +272,10 @@ Token NumberParser::finishValue(Token firstToken, bool singleToken) {
                 sizeBits = std::max(32u, bits);
             }
             else {
-                // We should warn about overflow here, but the spec says it is valid and
-                // the literal gets truncated. Definitely a warning though.
-                addDiag(diag::VectorLiteralOverflow, firstLocation);
+                auto result = SVInt::fromDigits(sizeBits, literalBase, signFlag, hasUnknown,
+                                                digits);
+                checkOverflow(bits, result);
+                return createResult(std::move(result));
             }
         }
     }
@@ -268,16 +292,16 @@ void NumberParser::addDigit(logic_t digit, int maxValue) {
         hasUnknown = true;
     }
     else {
-        ASSERT(digit.value < maxValue);
+        SLANG_ASSERT(digit.value < maxValue);
         if (digits.size() == 1 && digits.front().value == 0) {
             if (digit.value == 0)
                 return; // at most one leading zero
 
             // If first nonzero not unknown, no leading zeros
-            digits.pop();
+            digits.pop_back();
         }
     }
-    digits.append(digit);
+    digits.push_back(digit);
 }
 
 Diagnostic& NumberParser::addDiag(DiagCode code, SourceLocation location) {
@@ -288,8 +312,10 @@ NumberParser::IntResult NumberParser::reportMissingDigits(Token sizeToken, Token
                                                           Token first) {
     // If we issued this error in response to seeing an EOF token, back up and put
     // the error on the last consumed token instead.
-    SourceLocation errLoc = first.location();
-    if (first.kind == TokenKind::EndOfFile)
+    SourceLocation errLoc;
+    if (first && first.kind != TokenKind::EndOfFile)
+        errLoc = first.location();
+    else
         errLoc = baseToken.location() + baseToken.rawText().size();
 
     addDiag(diag::ExpectedVectorDigits, errLoc);
@@ -304,4 +330,4 @@ void NumberParser::reportIntOverflow(Token token) {
     addDiag(diag::SignedIntegerOverflow, token.location()) << val;
 }
 
-} // namespace slang
+} // namespace slang::parsing
