@@ -2,31 +2,31 @@
 // Parser.cpp
 // SystemVerilog language parser
 //
-// File is under the MIT license; see LICENSE for details
+// SPDX-FileCopyrightText: Michael Popoloski
+// SPDX-License-Identifier: MIT
 //------------------------------------------------------------------------------
 #include "slang/parsing/Parser.h"
 
 #include "slang/diagnostics/ParserDiags.h"
-#include "slang/parsing/Preprocessor.h"
 
-namespace slang {
+namespace slang::parsing {
+
+using namespace syntax;
 
 Parser::Parser(Preprocessor& preprocessor, const Bag& options) :
     ParserBase::ParserBase(preprocessor), factory(alloc),
-    parseOptions(options.getOrDefault<ParserOptions>()), numberParser(getDiagnostics(), alloc) {
+    parseOptions(options.getOrDefault<ParserOptions>()),
+    numberParser(getDiagnostics(), alloc, parseOptions.languageVersion) {
 }
 
 SyntaxNode& Parser::parseGuess() {
     // First try to parse as some kind of declaration.
-    auto attributes = parseAttributes();
-    if (isHierarchyInstantiation(/* requireName */ true))
-        return parseHierarchyInstantiation(attributes);
-    if (isNetDeclaration())
-        return parseNetDeclaration(attributes);
-    if (isVariableDeclaration())
-        return parseVariableDeclaration(attributes);
-
-    ASSERT(attributes.empty());
+    if (isMember()) {
+        bool anyLocalModules = false;
+        auto member = parseMember(SyntaxKind::CompilationUnit, anyLocalModules);
+        SLANG_ASSERT(member);
+        return *member;
+    }
 
     // Try to parse as an expression if possible.
     if (isPossibleExpression(peek().kind)) {
@@ -44,7 +44,7 @@ SyntaxNode& Parser::parseGuess() {
     if (statement.kind == SyntaxKind::EmptyStatement &&
         statement.as<EmptyStatementSyntax>().semicolon.isMissing()) {
 
-        getDiagnostics().pop();
+        getDiagnostics().pop_back();
         auto& unit = parseCompilationUnit();
 
         // If there's only one member, pull it out for convenience
@@ -77,10 +77,11 @@ Token Parser::parseLifetime() {
 
 AnsiPortListSyntax& Parser::parseAnsiPortList(Token openParen) {
     Token closeParen;
-    SmallVectorSized<TokenOrSyntax, 8> buffer;
-    parseList<isPossibleAnsiPort, isEndOfParenList>(
-        buffer, TokenKind::CloseParenthesis, TokenKind::Comma, closeParen, RequireItems::False,
-        diag::ExpectedAnsiPort, [this] { return &parseAnsiPort(); });
+    SmallVector<TokenOrSyntax, 8> buffer;
+    parseList<isPossibleAnsiPort, isEndOfParenList>(buffer, TokenKind::CloseParenthesis,
+                                                    TokenKind::Comma, closeParen,
+                                                    RequireItems::False, diag::ExpectedAnsiPort,
+                                                    [this] { return &parseAnsiPort(); });
     return factory.ansiPortList(openParen, buffer.copy(alloc), closeParen);
 }
 
@@ -94,14 +95,15 @@ ModuleHeaderSyntax& Parser::parseModuleHeader() {
     PortListSyntax* ports = nullptr;
     if (peek(TokenKind::OpenParenthesis)) {
         auto openParen = consume();
-        if (peek(TokenKind::DotStar)) {
-            auto dotStar = consume();
-            ports =
-                &factory.wildcardPortList(openParen, dotStar, expect(TokenKind::CloseParenthesis));
+        if (peek(TokenKind::Dot) && peek(1).kind == TokenKind::Star) {
+            auto dot = consume();
+            auto star = consume();
+            ports = &factory.wildcardPortList(openParen, dot, star,
+                                              expect(TokenKind::CloseParenthesis));
         }
         else if (isNonAnsiPort()) {
             Token closeParen;
-            SmallVectorSized<TokenOrSyntax, 8> buffer;
+            SmallVector<TokenOrSyntax, 8> buffer;
             parseList<isPossibleNonAnsiPort, isEndOfParenList>(
                 buffer, TokenKind::CloseParenthesis, TokenKind::Comma, closeParen,
                 RequireItems::True, diag::ExpectedNonAnsiPort,
@@ -114,7 +116,7 @@ ModuleHeaderSyntax& Parser::parseModuleHeader() {
     }
 
     if (moduleKeyword.kind == TokenKind::PackageKeyword) {
-        optional<SourceRange> errorRange;
+        std::optional<SourceRange> errorRange;
         if (!imports.empty())
             errorRange = imports[0]->sourceRange();
         else if (parameterList)
@@ -142,7 +144,7 @@ ParameterPortListSyntax* Parser::parseParameterPortList() {
 
     Token openParen;
     Token closeParen;
-    span<TokenOrSyntax> parameters;
+    std::span<TokenOrSyntax> parameters;
     parseList<isPossibleParameter, isEndOfParameterList>(
         TokenKind::OpenParenthesis, TokenKind::CloseParenthesis, TokenKind::Comma, openParen,
         parameters, closeParen, RequireItems::False, diag::ExpectedParameterPort,
@@ -164,7 +166,7 @@ PortReferenceSyntax& Parser::parsePortReference() {
 PortExpressionSyntax& Parser::parsePortExpression() {
     if (peek(TokenKind::OpenBrace)) {
         Token openBrace, closeBrace;
-        span<TokenOrSyntax> items;
+        std::span<TokenOrSyntax> items;
 
         parseList<isIdentifierOrComma, isEndOfBracedList>(
             TokenKind::OpenBrace, TokenKind::CloseBrace, TokenKind::Comma, openBrace, items,
@@ -233,7 +235,10 @@ PortHeaderSyntax& Parser::parsePortHeader(Token constKeyword, Token direction) {
         if (!constKeyword && peek(1).kind == TokenKind::Dot &&
             peek(2).kind == TokenKind::Identifier && peek(3).kind == TokenKind::Identifier) {
             auto name = consume();
-            return factory.interfacePortHeader(name, parseDotMemberClause());
+            InterfacePortHeaderSyntax* header =
+                &factory.interfacePortHeader(name, parseDotMemberClause());
+            meta.interfacePorts.push_back(header);
+            return *header;
         }
 
         DataTypeSyntax* type;
@@ -285,7 +290,7 @@ PortDeclarationSyntax& Parser::parsePortDeclaration(AttrList attributes) {
         direction = consume();
 
     // Callers must ensure we don't call with 'const' unless also 'ref'.
-    ASSERT(!constKeyword || direction.kind == TokenKind::RefKeyword);
+    SLANG_ASSERT(!constKeyword || direction.kind == TokenKind::RefKeyword);
 
     auto& header = parsePortHeader(constKeyword, direction);
 
@@ -378,7 +383,7 @@ VariableDimensionSyntax* Parser::parseDimension() {
         }
         default: {
             auto selector = parseElementSelector();
-            ASSERT(selector);
+            SLANG_ASSERT(selector);
             specifier = &factory.rangeDimensionSpecifier(*selector);
             break;
         }
@@ -388,13 +393,13 @@ VariableDimensionSyntax* Parser::parseDimension() {
     return &factory.variableDimension(openBracket, specifier, closeBracket);
 }
 
-span<VariableDimensionSyntax*> Parser::parseDimensionList() {
-    SmallVectorSized<VariableDimensionSyntax*, 4> buffer;
+std::span<VariableDimensionSyntax*> Parser::parseDimensionList() {
+    SmallVector<VariableDimensionSyntax*> buffer;
     while (true) {
         auto dim = parseDimension();
         if (!dim)
             break;
-        buffer.append(dim);
+        buffer.push_back(dim);
     }
     return buffer.copy(alloc);
 }
@@ -409,13 +414,17 @@ DotMemberClauseSyntax* Parser::parseDotMemberClause() {
 
 StructUnionTypeSyntax& Parser::parseStructUnion(SyntaxKind syntaxKind) {
     auto keyword = consume();
-    auto tagged = consumeIf(TokenKind::TaggedKeyword);
+
+    Token taggedOrSoft;
+    if (peek(TokenKind::TaggedKeyword) || peek(TokenKind::SoftKeyword))
+        taggedOrSoft = consume();
+
     auto packed = consumeIf(TokenKind::PackedKeyword);
     auto signing = parseSigning();
     auto openBrace = expect(TokenKind::OpenBrace);
 
     Token closeBrace;
-    SmallVectorSized<StructUnionMemberSyntax*, 8> buffer;
+    SmallVector<StructUnionMemberSyntax*> buffer;
 
     if (openBrace.isMissing())
         closeBrace = missingToken(TokenKind::CloseBrace, openBrace.location());
@@ -439,15 +448,17 @@ StructUnionTypeSyntax& Parser::parseStructUnion(SyntaxKind syntaxKind) {
             }
 
             bitmask<TypeOptions> typeOptions;
-            if (tagged.valid() && keyword.kind == TokenKind::UnionKeyword)
+            if (taggedOrSoft.kind == TokenKind::TaggedKeyword &&
+                keyword.kind == TokenKind::UnionKeyword) {
                 typeOptions = TypeOptions::AllowVoid;
+            }
 
             auto& type = parseDataType(typeOptions);
 
             Token semi;
             auto declarators = parseDeclarators(semi);
 
-            buffer.append(
+            buffer.push_back(
                 &factory.structUnionMember(attributes, randomQualifier, type, declarators, semi));
 
             // If we failed to consume any tokens for this member, skip whatever token is
@@ -468,8 +479,8 @@ StructUnionTypeSyntax& Parser::parseStructUnion(SyntaxKind syntaxKind) {
     auto dims = parseDimensionList();
     if (!packed) {
         if (!dims.empty()) {
-            SourceRange range{ dims.front()->getFirstToken().location(),
-                               dims.back()->getLastToken().range().end() };
+            SourceRange range{dims.front()->getFirstToken().location(),
+                              dims.back()->getLastToken().range().end()};
             addDiag(diag::PackedDimsOnUnpacked, range);
         }
 
@@ -477,10 +488,16 @@ StructUnionTypeSyntax& Parser::parseStructUnion(SyntaxKind syntaxKind) {
             addDiag(diag::UnpackedSigned, signing.range());
     }
 
-    if (keyword.kind == TokenKind::StructKeyword && tagged.valid())
-        addDiag(diag::TaggedStruct, tagged.range());
+    if (keyword.kind == TokenKind::StructKeyword && taggedOrSoft.valid()) {
+        addDiag(diag::TaggedStruct, taggedOrSoft.range()) << taggedOrSoft.valueText();
+    }
+    else if (taggedOrSoft.kind == TokenKind::SoftKeyword &&
+             parseOptions.languageVersion < LanguageVersion::v1800_2023) {
+        addDiag(diag::WrongLanguageVersion, taggedOrSoft.range())
+            << toString(parseOptions.languageVersion);
+    }
 
-    return factory.structUnionType(syntaxKind, keyword, tagged, packed, signing, openBrace,
+    return factory.structUnionType(syntaxKind, keyword, taggedOrSoft, packed, signing, openBrace,
                                    buffer.copy(alloc), closeBrace, dims);
 }
 
@@ -497,7 +514,7 @@ EnumTypeSyntax& Parser::parseEnum() {
     auto openBrace = expect(TokenKind::OpenBrace);
 
     Token closeBrace;
-    span<TokenOrSyntax> declarators;
+    std::span<TokenOrSyntax> declarators;
     if (openBrace.isMissing())
         closeBrace = missingToken(TokenKind::CloseBrace, openBrace.location());
     else
@@ -678,7 +695,7 @@ MemberSyntax& Parser::parseNetDeclaration(AttrList attributes) {
         // itself will only do so for an identifier if the following token is a hash,
         // indicating a timing control.
         auto delay = parseTimingControl();
-        ASSERT(delay);
+        SLANG_ASSERT(delay);
 
         Token semi;
         auto declarators = parseDeclarators(semi);
@@ -738,41 +755,38 @@ MemberSyntax& Parser::parseNetDeclaration(AttrList attributes) {
                                   declarators, semi);
 }
 
+ForwardTypeRestrictionSyntax* Parser::parseTypeRestriction(bool isExpected) {
+    switch (peek().kind) {
+        case TokenKind::EnumKeyword:
+        case TokenKind::StructKeyword:
+        case TokenKind::UnionKeyword:
+        case TokenKind::ClassKeyword:
+            if (isExpected ||
+                (peek(1).kind == TokenKind::Identifier && peek(2).kind == TokenKind::Semicolon)) {
+                return &factory.forwardTypeRestriction(consume(), Token());
+            }
+            break;
+        case TokenKind::InterfaceKeyword: {
+            auto interfaceKeyword = consume();
+            auto classKeyword = expect(TokenKind::ClassKeyword);
+            return &factory.forwardTypeRestriction(interfaceKeyword, classKeyword);
+        }
+        default:
+            break;
+    }
+    return nullptr;
+}
+
 MemberSyntax& Parser::parseVariableDeclaration(AttrList attributes) {
     switch (peek().kind) {
         case TokenKind::TypedefKeyword: {
             auto typedefKeyword = consume();
-            switch (peek().kind) {
-                case TokenKind::EnumKeyword:
-                case TokenKind::StructKeyword:
-                case TokenKind::UnionKeyword:
-                case TokenKind::ClassKeyword:
-                    if (peek(1).kind == TokenKind::Identifier &&
-                        peek(2).kind == TokenKind::Semicolon) {
-
-                        auto keyword = consume();
-                        auto name = consume();
-                        return factory.forwardTypedefDeclaration(attributes, typedefKeyword,
-                                                                 keyword, name, consume());
-                    }
-                    break;
-                case TokenKind::InterfaceKeyword: {
-                    auto interfaceKeyword = consume();
-                    auto classKeyword = expect(TokenKind::ClassKeyword);
-                    auto name = expect(TokenKind::Identifier);
-                    return factory.forwardInterfaceClassTypedefDeclaration(
-                        attributes, typedefKeyword, interfaceKeyword, classKeyword, name,
-                        expect(TokenKind::Semicolon));
-                }
-                case TokenKind::Identifier:
-                    if (peek(1).kind == TokenKind::Semicolon) {
-                        auto name = consume();
-                        return factory.forwardTypedefDeclaration(attributes, typedefKeyword,
-                                                                 Token(), name, consume());
-                    }
-                    break;
-                default:
-                    break;
+            auto restriction = parseTypeRestriction(/* isExpected */ false);
+            if (restriction ||
+                (peek(TokenKind::Identifier) && peek(1).kind == TokenKind::Semicolon)) {
+                auto name = expect(TokenKind::Identifier);
+                return factory.forwardTypedefDeclaration(attributes, typedefKeyword, restriction,
+                                                         name, expect(TokenKind::Semicolon));
             }
 
             auto& type = parseDataType();
@@ -806,7 +820,7 @@ MemberSyntax& Parser::parseVariableDeclaration(AttrList attributes) {
 }
 
 DataDeclarationSyntax& Parser::parseDataDeclaration(AttrList attributes) {
-    SmallVectorSized<Token, 4> modifiers;
+    SmallVector<Token, 4> modifiers;
     SmallMap<TokenKind, Token, 4> modifierSet;
     Token lastLifetime;
     bool hasVar = false;
@@ -816,7 +830,7 @@ DataDeclarationSyntax& Parser::parseDataDeclaration(AttrList attributes) {
 
     while (isDeclarationModifier(peek().kind)) {
         Token t = consume();
-        modifiers.append(t);
+        modifiers.push_back(t);
         if (t.kind == TokenKind::VarKeyword)
             hasVar = true;
 
@@ -893,37 +907,44 @@ DeclaratorSyntax& Parser::parseDeclarator(bool allowMinTypMax, bool requireIniti
 }
 
 template<bool (*IsEnd)(TokenKind)>
-span<TokenOrSyntax> Parser::parseDeclarators(TokenKind endKind, Token& end, bool allowMinTypMax,
-                                             bool requireInitializers) {
-    SmallVectorSized<TokenOrSyntax, 4> buffer;
-    parseList<isIdentifierOrComma, IsEnd>(
-        buffer, endKind, TokenKind::Comma, end, RequireItems::True, diag::ExpectedDeclarator,
-        [this, allowMinTypMax, requireInitializers] {
-            return &parseDeclarator(allowMinTypMax, requireInitializers);
-        });
+std::span<TokenOrSyntax> Parser::parseDeclarators(TokenKind endKind, Token& end,
+                                                  bool allowMinTypMax, bool requireInitializers) {
+    SmallVector<TokenOrSyntax, 4> buffer;
+    parseList<isIdentifierOrComma, IsEnd>(buffer, endKind, TokenKind::Comma, end,
+                                          RequireItems::True, diag::ExpectedDeclarator,
+                                          [this, allowMinTypMax, requireInitializers] {
+                                              return &parseDeclarator(allowMinTypMax,
+                                                                      requireInitializers);
+                                          });
 
     return buffer.copy(alloc);
 }
 
-span<TokenOrSyntax> Parser::parseDeclarators(Token& semi, bool allowMinTypMax,
-                                             bool requireInitializers) {
+std::span<TokenOrSyntax> Parser::parseDeclarators(Token& semi, bool allowMinTypMax,
+                                                  bool requireInitializers) {
     return parseDeclarators<isNotIdOrComma>(TokenKind::Semicolon, semi, allowMinTypMax,
                                             requireInitializers);
 }
 
 Parser::AttrList Parser::parseAttributes() {
-    SmallVectorSized<AttributeInstanceSyntax*, 4> buffer;
-    while (peek(TokenKind::OpenParenthesisStar)) {
-        Token openParen;
-        Token closeParen;
-        span<TokenOrSyntax> list;
+    SmallVector<AttributeInstanceSyntax*> buffer;
+    while (isStartOfAttrs(0)) {
+        Token openParen = consume();
+        Token closeParen, openStar, closeStar;
 
+        std::span<TokenOrSyntax> list;
         parseList<isIdentifierOrComma, isEndOfAttribute>(
-            TokenKind::OpenParenthesisStar, TokenKind::StarCloseParenthesis, TokenKind::Comma,
-            openParen, list, closeParen, RequireItems::True, diag::ExpectedAttribute,
-            [this] { return &parseAttributeSpec(); });
+            TokenKind::Star, TokenKind::Star, TokenKind::Comma, openStar, list, closeStar,
+            RequireItems::True, diag::ExpectedAttribute, [this] { return &parseAttributeSpec(); });
 
-        buffer.append(&factory.attributeInstance(openParen, list, closeParen));
+        if (!closeStar.isMissing()) {
+            closeParen = expect(TokenKind::CloseParenthesis);
+            if (!closeParen.isMissing() && !closeParen.trivia().empty())
+                addDiag(diag::ExpectedToken, closeStar.location()) << "*)"sv;
+        }
+
+        buffer.push_back(
+            &factory.attributeInstance(openParen, openStar, list, closeStar, closeParen));
     }
     return buffer.copy(alloc);
 }
@@ -980,38 +1001,66 @@ ParameterDeclarationBaseSyntax& Parser::parseParameterDecl(Token keyword, Token*
     // this is actually a type reference for a normal parameter.
     if (peek(TokenKind::TypeKeyword) && peek(1).kind != TokenKind::OpenParenthesis) {
         auto typeKeyword = consume();
+        auto restriction = parseTypeRestriction(/* isExpected */ true);
+        if (restriction && parseOptions.languageVersion < LanguageVersion::v1800_2023) {
+            addDiag(diag::WrongLanguageVersion, restriction->sourceRange())
+                << toString(parseOptions.languageVersion);
+        }
 
-        SmallVectorSized<TokenOrSyntax, 4> decls;
+        SmallVector<TokenOrSyntax, 4> decls;
         if (semi) {
-            parseList<isIdentifierOrComma, isSemicolon>(
-                decls, TokenKind::Semicolon, TokenKind::Comma, *semi, RequireItems::True,
-                diag::ExpectedParameterPort, [this] { return &parseTypeAssignment(); });
+            parseList<isIdentifierOrComma, isSemicolon>(decls, TokenKind::Semicolon,
+                                                        TokenKind::Comma, *semi, RequireItems::True,
+                                                        diag::ExpectedParameterPort,
+                                                        [this] { return &parseTypeAssignment(); });
         }
         else {
             while (true) {
-                decls.append(&parseTypeAssignment());
+                decls.push_back(&parseTypeAssignment());
                 if (!peek(TokenKind::Comma) || peek(1).kind != TokenKind::Identifier ||
                     (peek(2).kind != TokenKind::Equals && peek(2).kind != TokenKind::Comma)) {
                     break;
                 }
 
-                decls.append(consume());
+                decls.push_back(consume());
             }
         }
 
-        return factory.typeParameterDeclaration(keyword, typeKeyword, decls.copy(alloc));
+        return factory.typeParameterDeclaration(keyword, typeKeyword, restriction,
+                                                decls.copy(alloc));
     }
     else {
         auto& type = parseDataType(TypeOptions::AllowImplicit);
 
-        // If the semi pointer is given, we should parse a list of decls.
-        // Otherwise we're in a parameter port list and should just parse one.
-        span<TokenOrSyntax> decls;
-        if (semi)
+        // If the semi pointer is given, we should parse a simple list of decls.
+        // Otherwise we're in a parameter port list and don't know if we'll encounter
+        // other non-decl things, so do the parsing manually.
+        std::span<TokenOrSyntax> decls;
+        if (semi) {
             decls = parseDeclarators(*semi, /* allowMinTypMax */ true);
+        }
         else {
-            SmallVectorSized<TokenOrSyntax, 2> buffer;
-            buffer.append(&parseDeclarator(/* allowMinTypMax */ true));
+            SmallVector<TokenOrSyntax, 2> buffer;
+            while (true) {
+                buffer.push_back(&parseDeclarator(/* allowMinTypMax */ true));
+                if (!peek(TokenKind::Comma) || peek(1).kind != TokenKind::Identifier)
+                    break;
+
+                // We need to disambiguate here between another decl which has an optional
+                // unpacked dimension list from a totally new parameter being declared with
+                // an identifier type and an optional packed dimension list. If it's not the
+                // former we will bail out and let our parent parse a new parameter.
+                uint32_t index = 2;
+                if (!scanDimensionList(index))
+                    break;
+
+                if (auto nk = peek(index).kind; nk != TokenKind::Comma && nk != TokenKind::Equals &&
+                                                nk != TokenKind::CloseParenthesis) {
+                    break;
+                }
+
+                buffer.push_back(consume());
+            }
             decls = buffer.copy(alloc);
         }
 
@@ -1026,11 +1075,12 @@ PortConnectionSyntax& Parser::parsePortConnection() {
     if (peek(TokenKind::Comma) || peek(TokenKind::CloseParenthesis))
         return factory.emptyPortConnection(attributes, placeholderToken());
 
-    if (peek(TokenKind::DotStar))
-        return factory.wildcardPortConnection(attributes, consume());
-
     if (peek(TokenKind::Dot)) {
         auto dot = consume();
+
+        if (peek(TokenKind::Star))
+            return factory.wildcardPortConnection(attributes, dot, consume());
+
         auto name = expect(TokenKind::Identifier);
 
         PropertyExprSyntax* expr = nullptr;
@@ -1048,6 +1098,64 @@ PortConnectionSyntax& Parser::parsePortConnection() {
     return factory.orderedPortConnection(attributes, parsePropertyExpr(0));
 }
 
+bool Parser::isMember() {
+    // Any attributes found should indicate a member.
+    uint32_t index = 0;
+    scanAttributes(index);
+    if (index > 0)
+        return true;
+
+    if (isHierarchyInstantiation(/* requireName */ true) || isNetDeclaration() ||
+        isVariableDeclaration()) {
+        return true;
+    }
+
+    switch (peek().kind) {
+        case TokenKind::GenerateKeyword:
+        case TokenKind::TimeUnitKeyword:
+        case TokenKind::TimePrecisionKeyword:
+        case TokenKind::ModuleKeyword:
+        case TokenKind::MacromoduleKeyword:
+        case TokenKind::ProgramKeyword:
+        case TokenKind::PackageKeyword:
+        case TokenKind::InterfaceKeyword:
+        case TokenKind::ModPortKeyword:
+        case TokenKind::BindKeyword:
+        case TokenKind::SpecParamKeyword:
+        case TokenKind::AliasKeyword:
+        case TokenKind::SpecifyKeyword:
+        case TokenKind::AssignKeyword:
+        case TokenKind::InitialKeyword:
+        case TokenKind::FinalKeyword:
+        case TokenKind::AlwaysKeyword:
+        case TokenKind::AlwaysCombKeyword:
+        case TokenKind::AlwaysFFKeyword:
+        case TokenKind::AlwaysLatchKeyword:
+        case TokenKind::GenVarKeyword:
+        case TokenKind::TaskKeyword:
+        case TokenKind::FunctionKeyword:
+        case TokenKind::CoverGroupKeyword:
+        case TokenKind::ClassKeyword:
+        case TokenKind::VirtualKeyword:
+        case TokenKind::DefParamKeyword:
+        case TokenKind::ImportKeyword:
+        case TokenKind::ExportKeyword:
+        case TokenKind::PropertyKeyword:
+        case TokenKind::SequenceKeyword:
+        case TokenKind::CheckerKeyword:
+        case TokenKind::GlobalKeyword:
+        case TokenKind::DefaultKeyword:
+        case TokenKind::ClockingKeyword:
+        case TokenKind::ConstraintKeyword:
+        case TokenKind::PrimitiveKeyword:
+        case TokenKind::ConfigKeyword:
+        case TokenKind::Semicolon:
+            return true;
+        default:
+            return isGateType(peek().kind);
+    }
+}
+
 bool Parser::isPortDeclaration(bool inStatement) {
     uint32_t index = 0;
     if (!scanAttributes(index))
@@ -1275,7 +1383,7 @@ bool Parser::isHierarchyInstantiation(bool requireName) {
     if (peek(index++).kind != TokenKind::Identifier)
         return false;
 
-    // skip over optional parameter value assignment
+    // skip over std::optional parameter value assignment
     if (peek(index).kind == TokenKind::Hash) {
         if (peek(++index).kind != TokenKind::OpenParenthesis)
             return false;
@@ -1344,26 +1452,39 @@ bool Parser::scanQualifiedName(uint32_t& index, bool allowNew) {
 }
 
 bool Parser::scanAttributes(uint32_t& index) {
-    while (peek(index).kind == TokenKind::OpenParenthesisStar) {
+    while (isStartOfAttrs(index)) {
         // scan over attributes
+        index++;
         while (true) {
             auto kind = peek(++index).kind;
             if (kind == TokenKind::EndOfFile)
                 return false;
-            if (kind == TokenKind::StarCloseParenthesis)
+            if (kind == TokenKind::Star && peek(index + 1).kind == TokenKind::CloseParenthesis)
                 break;
         }
-        index++;
+        index += 2;
     }
     return true;
 }
 
-void Parser::errorIfAttributes(AttrList attributes) {
-    if (!attributes.empty())
-        addDiag(diag::AttributesNotAllowed, peek().location());
+bool Parser::isStartOfAttrs(uint32_t index) {
+    if (peek(index).kind == TokenKind::OpenParenthesis) {
+        auto t = peek(index + 1);
+        return t.kind == TokenKind::Star && t.trivia().empty();
+    }
+    return false;
 }
 
-void Parser::checkBlockNames(string_view begin, string_view end, SourceLocation loc) {
+void Parser::errorIfAttributes(AttrList attributes) {
+    if (!attributes.empty()) {
+        auto last = attributes.back()->getLastToken();
+        SourceRange range{attributes.front()->getFirstToken().location(),
+                          last.location() + last.rawText().length()};
+        addDiag(diag::AttributesNotAllowed, range);
+    }
+}
+
+void Parser::checkBlockNames(std::string_view begin, std::string_view end, SourceLocation loc) {
     if (begin.empty() || end.empty())
         return;
 
@@ -1406,7 +1527,7 @@ void Parser::checkBlockNames(const NamedBlockClauseSyntax* beginBlock,
 
 void Parser::handleTooDeep() {
     addDiag(diag::ParseTreeTooDeep, peek().location());
-    throw RecursionException("");
+    SLANG_THROW(RecursionException(""));
 }
 
-} // namespace slang
+} // namespace slang::parsing
