@@ -2,14 +2,17 @@
 // Parser_expressions.cpp
 // Expression-related parsing methods
 //
-// File is under the MIT license; see LICENSE for details
+// SPDX-FileCopyrightText: Michael Popoloski
+// SPDX-License-Identifier: MIT
 //------------------------------------------------------------------------------
 #include "slang/diagnostics/ParserDiags.h"
 #include "slang/parsing/Lexer.h"
 #include "slang/parsing/Parser.h"
 #include "slang/parsing/Preprocessor.h"
 
-namespace slang {
+namespace slang::parsing {
+
+using namespace syntax;
 
 ExpressionSyntax& Parser::parseExpression() {
     return parseSubExpression(ExpressionOptions::None, 0);
@@ -50,7 +53,7 @@ ExpressionSyntax& Parser::parseSubExpression(bitmask<ExpressionOptions> options,
     auto current = peek();
     if (isPossibleDelayOrEventControl(current.kind)) {
         auto timingControl = parseTimingControl();
-        ASSERT(timingControl);
+        SLANG_ASSERT(timingControl);
 
         auto& expr = factory.timingControlExpression(*timingControl, parseExpression());
         return parsePostfixExpression(expr, options);
@@ -60,8 +63,12 @@ ExpressionSyntax& Parser::parseSubExpression(bitmask<ExpressionOptions> options,
         auto member = expect(TokenKind::Identifier);
 
         ExpressionSyntax* expr = nullptr;
-        if (isPossibleExpression(peek().kind))
-            expr = &parseExpression();
+        if (isPossibleExpression(peek().kind)) {
+            // This is not allowed to be a binary expression,
+            // so use a high precedence to avoid taking any
+            // binary operators that may trail our tag expr.
+            expr = &parseSubExpression(ExpressionOptions::None, INT_MAX);
+        }
 
         return factory.taggedUnionExpression(tagged, member, expr);
     }
@@ -99,8 +106,9 @@ ExpressionSyntax& Parser::parseBinaryExpression(ExpressionSyntax* left,
         // either a binary operator, or we're done
         current = peek();
         auto opKind = getBinaryExpression(current.kind);
-        if (opKind == SyntaxKind::Unknown)
+        if (opKind == SyntaxKind::Unknown) {
             break;
+        }
         else if (opKind == SyntaxKind::LogicalImplicationExpression &&
                  options.has(ExpressionOptions::ConstraintContext)) {
             // the implication operator in constraint blocks is special, we don't handle it here
@@ -113,6 +121,11 @@ ExpressionSyntax& Parser::parseBinaryExpression(ExpressionSyntax* left,
             // and not part of this nested sub expression.
             break;
         }
+        else if (opKind == SyntaxKind::MultiplyExpression &&
+                 peek(1).kind == TokenKind::CloseParenthesis && peek(1).trivia().empty()) {
+            // This is an end-of-attribute token, not a multiply.
+            break;
+        }
         else if (opKind == SyntaxKind::LessThanEqualExpression &&
                  options.has(ExpressionOptions::ProceduralAssignmentContext)) {
             // we have to special case '<=', which can be less than or nonblocking assignment
@@ -149,6 +162,9 @@ ExpressionSyntax& Parser::parseBinaryExpression(ExpressionSyntax* left,
             auto attributes = parseAttributes();
             auto& rightOperand = parseSubExpression(options, newPrecedence);
             left = &factory.binaryExpression(opKind, *left, opToken, attributes, rightOperand);
+
+            if (!attributes.empty() && isAssignmentOperator(opKind))
+                errorIfAttributes(attributes);
         }
     }
 
@@ -286,8 +302,8 @@ ExpressionSyntax& Parser::parsePrimaryExpression(bitmask<ExpressionOptions> opti
 }
 
 ExpressionSyntax& Parser::parseIntegerExpression(bool disallowVector) {
-    auto result =
-        disallowVector ? numberParser.parseSimpleInt(*this) : numberParser.parseInteger(*this);
+    auto result = disallowVector ? numberParser.parseSimpleInt(*this)
+                                 : numberParser.parseInteger(*this);
 
     if (result.isSimple)
         return factory.literalExpression(SyntaxKind::IntegerLiteralExpression, result.value);
@@ -296,7 +312,7 @@ ExpressionSyntax& Parser::parseIntegerExpression(bool disallowVector) {
 }
 
 void Parser::handleExponentSplit(Token token, size_t offset) {
-    SmallVectorSized<Token, 4> split;
+    SmallVector<Token, 4> split;
     Lexer::splitTokens(alloc, getDiagnostics(), getPP().getSourceManager(), token, offset,
                        getPP().getCurrentKeywordVersion(), split);
 
@@ -305,46 +321,58 @@ void Parser::handleExponentSplit(Token token, size_t offset) {
 
 ExpressionSyntax& Parser::parseInsideExpression(ExpressionSyntax& expr) {
     auto inside = expect(TokenKind::InsideKeyword);
-    auto& list = parseOpenRangeList();
+    auto& list = parseRangeList();
     return factory.insideExpression(expr, inside, list);
 }
 
-OpenRangeListSyntax& Parser::parseOpenRangeList() {
+RangeListSyntax& Parser::parseRangeList() {
     Token openBrace;
     Token closeBrace;
-    span<TokenOrSyntax> list;
+    std::span<TokenOrSyntax> list;
 
-    parseList<isPossibleOpenRangeElement, isEndOfBracedList>(
+    parseList<isPossibleValueRangeElement, isEndOfBracedList>(
         TokenKind::OpenBrace, TokenKind::CloseBrace, TokenKind::Comma, openBrace, list, closeBrace,
-        RequireItems::True, diag::ExpectedOpenRangeElement,
-        [this] { return &parseOpenRangeElement(); });
+        RequireItems::True, diag::ExpectedValueRangeElement,
+        [this] { return &parseValueRangeElement(); });
 
-    return factory.openRangeList(openBrace, list, closeBrace);
+    return factory.rangeList(openBrace, list, closeBrace);
 }
 
-ExpressionSyntax& Parser::parseOpenRangeElement(bitmask<ExpressionOptions> options) {
+ExpressionSyntax& Parser::parseValueRangeElement(bitmask<ExpressionOptions> options) {
     if (!peek(TokenKind::OpenBracket))
         return parseSubExpression(options, 0);
 
     auto openBracket = consume();
     auto& left = parseExpression();
-    auto colon = expect(TokenKind::Colon);
+
+    Token op;
+    if (peek(TokenKind::PlusDivMinus) || peek(TokenKind::PlusModMinus)) {
+        op = consume();
+        if (parseOptions.languageVersion < LanguageVersion::v1800_2023) {
+            addDiag(diag::WrongLanguageVersion, op.range())
+                << toString(parseOptions.languageVersion);
+        }
+    }
+    else {
+        op = expect(TokenKind::Colon);
+    }
+
     auto& right = parseExpression();
     auto closeBracket = expect(TokenKind::CloseBracket);
-    return factory.openRangeExpression(openBracket, left, colon, right, closeBracket);
+    return factory.valueRangeExpression(openBracket, left, op, right, closeBracket);
 }
 
 ConcatenationExpressionSyntax& Parser::parseConcatenation(Token openBrace,
                                                           ExpressionSyntax* first) {
-    SmallVectorSized<TokenOrSyntax, 8> buffer;
+    SmallVector<TokenOrSyntax, 8> buffer;
     if (first) {
         // it's possible to have just one element in the concatenation list, so check for a close
         // brace
-        buffer.append(first);
+        buffer.push_back(first);
         if (peek(TokenKind::CloseBrace))
             return factory.concatenationExpression(openBrace, buffer.copy(alloc), consume());
 
-        buffer.append(expect(TokenKind::Comma));
+        buffer.push_back(expect(TokenKind::Comma));
     }
 
     Token closeBrace;
@@ -362,7 +390,7 @@ StreamingConcatenationExpressionSyntax& Parser::parseStreamConcatenation(Token o
 
     Token openBraceInner;
     Token closeBraceInner;
-    span<TokenOrSyntax> list;
+    std::span<TokenOrSyntax> list;
 
     parseList<isPossibleExpressionOrComma, isEndOfBracedList>(
         TokenKind::OpenBrace, TokenKind::CloseBrace, TokenKind::Comma, openBraceInner, list,
@@ -398,8 +426,8 @@ AssignmentPatternExpressionSyntax& Parser::parseAssignmentPatternExpression(Data
         // This is an empty pattern -- we'll just warn and continue on.
         addDiag(diag::EmptyAssignmentPattern, openBrace.location());
 
-        auto pattern =
-            &factory.simpleAssignmentPattern(openBrace, span<TokenOrSyntax>{}, consume());
+        auto pattern = &factory.simpleAssignmentPattern(openBrace, std::span<TokenOrSyntax>{},
+                                                        consume());
         return factory.assignmentPatternExpression(type, *pattern);
     }
     else {
@@ -408,13 +436,13 @@ AssignmentPatternExpressionSyntax& Parser::parseAssignmentPatternExpression(Data
 
     Token closeBrace;
     AssignmentPatternSyntax* pattern;
-    SmallVectorSized<TokenOrSyntax, 8> buffer;
+    SmallVector<TokenOrSyntax, 8> buffer;
 
     switch (peek().kind) {
         case TokenKind::Colon:
-            buffer.append(&parseAssignmentPatternItem(firstExpr));
+            buffer.push_back(&parseAssignmentPatternItem(firstExpr));
             if (peek(TokenKind::Comma)) {
-                buffer.append(consume());
+                buffer.push_back(consume());
 
                 parseList<isPossibleExpressionOrCommaOrDefault, isEndOfBracedList>(
                     buffer, TokenKind::CloseBrace, TokenKind::Comma, closeBrace,
@@ -425,8 +453,8 @@ AssignmentPatternExpressionSyntax& Parser::parseAssignmentPatternExpression(Data
                 closeBrace = expect(TokenKind::CloseBrace);
             }
 
-            pattern =
-                &factory.structuredAssignmentPattern(openBrace, buffer.copy(alloc), closeBrace);
+            pattern = &factory.structuredAssignmentPattern(openBrace, buffer.copy(alloc),
+                                                           closeBrace);
             break;
         case TokenKind::OpenBrace: {
             auto innerOpenBrace = consume();
@@ -440,8 +468,8 @@ AssignmentPatternExpressionSyntax& Parser::parseAssignmentPatternExpression(Data
             break;
         }
         case TokenKind::Comma:
-            buffer.append(firstExpr);
-            buffer.append(consume());
+            buffer.push_back(firstExpr);
+            buffer.push_back(consume());
 
             parseList<isPossibleExpressionOrComma, isEndOfBracedList>(
                 buffer, TokenKind::CloseBrace, TokenKind::Comma, closeBrace, RequireItems::True,
@@ -449,14 +477,14 @@ AssignmentPatternExpressionSyntax& Parser::parseAssignmentPatternExpression(Data
             pattern = &factory.simpleAssignmentPattern(openBrace, buffer.copy(alloc), closeBrace);
             break;
         case TokenKind::CloseBrace:
-            buffer.append(firstExpr);
+            buffer.push_back(firstExpr);
             closeBrace = consume();
             pattern = &factory.simpleAssignmentPattern(openBrace, buffer.copy(alloc), closeBrace);
             break;
         default:
             // This is an error case; let the list handling code get us out of it.
-            buffer.append(firstExpr);
-            buffer.append(expect(TokenKind::Comma));
+            buffer.push_back(firstExpr);
+            buffer.push_back(expect(TokenKind::Comma));
 
             parseList<isPossibleExpressionOrComma, isEndOfBracedList>(
                 buffer, TokenKind::CloseBrace, TokenKind::Comma, closeBrace, RequireItems::False,
@@ -464,7 +492,7 @@ AssignmentPatternExpressionSyntax& Parser::parseAssignmentPatternExpression(Data
             pattern = &factory.simpleAssignmentPattern(openBrace, buffer.copy(alloc), closeBrace);
             break;
     }
-    ASSERT(pattern);
+    SLANG_ASSERT(pattern);
     return factory.assignmentPatternExpression(type, *pattern);
 }
 
@@ -529,23 +557,79 @@ bool Parser::isSequenceRepetition() {
 ExpressionSyntax& Parser::parsePostfixExpression(ExpressionSyntax& lhs,
                                                  bitmask<ExpressionOptions> options) {
     ExpressionSyntax* expr = &lhs;
+
+    auto isLiteral = [&] {
+        switch (expr->kind) {
+            case SyntaxKind::NullLiteralExpression:
+            case SyntaxKind::WildcardLiteralExpression:
+            case SyntaxKind::StringLiteralExpression:
+            case SyntaxKind::RealLiteralExpression:
+            case SyntaxKind::TimeLiteralExpression:
+            case SyntaxKind::IntegerLiteralExpression:
+            case SyntaxKind::UnbasedUnsizedLiteralExpression:
+            case SyntaxKind::IntegerVectorExpression:
+                return true;
+            default:
+                return false;
+        }
+    };
+
     while (true) {
         switch (peek().kind) {
             case TokenKind::OpenBracket:
-                if (options.has(ExpressionOptions::SequenceExpr) && isSequenceRepetition())
+                if (isLiteral() ||
+                    (options.has(ExpressionOptions::SequenceExpr) && isSequenceRepetition())) {
                     return *expr;
+                }
 
                 expr = &factory.elementSelectExpression(*expr, parseElementSelect());
                 break;
             case TokenKind::Dot: {
                 auto dot = consume();
-                auto name = expect(TokenKind::Identifier);
+
+                Token name;
+                switch (peek().kind) {
+                    case TokenKind::UniqueKeyword:
+                    case TokenKind::AndKeyword:
+                    case TokenKind::OrKeyword:
+                    case TokenKind::XorKeyword:
+                        name = consume();
+                        break;
+                    default:
+                        name = expect(TokenKind::Identifier);
+                        break;
+                }
+
                 expr = &factory.memberAccessExpression(*expr, dot, name);
                 break;
             }
             case TokenKind::OpenParenthesis: {
-                auto& args = parseArgumentList();
-                expr = &factory.invocationExpression(*expr, nullptr, &args);
+                if (isStartOfAttrs(0)) {
+                    auto attributes = parseAttributes();
+                    switch (peek().kind) {
+                        case TokenKind::DoublePlus:
+                        case TokenKind::DoubleMinus: {
+                            auto op = consume();
+                            return factory.postfixUnaryExpression(
+                                getUnaryPostfixExpression(op.kind), *expr, attributes, op);
+                        }
+                        case TokenKind::OpenParenthesis:
+                            expr = &factory.invocationExpression(*expr, attributes,
+                                                                 &parseArgumentList());
+                            break;
+                        default:
+                            // otherwise, this has to be a function call without any arguments
+                            expr = &factory.invocationExpression(*expr, attributes, nullptr);
+                            break;
+                    }
+                }
+                else {
+                    if (isLiteral())
+                        return *expr;
+
+                    auto& args = parseArgumentList();
+                    expr = &factory.invocationExpression(*expr, nullptr, &args);
+                }
                 break;
             }
             case TokenKind::DoublePlus:
@@ -564,26 +648,6 @@ ExpressionSyntax& Parser::parsePostfixExpression(ExpressionSyntax& lhs,
                 expr = &factory.castExpression(*expr, apostrophe, parenExpr);
                 break;
             }
-            case TokenKind::OpenParenthesisStar: {
-                auto attributes = parseAttributes();
-                switch (peek().kind) {
-                    case TokenKind::DoublePlus:
-                    case TokenKind::DoubleMinus: {
-                        auto op = consume();
-                        return factory.postfixUnaryExpression(getUnaryPostfixExpression(op.kind),
-                                                              *expr, attributes, op);
-                    }
-                    case TokenKind::OpenParenthesis:
-                        expr =
-                            &factory.invocationExpression(*expr, attributes, &parseArgumentList());
-                        break;
-                    default:
-                        // otherwise, this has to be a function call without any arguments
-                        expr = &factory.invocationExpression(*expr, attributes, nullptr);
-                        break;
-                }
-                break;
-            }
             case TokenKind::WithKeyword:
                 // If we see bracket right after the with keyword, this is actually part of a stream
                 // expression -- return and let the call further up the stack handle it.
@@ -666,10 +730,9 @@ NameSyntax& Parser::parseName(bitmask<NameOptions> options) {
         kind = peek().kind;
     }
 
-    // If we saw $unit, $root, super, or local, make sure the correct token follows it.
+    // If we saw super or local, make sure the correct token follows it.
     TokenKind expectedKind = TokenKind::Unknown;
     switch (name->kind) {
-        case SyntaxKind::UnitScope:
         case SyntaxKind::LocalScope:
             expectedKind = TokenKind::DoubleColon;
             break;
@@ -735,8 +798,11 @@ NameSyntax& Parser::parseNamePart(bitmask<NameOptions> options) {
 
     switch (peek().kind) {
         case TokenKind::Hash: {
+            if (options.has(NameOptions::NoClassScope))
+                return factory.identifierName(identifier);
+
             auto parameterValues = parseParameterValueAssignment();
-            ASSERT(parameterValues);
+            SLANG_ASSERT(parameterValues);
             return factory.className(identifier, *parameterValues);
         }
         case TokenKind::OpenBracket: {
@@ -745,7 +811,7 @@ NameSyntax& Parser::parseNamePart(bitmask<NameOptions> options) {
                 // brackets need to be parsed specially because they declare
                 // loop variable names. All the selectors prior can be
                 // parsed as normal selectors.
-                SmallVectorSized<ElementSelectSyntax*, 4> buffer;
+                SmallVector<ElementSelectSyntax*> buffer;
                 do {
                     uint32_t index = 1;
                     scanTypePart<isSemicolon>(index, TokenKind::OpenBracket,
@@ -755,7 +821,7 @@ NameSyntax& Parser::parseNamePart(bitmask<NameOptions> options) {
                         break;
                     }
 
-                    buffer.append(&parseElementSelect());
+                    buffer.push_back(&parseElementSelect());
                 } while (peek(TokenKind::OpenBracket));
 
                 if (buffer.empty())
@@ -764,7 +830,7 @@ NameSyntax& Parser::parseNamePart(bitmask<NameOptions> options) {
                 return factory.identifierSelectName(identifier, buffer.copy(alloc));
             }
             else {
-                SmallVectorSized<ElementSelectSyntax*, 4> buffer;
+                SmallVector<ElementSelectSyntax*> buffer;
                 do {
                     // Inside a sequence expression this could be a repetition directive
                     // instead of a selection.
@@ -775,7 +841,7 @@ NameSyntax& Parser::parseNamePart(bitmask<NameOptions> options) {
                             return factory.identifierSelectName(identifier, buffer.copy(alloc));
                     }
 
-                    buffer.append(&parseElementSelect());
+                    buffer.push_back(&parseElementSelect());
                 } while (peek(TokenKind::OpenBracket));
 
                 return factory.identifierSelectName(identifier, buffer.copy(alloc));
@@ -795,7 +861,7 @@ ParameterValueAssignmentSyntax* Parser::parseParameterValueAssignment() {
 
     Token openParen;
     Token closeParen;
-    span<TokenOrSyntax> list;
+    std::span<TokenOrSyntax> list;
     parseList<isPossibleParamAssignment, isEndOfParenList>(
         TokenKind::OpenParenthesis, TokenKind::CloseParenthesis, TokenKind::Comma, openParen, list,
         closeParen, RequireItems::False, diag::ExpectedArgument,
@@ -810,9 +876,9 @@ ParamAssignmentSyntax& Parser::parseParamValue() {
         auto dot = consume();
         auto name = expect(TokenKind::Identifier);
 
-        auto [innerOpenParen, innerCloseParen, expr] =
-            parseGroupOrSkip(TokenKind::OpenParenthesis, TokenKind::CloseParenthesis,
-                             [this]() { return &parseMinTypMaxExpression(); });
+        auto [innerOpenParen, innerCloseParen,
+              expr] = parseGroupOrSkip(TokenKind::OpenParenthesis, TokenKind::CloseParenthesis,
+                                       [this]() { return &parseMinTypMaxExpression(); });
 
         return factory.namedParamAssignment(dot, name, innerOpenParen, expr, innerCloseParen);
     }
@@ -823,7 +889,7 @@ ParamAssignmentSyntax& Parser::parseParamValue() {
 ArgumentListSyntax& Parser::parseArgumentList() {
     Token openParen;
     Token closeParen;
-    span<TokenOrSyntax> list;
+    std::span<TokenOrSyntax> list;
 
     parseList<isPossibleArgument, isEndOfParenList>(
         TokenKind::OpenParenthesis, TokenKind::CloseParenthesis, TokenKind::Comma, openParen, list,
@@ -843,9 +909,9 @@ ArgumentSyntax& Parser::parseArgument() {
         auto dot = consume();
         auto name = expect(TokenKind::Identifier);
 
-        auto [innerOpenParen, innerCloseParen, expr] =
-            parseGroupOrSkip(TokenKind::OpenParenthesis, TokenKind::CloseParenthesis,
-                             [this]() { return &parsePropertyExpr(0); });
+        auto [innerOpenParen, innerCloseParen,
+              expr] = parseGroupOrSkip(TokenKind::OpenParenthesis, TokenKind::CloseParenthesis,
+                                       [this]() { return &parsePropertyExpr(0); });
 
         return factory.namedArgument(dot, name, innerOpenParen, expr, innerCloseParen);
     }
@@ -861,10 +927,10 @@ PatternSyntax& Parser::parsePattern() {
             return factory.parenthesizedPattern(openParen, pattern,
                                                 expect(TokenKind::CloseParenthesis));
         }
-        case TokenKind::DotStar:
-            return factory.wildcardPattern(consume());
         case TokenKind::Dot: {
             auto dot = consume();
+            if (peek(TokenKind::Star))
+                return factory.wildcardPattern(dot, consume());
             return factory.variablePattern(dot, expect(TokenKind::Identifier));
         }
         case TokenKind::TaggedKeyword: {
@@ -880,7 +946,7 @@ PatternSyntax& Parser::parsePattern() {
         case TokenKind::ApostropheOpenBrace: {
             auto openBrace = consume();
             Token closeBrace;
-            SmallVectorSized<TokenOrSyntax, 4> buffer;
+            SmallVector<TokenOrSyntax, 4> buffer;
 
             if (peek(TokenKind::Identifier) && peek(1).kind == TokenKind::Colon) {
                 parseList<isIdentifierOrComma, isCloseBrace>(
@@ -915,7 +981,7 @@ StructurePatternMemberSyntax& Parser::parseMemberPattern() {
 
 ConditionalPredicateSyntax& Parser::parseConditionalPredicate(ExpressionSyntax& first,
                                                               TokenKind endKind, Token& end) {
-    SmallVectorSized<TokenOrSyntax, 4> buffer;
+    SmallVector<TokenOrSyntax, 4> buffer;
 
     MatchesClauseSyntax* matchesClause = nullptr;
     if (peek(TokenKind::MatchesKeyword)) {
@@ -923,10 +989,10 @@ ConditionalPredicateSyntax& Parser::parseConditionalPredicate(ExpressionSyntax&
         matchesClause = &factory.matchesClause(matches, parsePattern());
     }
 
-    buffer.append(&factory.conditionalPattern(first, matchesClause));
+    buffer.push_back(&factory.conditionalPattern(first, matchesClause));
 
     if (peek(TokenKind::TripleAnd)) {
-        buffer.append(consume());
+        buffer.push_back(consume());
         parseList<isPossibleExpressionOrTripleAnd, isEndOfConditionalPredicate>(
             buffer, endKind, TokenKind::TripleAnd, end, RequireItems::True,
             diag::ExpectedConditionalPattern, [this] { return &parseConditionalPattern(); });
@@ -978,8 +1044,8 @@ EventExpressionSyntax& Parser::parseEventExpression() {
         if (expr.kind == SyntaxKind::SignalEventExpression) {
             auto& see = expr.as<SignalEventExpressionSyntax>();
             if (!see.edge && !see.iffClause) {
-                ExpressionSyntax* newExpr =
-                    &factory.parenthesizedExpression(openParen, *see.expr, closeParen);
+                ExpressionSyntax* newExpr = &factory.parenthesizedExpression(openParen, *see.expr,
+                                                                             closeParen);
 
                 newExpr = &parsePostfixExpression(*newExpr, ExpressionOptions::None);
                 newExpr = &parseBinaryExpression(newExpr, ExpressionOptions::None, 0);
@@ -1024,11 +1090,13 @@ ExpressionSyntax& Parser::parseNewExpression(NameSyntax& newKeyword,
     }
 
     // Enforce rules for super.new placement.
+    bool isSuperNew = false;
     if (newKeyword.kind == SyntaxKind::ScopedName) {
         auto& scoped = newKeyword.as<ScopedNameSyntax>();
         if (scoped.right->kind == SyntaxKind::ConstructorName &&
             scoped.left->getLastToken().kind == TokenKind::SuperKeyword) {
-            if ((options & ExpressionOptions::AllowSuperNewCall) == 0) {
+            isSuperNew = true;
+            if (!options.has(ExpressionOptions::AllowSuperNewCall)) {
                 addDiag(diag::InvalidSuperNew, scoped.right->getFirstToken().location())
                     << newKeyword.sourceRange();
             }
@@ -1039,8 +1107,23 @@ ExpressionSyntax& Parser::parseNewExpression(NameSyntax& newKeyword,
     // new-class has an optional argument list, copy-class has a required expression.
     // An open paren here would be ambiguous between an arg list and a parenthesized
     // expression -- we resolve by always taking the arg list.
-    if (kind == TokenKind::OpenParenthesis)
+    if (kind == TokenKind::OpenParenthesis) {
+        // 1800-2023 added the construct super.new(default) -- check for that here.
+        if (isSuperNew && peek(1).kind == TokenKind::DefaultKeyword) {
+            auto openParen = consume();
+            auto defaultKeyword = consume();
+            auto& result = factory.superNewDefaultedArgsExpression(
+                newKeyword, openParen, defaultKeyword, expect(TokenKind::CloseParenthesis));
+
+            if (parseOptions.languageVersion < LanguageVersion::v1800_2023) {
+                addDiag(diag::WrongLanguageVersion, result.sourceRange())
+                    << toString(parseOptions.languageVersion);
+            }
+
+            return result;
+        }
         return factory.newClassExpression(newKeyword, &parseArgumentList());
+    }
 
     if (isPossibleExpression(kind)) {
         if (newKeyword.kind != SyntaxKind::ConstructorName)
@@ -1112,16 +1195,13 @@ TimingControlSyntax* Parser::parseTimingControl() {
                     return &factory.eventControlWithExpression(
                         at, factory.parenthesizedEventExpression(openParen, eventExpr, closeParen));
                 }
-                case TokenKind::OpenParenthesisStar: {
-                    // Special case since @(*) will be lexed as '@' '(*' ')'
-                    auto openParen = consume();
-                    return &factory.implicitEventControl(at, openParen, Token(),
-                                                         expect(TokenKind::CloseParenthesis));
-                }
                 case TokenKind::Star:
                     return &factory.implicitEventControl(at, Token(), consume(), Token());
+                case TokenKind::SystemIdentifier:
+                    return &factory.eventControl(at,
+                                                 parsePrimaryExpression(ExpressionOptions::None));
                 default:
-                    return &factory.eventControl(at, parseName());
+                    return &factory.eventControl(at, parseName(NameOptions::NoClassScope));
             }
         }
         case TokenKind::RepeatKeyword: {
@@ -1143,7 +1223,7 @@ ExpressionSyntax& Parser::parseArrayOrRandomizeMethod(ExpressionSyntax& expr) {
     ParenExpressionListSyntax* args = nullptr;
     if (peek(TokenKind::OpenParenthesis)) {
         Token openParen, closeParen;
-        span<TokenOrSyntax> items;
+        std::span<TokenOrSyntax> items;
         parseList<isPossibleExpressionOrComma, isEndOfParenList>(
             TokenKind::OpenParenthesis, TokenKind::CloseParenthesis, TokenKind::Comma, openParen,
             items, closeParen, RequireItems::False, diag::ExpectedExpression,
@@ -1208,7 +1288,7 @@ SelectorSyntax* Parser::parseSequenceRange() {
 }
 
 SequenceExprSyntax& Parser::parseDelayedSequenceExpr(SequenceExprSyntax* first) {
-    SmallVectorSized<DelayedSequenceElementSyntax*, 4> elements;
+    SmallVector<DelayedSequenceElementSyntax*> elements;
     do {
         Token op, openBracket, closeBracket;
         SelectorSyntax* selector = nullptr;
@@ -1232,8 +1312,8 @@ SequenceExprSyntax& Parser::parseDelayedSequenceExpr(SequenceExprSyntax* first)
         }
 
         auto& expr = parseSequencePrimary();
-        elements.append(&factory.delayedSequenceElement(hash, delayVal, openBracket, op, selector,
-                                                        closeBracket, expr));
+        elements.push_back(&factory.delayedSequenceElement(hash, delayVal, openBracket, op,
+                                                           selector, closeBracket, expr));
 
     } while (peek(TokenKind::DoubleHash));
 
@@ -1248,7 +1328,6 @@ static bool isBinaryOrPostfixExpression(TokenKind kind) {
     switch (kind) {
         case TokenKind::Dot:
         case TokenKind::OpenParenthesis:
-        case TokenKind::OpenParenthesisStar:
         case TokenKind::Apostrophe:
         case TokenKind::DistKeyword:
         case TokenKind::Question:
@@ -1278,10 +1357,11 @@ SequenceMatchListSyntax* Parser::parseSequenceMatchList(Token& closeParen) {
     }
 
     Token comma;
-    span<TokenOrSyntax> list;
-    parseList<isPossibleArgument, isEndOfParenList>(
-        TokenKind::Comma, TokenKind::CloseParenthesis, TokenKind::Comma, comma, list, closeParen,
-        RequireItems::True, diag::ExpectedExpression, [this] { return &parsePropertyExpr(0); });
+    std::span<TokenOrSyntax> list;
+    parseList<isPossibleArgument, isEndOfParenList>(TokenKind::Comma, TokenKind::CloseParenthesis,
+                                                    TokenKind::Comma, comma, list, closeParen,
+                                                    RequireItems::True, diag::ExpectedExpression,
+                                                    [this] { return &parsePropertyExpr(0); });
 
     return &factory.sequenceMatchList(comma, list);
 }
@@ -1358,7 +1438,7 @@ SequenceExprSyntax& Parser::parseSequencePrimary() {
             return parseDelayedSequenceExpr(nullptr);
         case TokenKind::At: {
             auto event = parseTimingControl();
-            ASSERT(event);
+            SLANG_ASSERT(event);
             return factory.clockingSequenceExpr(*event,
                                                 parseSequenceExpr(0, /* isInProperty */ false));
         }
@@ -1437,7 +1517,7 @@ PropertyExprSyntax& Parser::parseCasePropertyExpr() {
     auto& condition = parseExpressionOrDist();
     auto closeParen = expect(TokenKind::CloseParenthesis);
 
-    SmallVectorSized<PropertyCaseItemSyntax*, 8> itemBuffer;
+    SmallVector<PropertyCaseItemSyntax*> itemBuffer;
     SourceLocation lastDefault;
     bool errored = false;
 
@@ -1456,18 +1536,18 @@ PropertyExprSyntax& Parser::parseCasePropertyExpr() {
             auto colon = consumeIf(TokenKind::Colon);
             auto& expr = parsePropertyExpr(0);
             auto semi = expect(TokenKind::Semicolon);
-            itemBuffer.append(&factory.defaultPropertyCaseItem(def, colon, expr, semi));
+            itemBuffer.push_back(&factory.defaultPropertyCaseItem(def, colon, expr, semi));
         }
         else if (isPossibleExpression(kind)) {
             Token colon;
-            SmallVectorSized<TokenOrSyntax, 8> buffer;
+            SmallVector<TokenOrSyntax, 8> buffer;
             parseList<isPossibleExpressionOrComma, isEndOfCaseItem>(
                 buffer, TokenKind::Colon, TokenKind::Comma, colon, RequireItems::True,
                 diag::ExpectedExpression, [this] { return &parseExpressionOrDist(); });
 
             auto& expr = parsePropertyExpr(0);
             auto semi = expect(TokenKind::Semicolon);
-            itemBuffer.append(
+            itemBuffer.push_back(
                 &factory.standardPropertyCaseItem(buffer.copy(alloc), colon, expr, semi));
         }
         else {
@@ -1488,7 +1568,7 @@ PropertyExprSyntax& Parser::parsePropertyPrimary() {
     switch (current.kind) {
         case TokenKind::At: {
             auto event = parseTimingControl();
-            ASSERT(event);
+            SLANG_ASSERT(event);
 
             // To support clocking events in sampled value system calls,
             // check for an end of an argument and just return if so.
@@ -1530,7 +1610,7 @@ PropertyExprSyntax& Parser::parsePropertyPrimary() {
             // when parsing a property instance argument, we may be looking at an event
             // expression here. The first term of an event expression may be an `iff`
             // expression that gets parsed as a property expression, and we want to allow
-            // a comma separated list of events to follow. Binding will disambiguate this
+            // a comma separated list of events to follow. AST will disambiguate this
             // later when we learn what kind of expression we're actually being used in.
             Token closeParen;
             auto matchList = parseSequenceMatchList(closeParen);
@@ -1665,11 +1745,12 @@ PropertyExprSyntax& Parser::parsePropertyExpr(int precedence) {
             left->kind == SyntaxKind::SimplePropertyExpr &&
             right.kind == SyntaxKind::SimplePropertyExpr) {
 
-            auto& seqExpr = factory.binarySequenceExpr(
-                opKind == SyntaxKind::AndPropertyExpr ? SyntaxKind::AndSequenceExpr
-                                                      : SyntaxKind::OrSequenceExpr,
-                *left->as<SimplePropertyExprSyntax>().expr, opToken,
-                *right.as<SimplePropertyExprSyntax>().expr);
+            auto& seqExpr = factory.binarySequenceExpr(opKind == SyntaxKind::AndPropertyExpr
+                                                           ? SyntaxKind::AndSequenceExpr
+                                                           : SyntaxKind::OrSequenceExpr,
+                                                       *left->as<SimplePropertyExprSyntax>().expr,
+                                                       opToken,
+                                                       *right.as<SimplePropertyExprSyntax>().expr);
 
             left = &factory.simplePropertyExpr(seqExpr);
         }
@@ -1681,4 +1762,4 @@ PropertyExprSyntax& Parser::parsePropertyExpr(int precedence) {
     return *left;
 }
 
-} // namespace slang
+} // namespace slang::parsing
