@@ -2,13 +2,16 @@
 // Parser_statements.cpp
 // Statement-related parsing methods
 //
-// File is under the MIT license; see LICENSE for details
+// SPDX-FileCopyrightText: Michael Popoloski
+// SPDX-License-Identifier: MIT
 //------------------------------------------------------------------------------
 #include "slang/diagnostics/ParserDiags.h"
 #include "slang/parsing/LexerFacts.h"
 #include "slang/parsing/Parser.h"
 
-namespace slang {
+namespace slang::parsing {
+
+using namespace syntax;
 
 StatementSyntax& Parser::parseStatement(bool allowEmpty, bool allowSuperNew) {
     auto dg = setDepthGuard();
@@ -72,7 +75,7 @@ StatementSyntax& Parser::parseStatement(bool allowEmpty, bool allowSuperNew) {
         case TokenKind::DoubleHash:
         case TokenKind::At: {
             auto timingControl = parseTimingControl();
-            ASSERT(timingControl);
+            SLANG_ASSERT(timingControl);
             return factory.timingControlStatement(label, attributes, *timingControl,
                                                   parseStatement());
         }
@@ -166,7 +169,12 @@ StatementSyntax& Parser::parseStatement(bool allowEmpty, bool allowSuperNew) {
 ElseClauseSyntax* Parser::parseElseClause() {
     if (peek(TokenKind::ElseKeyword)) {
         auto elseKeyword = consume();
-        return &factory.elseClause(elseKeyword, parseStatement());
+        auto& stmt = parseStatement();
+        if (stmt.kind == SyntaxKind::ConditionalStatement) {
+            if (auto tok = stmt.as<ConditionalStatementSyntax>().uniqueOrPriority)
+                addDiag(diag::UniquePriorityAfterElse, tok.range());
+        }
+        return &factory.elseClause(elseKeyword, stmt);
     }
     return nullptr;
 }
@@ -178,8 +186,8 @@ ConditionalStatementSyntax& Parser::parseConditionalStatement(NamedLabelSyntax*
     auto openParen = expect(TokenKind::OpenParenthesis);
 
     Token closeParen;
-    auto& predicate =
-        parseConditionalPredicate(parseExpression(), TokenKind::CloseParenthesis, closeParen);
+    auto& predicate = parseConditionalPredicate(parseExpression(), TokenKind::CloseParenthesis,
+                                                closeParen);
     auto& statement = parseStatement();
     auto elseClause = parseElseClause();
 
@@ -193,13 +201,14 @@ ConditionalStatementSyntax& Parser::parseConditionalStatement(NamedLabelSyntax*
 }
 
 template<typename IsItemFunc, typename ParseItemFunc>
-bool Parser::parseCaseItems(TokenKind caseKind, SmallVector<CaseItemSyntax*>& itemBuffer,
+bool Parser::parseCaseItems(TokenKind caseKind, SmallVectorBase<CaseItemSyntax*>& itemBuffer,
                             IsItemFunc&& isItem, ParseItemFunc&& parseItem) {
     SourceLocation lastDefault;
     bool errored = false;
 
     while (true) {
-        auto kind = peek().kind;
+        auto current = peek();
+        auto kind = current.kind;
         if (kind == TokenKind::DefaultKeyword) {
             if (lastDefault && !errored) {
                 auto& diag = addDiag(diag::MultipleDefaultCases, peek().location());
@@ -209,10 +218,15 @@ bool Parser::parseCaseItems(TokenKind caseKind, SmallVector<CaseItemSyntax*>& it
             }
 
             lastDefault = peek().location();
-            itemBuffer.append(&parseDefaultCaseItem());
+            itemBuffer.push_back(&parseDefaultCaseItem());
         }
         else if (isItem(kind)) {
-            itemBuffer.append(parseItem());
+            itemBuffer.push_back(parseItem());
+            if (current == peek()) {
+                // parseItem() didn't consume any tokens, so we'll be stuck
+                // in an infinite loop if we don't skip a token here.
+                skipToken(std::nullopt);
+            }
         }
         else if (kind == TokenKind::EndOfFile || isEndKeyword(kind)) {
             break;
@@ -234,7 +248,7 @@ CaseStatementSyntax& Parser::parseCaseStatement(NamedLabelSyntax* label, AttrLis
     auto closeParen = expect(TokenKind::CloseParenthesis);
 
     Token matchesOrInside;
-    SmallVectorSized<CaseItemSyntax*, 16> itemBuffer;
+    SmallVector<CaseItemSyntax*> itemBuffer;
     bool errored = false;
 
     switch (peek().kind) {
@@ -263,15 +277,15 @@ CaseStatementSyntax& Parser::parseCaseStatement(NamedLabelSyntax* label, AttrLis
             matchesOrInside = consume();
             errored = parseCaseItems(
                 caseKeyword.kind, itemBuffer,
-                [](auto kind) { return isPossibleOpenRangeElement(kind); },
+                [](auto kind) { return isPossibleValueRangeElement(kind); },
                 [this] {
                     Token colon;
-                    SmallVectorSized<TokenOrSyntax, 8> buffer;
+                    SmallVector<TokenOrSyntax, 8> buffer;
 
-                    parseList<isPossibleOpenRangeElement, isEndOfCaseItem>(
+                    parseList<isPossibleValueRangeElement, isEndOfCaseItem>(
                         buffer, TokenKind::Colon, TokenKind::Comma, colon, RequireItems::True,
-                        diag::ExpectedOpenRangeElement,
-                        [this] { return &parseOpenRangeElement(); });
+                        diag::ExpectedValueRangeElement,
+                        [this] { return &parseValueRangeElement(); });
                     return &factory.standardCaseItem(buffer.copy(alloc), colon, parseStatement());
                 });
             break;
@@ -281,7 +295,7 @@ CaseStatementSyntax& Parser::parseCaseStatement(NamedLabelSyntax* label, AttrLis
                 caseKeyword.kind, itemBuffer, [](auto kind) { return isPossibleExpression(kind); },
                 [this] {
                     Token colon;
-                    SmallVectorSized<TokenOrSyntax, 8> buffer;
+                    SmallVector<TokenOrSyntax, 8> buffer;
 
                     parseList<isPossibleExpressionOrComma, isEndOfCaseItem>(
                         buffer, TokenKind::Colon, TokenKind::Comma, colon, RequireItems::True,
@@ -342,9 +356,9 @@ SyntaxNode& Parser::parseForInitializer() {
         auto varKeyword = consumeIf(TokenKind::VarKeyword);
         auto& type = parseDataType();
 
-        return factory.forVariableDeclaration(
-            varKeyword, &type,
-            parseDeclarator(/* allowMinTypMax */ false, /* requireInitializers */ true));
+        return factory.forVariableDeclaration(varKeyword, &type,
+                                              parseDeclarator(/* allowMinTypMax */ false,
+                                                              /* requireInitializers */ true));
     }
 
     return factory.forVariableDeclaration(Token(), nullptr, parseDeclarator());
@@ -356,7 +370,7 @@ ForLoopStatementSyntax& Parser::parseForLoopStatement(NamedLabelSyntax* label,
     auto openParen = expect(TokenKind::OpenParenthesis);
 
     Token semi1;
-    SmallVectorSized<TokenOrSyntax, 4> initializers;
+    SmallVector<TokenOrSyntax, 4> initializers;
 
     if (isVariableDeclaration()) {
         parseList<isPossibleForInitializer, isEndOfParenList>(
@@ -389,7 +403,7 @@ ForLoopStatementSyntax& Parser::parseForLoopStatement(NamedLabelSyntax* label,
     }
 
     Token closeParen;
-    SmallVectorSized<TokenOrSyntax, 4> steps;
+    SmallVector<TokenOrSyntax, 4> steps;
     parseList<isPossibleExpressionOrComma, isEndOfParenList>(
         steps, TokenKind::CloseParenthesis, TokenKind::Comma, closeParen, RequireItems::False,
         diag::ExpectedExpression, [this] { return &parseExpression(); });
@@ -418,7 +432,7 @@ ForeachLoopListSyntax& Parser::parseForeachLoopVariables() {
     if (arrayName.kind == SyntaxKind::IdentifierSelectName)
         addDiag(diag::NonstandardForeach, arrayName.sourceRange());
 
-    span<TokenOrSyntax> list;
+    std::span<TokenOrSyntax> list;
     Token openBracket;
     Token closeBracket;
     parseList<isIdentifierOrComma, isEndOfBracketedList>(
@@ -511,7 +525,7 @@ StatementSyntax& Parser::parseAssertionStatement(NamedLabelSyntax* label, AttrLi
             assertionKind = SyntaxKind::ImmediateCoverStatement;
             break;
         default:
-            THROW_UNREACHABLE;
+            SLANG_UNREACHABLE;
     }
 
     Token keyword = consume();
@@ -529,8 +543,8 @@ StatementSyntax& Parser::parseAssertionStatement(NamedLabelSyntax* label, AttrLi
 
     auto openParen = expect(TokenKind::OpenParenthesis);
     auto& expr = parseExpression();
-    auto& parenExpr =
-        factory.parenthesizedExpression(openParen, expr, expect(TokenKind::CloseParenthesis));
+    auto& parenExpr = factory.parenthesizedExpression(openParen, expr,
+                                                      expect(TokenKind::CloseParenthesis));
     auto& actionBlock = parseActionBlock();
     return factory.immediateAssertionStatement(assertionKind, label, attributes, keyword, deferred,
                                                parenExpr, actionBlock);
@@ -570,7 +584,7 @@ ConcurrentAssertionStatementSyntax& Parser::parseConcurrentAssertion(NamedLabelS
             kind = SyntaxKind::ExpectPropertyStatement;
             break;
         default:
-            THROW_UNREACHABLE;
+            SLANG_UNREACHABLE;
     }
 
     auto openParen = expect(TokenKind::OpenParenthesis);
@@ -578,8 +592,9 @@ ConcurrentAssertionStatementSyntax& Parser::parseConcurrentAssertion(NamedLabelS
     auto closeParen = expect(TokenKind::CloseParenthesis);
     auto& action = parseActionBlock();
 
-    return factory.concurrentAssertionStatement(
-        kind, label, attributes, keyword, propertyOrSequence, openParen, spec, closeParen, action);
+    return factory.concurrentAssertionStatement(kind, label, attributes, keyword,
+                                                propertyOrSequence, openParen, spec, closeParen,
+                                                action);
 }
 
 PropertySpecSyntax& Parser::parsePropertySpec() {
@@ -593,8 +608,8 @@ PropertySpecSyntax& Parser::parsePropertySpec() {
         auto iff = expect(TokenKind::IffKeyword);
         auto openParen = expect(TokenKind::OpenParenthesis);
         auto& expr = parseExpressionOrDist();
-        disable =
-            &factory.disableIff(keyword, iff, openParen, expr, expect(TokenKind::CloseParenthesis));
+        disable = &factory.disableIff(keyword, iff, openParen, expr,
+                                      expect(TokenKind::CloseParenthesis));
     }
 
     return factory.propertySpec(timing, disable, parsePropertyExpr(0));
@@ -608,7 +623,8 @@ ActionBlockSyntax& Parser::parseActionBlock() {
         elseClause = parseElseClause();
     else {
         statement = &parseStatement();
-        elseClause = parseElseClause();
+        if (statement->kind != SyntaxKind::EmptyStatement)
+            elseClause = parseElseClause();
     }
 
     return factory.actionBlock(statement, elseClause);
@@ -630,8 +646,8 @@ NamedBlockClauseSyntax* Parser::parseNamedBlockClause() {
     return nullptr;
 }
 
-span<SyntaxNode*> Parser::parseBlockItems(TokenKind endKind, Token& end, bool inConstructor) {
-    SmallVectorSized<SyntaxNode*, 16> buffer;
+std::span<SyntaxNode*> Parser::parseBlockItems(TokenKind endKind, Token& end, bool inConstructor) {
+    SmallVector<SyntaxNode*, 16> buffer;
     auto kind = peek().kind;
     bool errored = false;
     bool sawStatement = false;
@@ -665,7 +681,7 @@ span<SyntaxNode*> Parser::parseBlockItems(TokenKind endKind, Token& end, bool in
         }
 
         if (newNode) {
-            buffer.append(newNode);
+            buffer.push_back(newNode);
             errored = false;
 
             if (!erroredAboutDecls && !isStmt && sawStatement) {
@@ -728,12 +744,13 @@ WaitOrderStatementSyntax& Parser::parseWaitOrderStatement(NamedLabelSyntax* labe
                                                           AttrList attributes) {
     auto keyword = consume();
     auto openParen = expect(TokenKind::OpenParenthesis);
-    SmallVectorSized<TokenOrSyntax, 4> buffer;
+    SmallVector<TokenOrSyntax, 4> buffer;
 
     Token closeParen;
-    parseList<isIdentifierOrComma, isEndOfParenList>(
-        buffer, TokenKind::CloseParenthesis, TokenKind::Comma, closeParen, RequireItems::True,
-        diag::ExpectedIdentifier, [this] { return &parseName(); });
+    parseList<isIdentifierOrComma, isEndOfParenList>(buffer, TokenKind::CloseParenthesis,
+                                                     TokenKind::Comma, closeParen,
+                                                     RequireItems::True, diag::ExpectedIdentifier,
+                                                     [this] { return &parseName(); });
 
     return factory.waitOrderStatement(label, attributes, keyword, openParen, buffer.copy(alloc),
                                       closeParen, parseActionBlock());
@@ -742,18 +759,18 @@ WaitOrderStatementSyntax& Parser::parseWaitOrderStatement(NamedLabelSyntax* labe
 RandCaseStatementSyntax& Parser::parseRandCaseStatement(NamedLabelSyntax* label,
                                                         AttrList attributes) {
     auto randCase = consume();
-    SmallVectorSized<RandCaseItemSyntax*, 16> itemBuffer;
+    SmallVector<RandCaseItemSyntax*> itemBuffer;
 
     while (isPossibleExpression(peek().kind)) {
+        Token curr = peek();
         auto& expr = parseExpression();
         auto colon = expect(TokenKind::Colon);
-        const auto loc = peek().location();
         auto& stmt = parseStatement();
-        if (stmt.kind == SyntaxKind::EmptyStatement &&
-            stmt.as<EmptyStatementSyntax>().semicolon.isMissing() && loc == peek().location()) {
+        // If there is no consumed tokens then expression and statement were not parsed
+        if (curr == peek())
             skipToken(std::nullopt);
-        }
-        itemBuffer.append(&factory.randCaseItem(expr, colon, stmt));
+
+        itemBuffer.push_back(&factory.randCaseItem(expr, colon, stmt));
     }
 
     auto endcase = expect(TokenKind::EndCaseKeyword);
@@ -812,7 +829,7 @@ RsCaseSyntax& Parser::parseRsCase() {
     auto& condition = parseExpression();
     auto closeParen = expect(TokenKind::CloseParenthesis);
 
-    SmallVectorSized<RsCaseItemSyntax*, 8> itemBuffer;
+    SmallVector<RsCaseItemSyntax*> itemBuffer;
     SourceLocation lastDefault;
     bool errored = false;
 
@@ -831,18 +848,19 @@ RsCaseSyntax& Parser::parseRsCase() {
             auto colon = consumeIf(TokenKind::Colon);
             auto& item = parseRsProdItem();
             auto semi = expect(TokenKind::Semicolon);
-            itemBuffer.append(&factory.defaultRsCaseItem(def, colon, item, semi));
+            itemBuffer.push_back(&factory.defaultRsCaseItem(def, colon, item, semi));
         }
         else if (isPossibleExpression(kind)) {
             Token colon;
-            SmallVectorSized<TokenOrSyntax, 8> buffer;
+            SmallVector<TokenOrSyntax, 8> buffer;
             parseList<isPossibleExpressionOrComma, isEndOfCaseItem>(
                 buffer, TokenKind::Colon, TokenKind::Comma, colon, RequireItems::True,
                 diag::ExpectedExpression, [this] { return &parseExpression(); });
 
             auto& item = parseRsProdItem();
             auto semi = expect(TokenKind::Semicolon);
-            itemBuffer.append(&factory.standardRsCaseItem(buffer.copy(alloc), colon, item, semi));
+            itemBuffer.push_back(
+                &factory.standardRsCaseItem(buffer.copy(alloc), colon, item, semi));
         }
         else {
             break;
@@ -912,13 +930,13 @@ RsRuleSyntax& Parser::parseRsRule() {
         randJoin = &factory.randJoinClause(rand, join, parenExpr);
     }
 
-    SmallVectorSized<RsProdSyntax*, 16> prods;
+    SmallVector<RsProdSyntax*> prods;
     while (true) {
         auto prod = parseRsProd();
         if (!prod)
             break;
 
-        prods.append(prod);
+        prods.push_back(prod);
         if (randJoin && prod->kind != SyntaxKind::RsProdItem)
             addDiag(diag::RandJoinProdItem, prod->sourceRange());
     }
@@ -926,7 +944,7 @@ RsRuleSyntax& Parser::parseRsRule() {
     if (randJoin && prods.size() < 2) {
         SourceRange range = randJoin->sourceRange();
         if (!prods.empty())
-            range = SourceRange{ range.start(), prods.back()->getLastToken().range().end() };
+            range = SourceRange{range.start(), prods.back()->getLastToken().range().end()};
 
         addDiag(diag::RandJoinNotEnough, range);
     }
@@ -958,11 +976,11 @@ ProductionSyntax& Parser::parseProduction() {
     }
 
     auto name = expect(TokenKind::Identifier);
-    auto ports = parseFunctionPortList(/* allowEmptyNames */ false);
+    auto ports = parseFunctionPortList({});
     auto colon = expect(TokenKind::Colon);
 
     Token semi;
-    SmallVectorSized<TokenOrSyntax, 8> buffer;
+    SmallVector<TokenOrSyntax, 8> buffer;
     parseList<isPossibleRsRule, isSemicolon>(buffer, TokenKind::Semicolon, TokenKind::Or, semi,
                                              RequireItems::True, diag::ExpectedRsRule,
                                              [this] { return &parseRsRule(); });
@@ -976,9 +994,14 @@ StatementSyntax& Parser::parseRandSequenceStatement(NamedLabelSyntax* label, Att
     auto firstProd = consumeIf(TokenKind::Identifier);
     auto closeParen = expect(TokenKind::CloseParenthesis);
 
-    SmallVectorSized<ProductionSyntax*, 16> productions;
-    while (isPossibleDataType(peek().kind))
-        productions.append(&parseProduction());
+    SmallVector<ProductionSyntax*> productions;
+    while (isPossibleDataType(peek().kind)) {
+        Token curr = peek();
+        productions.push_back(&parseProduction());
+        // If there is no consumed tokens then production was not parsed
+        if (curr == peek())
+            skipToken(std::nullopt);
+    }
 
     if (productions.empty())
         addDiag(diag::ExpectedRsRule, peek().location());
@@ -993,7 +1016,8 @@ StatementSyntax& Parser::parseCheckerStatement(NamedLabelSyntax* label, AttrList
     return factory.checkerInstanceStatement(label, attributes, instance);
 }
 
-void Parser::checkEmptyBody(const SyntaxNode& syntax, Token prevToken, string_view syntaxName) {
+void Parser::checkEmptyBody(const SyntaxNode& syntax, Token prevToken,
+                            std::string_view syntaxName) {
     if (syntax.kind != SyntaxKind::EmptyStatement || prevToken.isMissing())
         return;
 
@@ -1006,4 +1030,4 @@ void Parser::checkEmptyBody(const SyntaxNode& syntax, Token prevToken, string_vi
     addDiag(diag::EmptyBody, ess.semicolon.location()) << syntaxName;
 }
 
-} // namespace slang
+} // namespace slang::parsing
