@@ -2,7 +2,8 @@
 // Preprocessor_macros.cpp
 // Macro-related preprocessor support
 //
-// File is under the MIT license; see LICENSE for details
+// SPDX-FileCopyrightText: Michael Popoloski
+// SPDX-License-Identifier: MIT
 //------------------------------------------------------------------------------
 #include "slang/diagnostics/LexerDiags.h"
 #include "slang/diagnostics/PreprocessorDiags.h"
@@ -12,12 +13,14 @@
 #include "slang/text/SourceManager.h"
 #include "slang/util/String.h"
 
-namespace slang {
+namespace slang::parsing {
+
+using namespace syntax;
 
 using LF = LexerFacts;
 
 Preprocessor::MacroDef Preprocessor::findMacro(Token directive) {
-    string_view name = directive.valueText().substr(1);
+    std::string_view name = directive.valueText().substr(1);
     if (!name.empty() && name[0] == '\\')
         name = name.substr(1);
 
@@ -27,60 +30,72 @@ Preprocessor::MacroDef Preprocessor::findMacro(Token directive) {
     return it->second;
 }
 
-void Preprocessor::createBuiltInMacro(string_view name, int value, string_view valueStr) {
+void Preprocessor::createBuiltInMacro(std::string_view name, int value, std::string_view valueStr) {
 #define NL SourceLocation::NoLocation
 
     if (valueStr.empty()) {
         auto str = std::to_string(value);
         auto ptr = (char*)alloc.allocate(str.length(), 1);
         memcpy(ptr, str.data(), str.length());
-        valueStr = string_view(ptr, str.length());
+        valueStr = std::string_view(ptr, str.length());
     }
 
     Token directive(alloc, TokenKind::Directive, {}, valueStr, NL, SyntaxKind::DefineDirective);
     Token nameTok(alloc, TokenKind::Identifier, {}, name, NL);
 
-    SmallVectorSized<Token, 2> body;
-    body.append(Token(alloc, TokenKind::IntegerLiteral, {}, valueStr, NL,
-                      SVInt(32, uint64_t(value), true)));
+    SmallVector<Token> body;
+    body.push_back(Token(alloc, TokenKind::IntegerLiteral, {}, valueStr, NL,
+                         SVInt(32, uint64_t(value), true)));
 
     MacroDef def;
-    def.syntax =
-        alloc.emplace<DefineDirectiveSyntax>(directive, nameTok, nullptr, body.copy(alloc));
+    def.syntax = alloc.emplace<DefineDirectiveSyntax>(directive, nameTok, nullptr,
+                                                      body.copy(alloc));
     def.builtIn = true;
     macros[name] = def;
 
 #undef NL
 }
 
-MacroActualArgumentListSyntax* Preprocessor::handleTopLevelMacro(Token directive) {
+std::pair<MacroActualArgumentListSyntax*, Trivia> Preprocessor::handleTopLevelMacro(
+    Token directive) {
     auto macro = findMacro(directive);
     if (!macro.valid()) {
+        if (options.ignoreDirectives.find(directive.valueText().substr(1)) !=
+            options.ignoreDirectives.end()) {
+            SmallVector<Token, 4> ignoreTokens;
+            while (peekSameLine() && peek().kind != TokenKind::EndOfFile)
+                ignoreTokens.push_back(consume());
+
+            if (ignoreTokens.empty())
+                return {nullptr, Trivia()};
+            else
+                return {nullptr, Trivia(TriviaKind::SkippedTokens, ignoreTokens.copy(alloc))};
+        }
         addDiag(diag::UnknownDirective, directive.location()) << directive.valueText();
 
         // If we see a parenthesis next, let's assume they tried to invoke a function-like macro
         // and skip over the tokens.
         if (peek(TokenKind::OpenParenthesis))
-            return MacroParser(*this).parseActualArgumentList(directive);
-        return nullptr;
+            return {MacroParser(*this).parseActualArgumentList(directive), Trivia()};
+        return {nullptr, Trivia()};
     }
 
     // if this assert fires, we failed to fully expand nested macros at a previous point
-    ASSERT(!currentMacroToken);
+    SLANG_ASSERT(!currentMacroToken);
 
     // parse arguments if necessary
     MacroActualArgumentListSyntax* actualArgs = nullptr;
     if (macro.needsArgs()) {
         actualArgs = MacroParser(*this).parseActualArgumentList(directive);
         if (!actualArgs)
-            return nullptr;
+            return {nullptr, Trivia()};
     }
 
     // Expand out the macro
-    SmallVectorSized<Token, 32> buffer;
-    MacroExpansion expansion{ sourceManager, alloc, buffer, directive, true };
+    SmallVector<Token, 32> buffer;
+    MacroExpansion expansion{sourceManager, alloc, buffer, directive, true};
     if (!expandMacro(macro, expansion, actualArgs))
-        return actualArgs;
+        return {actualArgs, Trivia()};
 
     // The macro is now expanded out into tokens, but some of those tokens might
     // be more macros that need to be expanded, or special characters that
@@ -91,7 +106,7 @@ MacroActualArgumentListSyntax* Preprocessor::handleTopLevelMacro(Token directive
     if (!macro.isIntrinsic())
         alreadyExpanded.insert(macro.syntax);
 
-    span<Token const> tokens = buffer.copy(alloc);
+    std::span<Token const> tokens = buffer.copy(alloc);
     while (true) {
         // Start by recursively expanding out all valid macro usages. We keep track of
         // the token pointer here so that we can detect if expandReplacementList actually
@@ -99,7 +114,7 @@ MacroActualArgumentListSyntax* Preprocessor::handleTopLevelMacro(Token directive
         // pass. This ensures that we don't miss expanding a constructed macro.
         const Token* ptr = tokens.data();
         if (!expandReplacementList(tokens, alreadyExpanded))
-            return actualArgs;
+            return {actualArgs, Trivia()};
 
         // Now that all macros have been expanded, handle token concatenation and stringification.
         expandedTokens.clear();
@@ -114,15 +129,16 @@ MacroActualArgumentListSyntax* Preprocessor::handleTopLevelMacro(Token directive
     if (!expandedTokens.empty())
         currentMacroToken = expandedTokens.begin();
 
-    return actualArgs;
+    return {actualArgs, Trivia()};
 }
 
-bool Preprocessor::applyMacroOps(span<Token const> tokens, SmallVector<Token>& dest) {
-    SmallVectorSized<Trivia, 16> emptyArgTrivia;
-    SmallVectorSized<Token, 16> stringifyBuffer;
-    SmallVectorSized<Token, 16> commentBuffer;
+bool Preprocessor::applyMacroOps(std::span<Token const> tokens, SmallVectorBase<Token>& dest) {
+    SmallVector<Trivia, 8> emptyArgTrivia;
+    SmallVector<Token, 8> stringifyBuffer;
+    SmallVector<Token, 8> commentBuffer;
     Token stringify;
     Token syntheticComment;
+    Token extraToAppend;
     bool anyNewMacros = false;
     bool didConcat = false;
 
@@ -136,15 +152,40 @@ bool Preprocessor::applyMacroOps(span<Token const> tokens, SmallVector<Token>& d
         Token token = tokens[i];
         switch (token.kind) {
             case TokenKind::MacroQuote:
+            case TokenKind::MacroTripleQuote:
                 if (!stringify) {
                     stringify = token;
                     stringifyBuffer.clear();
+
+                    if (token.kind == TokenKind::MacroTripleQuote &&
+                        options.languageVersion < LanguageVersion::v1800_2023) {
+                        addDiag(diag::WrongLanguageVersion, token.location())
+                            << toString(options.languageVersion);
+                    }
+                }
+                else if (token.kind == stringify.kind) {
+                    // all done stringifying; convert saved tokens to string
+                    newToken = Lexer::stringify(*lexerStack.back(), stringify, stringifyBuffer,
+                                                token);
+                    stringify = Token();
+                }
+                else if (stringify.kind == TokenKind::MacroTripleQuote) {
+                    // We found a `" inside of a triple quoted stringification.
+                    // Just keep it, let it become part of the string contents.
+                    newToken = token;
                 }
                 else {
-                    // all done stringifying; convert saved tokens to string
-                    newToken = Lexer::stringify(alloc, stringify.location(), stringify.trivia(),
-                                                stringifyBuffer.begin(), stringifyBuffer.end());
+                    // We found a `""" inside of a single quoted stringification.
+                    // The LRM doesn't say what to do, but I think it makes sense to
+                    // split the token, end the previous stringification, and then
+                    // append the essentially empty string literal after it.
+                    // This will cause an error down the line since two string literals
+                    // next to each other isn't ever valid.
+                    newToken = Lexer::stringify(*lexerStack.back(), stringify, stringifyBuffer,
+                                                token);
                     stringify = Token();
+                    extraToAppend = Token(alloc, TokenKind::StringLiteral, {}, "\"\"",
+                                          token.location() + 2, ""sv);
                 }
                 break;
             case TokenKind::MacroPaste:
@@ -158,18 +199,20 @@ bool Preprocessor::applyMacroOps(span<Token const> tokens, SmallVector<Token>& d
 
                     // We're ignoring this token, but don't lose its trivia or our
                     // spacing can get messed up.
-                    emptyArgTrivia.appendRange(token.trivia());
+                    emptyArgTrivia.append_range(token.trivia());
                 }
                 else if (stringify) {
                     // If this is right after the opening quote or right before the closing quote,
                     // we're trying to concatenate something with nothing.
-                    if (stringifyBuffer.empty() || tokens[i + 1].kind == TokenKind::MacroQuote)
+                    if (stringifyBuffer.empty() || tokens[i + 1].kind == TokenKind::MacroQuote ||
+                        tokens[i + 1].kind == TokenKind::MacroTripleQuote) {
                         addDiag(diag::IgnoredMacroPaste, token.location());
+                    }
                     else {
-                        newToken =
-                            Lexer::concatenateTokens(alloc, stringifyBuffer.back(), tokens[i + 1]);
+                        newToken = Lexer::concatenateTokens(alloc, stringifyBuffer.back(),
+                                                            tokens[i + 1]);
                         if (newToken) {
-                            stringifyBuffer.pop();
+                            stringifyBuffer.pop_back();
                             ++i;
                         }
                     }
@@ -179,12 +222,11 @@ bool Preprocessor::applyMacroOps(span<Token const> tokens, SmallVector<Token>& d
                     // since this is just going to become a comment anyway.
                     if (commentBuffer.back().kind == TokenKind::Star &&
                         tokens[i + 1].kind == TokenKind::Slash) {
-                        commentBuffer.append(tokens[i + 1]);
+                        commentBuffer.push_back(tokens[i + 1]);
                         i++;
 
-                        emptyArgTrivia.appendRange(syntheticComment.trivia());
-                        emptyArgTrivia.append(
-                            Lexer::commentify(alloc, commentBuffer.begin(), commentBuffer.end()));
+                        emptyArgTrivia.append_range(syntheticComment.trivia());
+                        emptyArgTrivia.push_back(Lexer::commentify(alloc, commentBuffer));
                         syntheticComment = Token();
                     }
                 }
@@ -199,16 +241,16 @@ bool Preprocessor::applyMacroOps(span<Token const> tokens, SmallVector<Token>& d
                     if (left.kind == TokenKind::Slash && right.kind == TokenKind::Star) {
                         commentBuffer.clear();
                         syntheticComment = left;
-                        dest.pop();
+                        dest.pop_back();
                         ++i;
 
-                        commentBuffer.append(left.withTrivia(alloc, {}));
+                        commentBuffer.push_back(left.withTrivia(alloc, {}));
                         newToken = right;
                     }
                     else {
                         newToken = Lexer::concatenateTokens(alloc, dest.back(), tokens[i + 1]);
                         if (newToken) {
-                            dest.pop();
+                            dest.pop_back();
                             ++i;
 
                             nextDidConcat = true;
@@ -225,7 +267,7 @@ bool Preprocessor::applyMacroOps(span<Token const> tokens, SmallVector<Token>& d
                 if (didConcat && token.trivia().empty() && emptyArgTrivia.empty()) {
                     newToken = Lexer::concatenateTokens(alloc, dest.back(), token);
                     if (newToken) {
-                        dest.pop();
+                        dest.pop_back();
                         nextDidConcat = true;
                         break;
                     }
@@ -245,51 +287,62 @@ bool Preprocessor::applyMacroOps(span<Token const> tokens, SmallVector<Token>& d
         // we find. Note that this can be left over at the end of applying ops; that's fine,
         // nothing is relying on observing this after the end of the macro's tokens.
         if (newToken.kind == TokenKind::EmptyMacroArgument) {
-            emptyArgTrivia.appendRange(newToken.trivia());
+            emptyArgTrivia.append_range(newToken.trivia());
             continue;
         }
 
         if (!emptyArgTrivia.empty()) {
-            emptyArgTrivia.appendRange(newToken.trivia());
+            emptyArgTrivia.append_range(newToken.trivia());
             newToken = newToken.withTrivia(alloc, emptyArgTrivia.copy(alloc));
             emptyArgTrivia.clear();
         }
 
         if (!stringify) {
-            if (syntheticComment)
-                commentBuffer.append(newToken);
-            else
-                dest.append(newToken);
+            if (syntheticComment) {
+                commentBuffer.push_back(newToken);
+            }
+            else {
+                dest.push_back(newToken);
+                if (extraToAppend) {
+                    dest.push_back(extraToAppend);
+                    extraToAppend = Token();
+                }
+            }
             continue;
         }
 
         // If this is an escaped identifier that includes a `" within it, we need to split the
         // token up to match the behavior of other simulators.
-        if (newToken.kind == TokenKind::Identifier && !newToken.rawText().empty() &&
-            newToken.rawText()[0] == '\\') {
-
-            size_t offset = newToken.rawText().find("`\"");
+        auto raw = newToken.rawText();
+        if (newToken.kind == TokenKind::Identifier && !raw.empty() && raw[0] == '\\') {
+            size_t offset = raw.find("`\"");
             if (offset != std::string_view::npos) {
-                // Split the token, finish the stringification.
-                Token split(alloc, TokenKind::Identifier, newToken.trivia(),
-                            newToken.rawText().substr(0, offset), newToken.location());
-                stringifyBuffer.append(split);
+                // Like above, we need to handle the case of mismatched delimiters.
+                // If we match, we complete the stringification. If we found a `"
+                // inside a triple quoted string, just keep going. If we found a `"""
+                // inside a single quoted string, split and end.
+                const bool startIsTriple = stringify.kind == TokenKind::MacroTripleQuote;
+                const bool endIsTriple = raw.substr(offset).starts_with(R"(`""")");
+                if (startIsTriple == endIsTriple || !startIsTriple) {
+                    stringifyBuffer.push_back(newToken.withRawText(alloc, raw.substr(0, offset)));
 
-                dest.append(Lexer::stringify(alloc, stringify.location(), stringify.trivia(),
-                                             stringifyBuffer.begin(), stringifyBuffer.end()));
-                stringify = Token();
+                    // Note: endToken parameter here doesn't matter,
+                    // we know there is no trivia to take.
+                    dest.push_back(
+                        Lexer::stringify(*lexerStack.back(), stringify, stringifyBuffer, Token()));
+                    stringify = Token();
 
-                // Now we have the unfortunate task of re-lexing the remaining stuff after the split
-                // and then appending those tokens to the destination as well.
-                SmallVectorSized<Token, 8> splits;
-                Lexer::splitTokens(alloc, diagnostics, sourceManager, newToken, offset + 2,
-                                   getCurrentKeywordVersion(), splits);
-                anyNewMacros |= applyMacroOps(splits, dest);
-                continue;
+                    // Now we have the unfortunate task of re-lexing the remaining stuff after the
+                    // split and then appending those tokens to the destination as well.
+                    SmallVector<Token, 8> splits;
+                    splitTokens(newToken, offset + (endIsTriple ? 4 : 2), splits);
+                    anyNewMacros |= applyMacroOps(splits, dest);
+                    continue;
+                }
             }
         }
 
-        stringifyBuffer.append(newToken);
+        stringifyBuffer.push_back(newToken);
     }
 
     if (stringify)
@@ -302,25 +355,25 @@ bool Preprocessor::expandMacro(MacroDef macro, MacroExpansion& expansion,
                                MacroActualArgumentListSyntax* actualArgs) {
     if (macro.isIntrinsic()) {
         // for now, no intrisics can have arguments
-        ASSERT(!actualArgs);
+        SLANG_ASSERT(!actualArgs);
         return expandIntrinsic(macro.intrinsic, expansion);
     }
 
     const DefineDirectiveSyntax* directive = macro.syntax;
-    ASSERT(directive);
+    SLANG_ASSERT(directive);
 
     // ignore empty macro
     const auto& body = directive->body;
     if (body.empty())
         return true;
 
-    string_view macroName = directive->name.valueText();
+    std::string_view macroName = directive->name.valueText();
 
     if (!directive->formalArguments) {
         // each macro expansion gets its own location entry
         SourceLocation start = body[0].location();
-        SourceLocation expansionLoc =
-            sourceManager.createExpansionLoc(start, expansion.getRange(), macroName);
+        SourceLocation expansionLoc = sourceManager.createExpansionLoc(start, expansion.getRange(),
+                                                                       macroName);
 
         // simple macro; just take body tokens
         for (auto token : body)
@@ -330,7 +383,7 @@ bool Preprocessor::expandMacro(MacroDef macro, MacroExpansion& expansion,
     }
 
     // match up actual arguments with formal parameters
-    ASSERT(actualArgs);
+    SLANG_ASSERT(actualArgs);
     auto& formalList = directive->formalArguments->args;
     auto& actualList = actualArgs->args;
     if (actualList.size() > formalList.size()) {
@@ -338,12 +391,12 @@ bool Preprocessor::expandMacro(MacroDef macro, MacroExpansion& expansion,
         return false;
     }
 
-    struct ArgTokens : public span<const Token> {
-        using span<const Token>::span;
-        using span<const Token>::operator=;
+    struct ArgTokens : public std::span<const Token> {
+        using std::span<const Token>::span;
+        using std::span<const Token>::operator=;
         bool isExpanded = false;
     };
-    SmallMap<string_view, ArgTokens, 8> argumentMap;
+    SmallMap<std::string_view, ArgTokens, 8> argumentMap;
 
     for (size_t i = 0; i < formalList.size(); i++) {
         auto formal = formalList[i];
@@ -374,8 +427,8 @@ bool Preprocessor::expandMacro(MacroDef macro, MacroExpansion& expansion,
                                endOfArgs.location() + endOfArgs.rawText().length());
 
     SourceLocation start = body[0].location();
-    SourceLocation expansionLoc =
-        sourceManager.createExpansionLoc(start, expansionRange, macroName);
+    SourceLocation expansionLoc = sourceManager.createExpansionLoc(start, expansionRange,
+                                                                   macroName);
 
     auto append = [&](Token token) {
         expansion.append(token, expansionLoc, start, expansionRange);
@@ -394,7 +447,7 @@ bool Preprocessor::expandMacro(MacroDef macro, MacroExpansion& expansion,
             return append(token);
         }
 
-        string_view text = token.valueText();
+        std::string_view text = token.valueText();
         if (token.kind == TokenKind::Directive && !text.empty()) {
             if (token.directiveKind() != SyntaxKind::MacroUsage) {
                 // If this is the start of a `define directive, note that fact because
@@ -419,7 +472,7 @@ bool Preprocessor::expandMacro(MacroDef macro, MacroExpansion& expansion,
         // Fully expand out arguments before substitution to make sure we can detect whether
         // a usage of a macro in a replacement list is valid or an illegal recursion.
         if (!it->second.isExpanded) {
-            span<const Token> argTokens = it->second;
+            std::span<const Token> argTokens = it->second;
             SmallSet<const DefineDirectiveSyntax*, 8> alreadyExpanded;
             if (!expandReplacementList(argTokens, alreadyExpanded))
                 return false;
@@ -505,10 +558,8 @@ bool Preprocessor::expandMacro(MacroDef macro, MacroExpansion& expansion,
                 if (!handleToken(first))
                     return false;
 
-                SmallVectorSized<Token, 8> splits;
-                Lexer::splitTokens(alloc, diagnostics, sourceManager, token, index,
-                                   getCurrentKeywordVersion(), splits);
-
+                SmallVector<Token, 8> splits;
+                splitTokens(token, index, splits);
                 for (auto t : splits) {
                     if (!handleToken(t))
                         return false;
@@ -517,8 +568,8 @@ bool Preprocessor::expandMacro(MacroDef macro, MacroExpansion& expansion,
                 // Add an empty argument in here so we can make sure a space ends
                 // the escaped identifier once it gets concatenated again.
                 if (!splits.empty()) {
-                    SmallVectorSized<Trivia, 2> triviaBuf;
-                    triviaBuf.emplace(TriviaKind::Whitespace, " "sv);
+                    SmallVector<Trivia> triviaBuf;
+                    triviaBuf.emplace_back(TriviaKind::Whitespace, " "sv);
 
                     auto loc = splits.back().location() + splits.back().rawText().length();
                     Token empty(alloc, TokenKind::EmptyMacroArgument, triviaBuf.copy(alloc), ""sv,
@@ -540,7 +591,7 @@ bool Preprocessor::expandMacro(MacroDef macro, MacroExpansion& expansion,
 }
 
 SourceRange Preprocessor::MacroExpansion::getRange() const {
-    return { usageSite.location(), usageSite.location() + usageSite.rawText().length() };
+    return {usageSite.location(), usageSite.location() + usageSite.rawText().length()};
 }
 
 SourceLocation Preprocessor::MacroExpansion::adjustLoc(Token token, SourceLocation& macroLoc,
@@ -576,23 +627,23 @@ void Preprocessor::MacroExpansion::append(Token token, SourceLocation location,
 
     // Line continuations get stripped out when we expand macros and become newline trivia instead.
     if (token.kind == TokenKind::LineContinuation && !allowLineContinuation) {
-        SmallVectorSized<Trivia, 8> newTrivia;
-        newTrivia.appendRange(token.trivia());
-        newTrivia.append(Trivia(TriviaKind::EndOfLine, token.rawText().substr(1)));
+        SmallVector<Trivia, 8> newTrivia;
+        newTrivia.append_range(token.trivia());
+        newTrivia.push_back(Trivia(TriviaKind::EndOfLine, token.rawText().substr(1)));
 
-        dest.append(
+        dest.push_back(
             Token(alloc, TokenKind::EmptyMacroArgument, newTrivia.copy(alloc), "", location));
     }
     else {
-        dest.append(token.withLocation(alloc, location));
+        dest.push_back(token.withLocation(alloc, location));
     }
 }
 
 bool Preprocessor::expandReplacementList(
-    span<Token const>& tokens, SmallSet<const DefineDirectiveSyntax*, 8>& alreadyExpanded) {
+    std::span<Token const>& tokens, SmallSet<const DefineDirectiveSyntax*, 8>& alreadyExpanded) {
 
-    SmallVectorSized<Token, 64> outBuffer;
-    SmallVectorSized<Token, 64> expansionBuffer;
+    SmallVector<Token, 16> outBuffer;
+    SmallVector<Token, 16> expansionBuffer;
 
     bool expandedSomething = false;
     MacroParser parser(*this);
@@ -602,7 +653,7 @@ bool Preprocessor::expandReplacementList(
     Token token;
     while ((token = parser.next())) {
         if (token.kind != TokenKind::Directive || token.directiveKind() != SyntaxKind::MacroUsage) {
-            outBuffer.append(token);
+            outBuffer.push_back(token);
             continue;
         }
 
@@ -611,7 +662,7 @@ bool Preprocessor::expandReplacementList(
         if (!macro.valid()) {
             // If we couldn't find the macro, just keep trucking.
             // It's possible that a future expansion will make this valid.
-            outBuffer.append(token);
+            outBuffer.push_back(token);
             continue;
         }
 
@@ -629,19 +680,19 @@ bool Preprocessor::expandReplacementList(
         }
 
         expansionBuffer.clear();
-        MacroExpansion expansion{ sourceManager, alloc, expansionBuffer, token, false };
+        MacroExpansion expansion{sourceManager, alloc, expansionBuffer, token, false};
         if (!expandMacro(macro, expansion, actualArgs))
             return false;
 
         // Recursively expand out nested macros; this ensures that we detect
         // any potentially recursive macros.
         alreadyExpanded.insert(macro.syntax);
-        span<const Token> expanded = expansionBuffer;
+        std::span<const Token> expanded = expansionBuffer;
         if (!expandReplacementList(expanded, alreadyExpanded))
             return false;
 
         alreadyExpanded.erase(macro.syntax);
-        outBuffer.appendRange(expanded);
+        outBuffer.append_range(expanded);
         expandedSomething = true;
     }
 
@@ -653,15 +704,15 @@ bool Preprocessor::expandReplacementList(
 
 bool Preprocessor::expandIntrinsic(MacroIntrinsic intrinsic, MacroExpansion& expansion) {
     auto loc = expansion.getRange().start();
-    SmallVectorSized<char, 64> text;
+    SmallVector<char> text;
     switch (intrinsic) {
         case MacroIntrinsic::File: {
-            string_view fileName = sourceManager.getFileName(loc);
-            text.append('"');
-            text.appendRange(fileName);
-            text.append('"');
+            std::string_view fileName = sourceManager.getFileName(loc);
+            text.push_back('"');
+            text.append_range(fileName);
+            text.push_back('"');
 
-            string_view rawText = toStringView(text.copy(alloc));
+            std::string_view rawText = toStringView(text.copy(alloc));
             Token token(alloc, TokenKind::StringLiteral, {}, rawText, loc, fileName);
             expansion.append(token, loc);
             break;
@@ -670,13 +721,13 @@ bool Preprocessor::expandIntrinsic(MacroIntrinsic intrinsic, MacroExpansion& exp
             size_t lineNum = sourceManager.getLineNumber(loc);
             uintToStr(text, static_cast<uint64_t>(lineNum));
 
-            string_view rawText = toStringView(text.copy(alloc));
+            std::string_view rawText = toStringView(text.copy(alloc));
             Token token(alloc, TokenKind::IntegerLiteral, {}, rawText, loc, lineNum);
             expansion.append(token, loc);
             break;
         }
         case MacroIntrinsic::None:
-            THROW_UNREACHABLE;
+            SLANG_UNREACHABLE;
     }
 
     return true;
@@ -687,13 +738,14 @@ bool Preprocessor::MacroDef::needsArgs() const {
 }
 
 MacroFormalArgumentListSyntax* Preprocessor::MacroParser::parseFormalArgumentList() {
-    // parse all formal arguments
     auto openParen = consume();
-    SmallVectorSized<TokenOrSyntax, 16> arguments;
-    parseArgumentList(arguments, [this]() { return parseFormalArgument(); });
+
+    Token closeParen;
+    SmallVector<TokenOrSyntax, 8> arguments;
+    parseArgumentList(arguments, [this]() { return parseFormalArgument(); }, closeParen);
 
     return pp.alloc.emplace<MacroFormalArgumentListSyntax>(openParen, arguments.copy(pp.alloc),
-                                                           expect(TokenKind::CloseParenthesis));
+                                                           closeParen);
 }
 
 MacroActualArgumentListSyntax* Preprocessor::MacroParser::parseActualArgumentList(Token prevToken) {
@@ -704,33 +756,31 @@ MacroActualArgumentListSyntax* Preprocessor::MacroParser::parseActualArgumentLis
     }
 
     auto openParen = consume();
-    SmallVectorSized<TokenOrSyntax, 16> arguments;
-    parseArgumentList(arguments, [this]() { return parseActualArgument(); });
+    Token closeParen;
+    SmallVector<TokenOrSyntax, 8> arguments;
+    parseArgumentList(arguments, [this]() { return parseActualArgument(); }, closeParen);
 
-    auto closeParen = expect(TokenKind::CloseParenthesis);
     return pp.alloc.emplace<MacroActualArgumentListSyntax>(openParen, arguments.copy(pp.alloc),
                                                            closeParen);
 }
 
 template<typename TFunc>
-void Preprocessor::MacroParser::parseArgumentList(SmallVector<TokenOrSyntax>& results,
-                                                  TFunc&& parseItem) {
+void Preprocessor::MacroParser::parseArgumentList(SmallVectorBase<TokenOrSyntax>& results,
+                                                  TFunc&& parseItem, Token& closeParen) {
     while (true) {
-        results.append(parseItem());
+        results.push_back(parseItem());
 
         if (peek().kind == TokenKind::Comma)
-            results.append(consume());
-        else {
-            // Just break out of the loop. Our caller will expect
-            // that there is a closing parenthesis here.
+            results.push_back(consume());
+        else
             break;
-        }
     }
+
+    closeParen = expect(TokenKind::CloseParenthesis);
 }
 
 MacroActualArgumentSyntax* Preprocessor::MacroParser::parseActualArgument() {
-    auto arg = parseTokenList(true);
-    return pp.alloc.emplace<MacroActualArgumentSyntax>(arg);
+    return pp.alloc.emplace<MacroActualArgumentSyntax>(parseTokenList(true));
 }
 
 MacroFormalArgumentSyntax* Preprocessor::MacroParser::parseFormalArgument() {
@@ -749,12 +799,13 @@ MacroFormalArgumentSyntax* Preprocessor::MacroParser::parseFormalArgument() {
     return pp.alloc.emplace<MacroFormalArgumentSyntax>(arg, argDef);
 }
 
-span<Token> Preprocessor::MacroParser::parseTokenList(bool allowNewlines) {
-    // comma and right parenthesis only end the default token list if they are
-    // not inside a nested pair of (), [], or {}
-    // otherwise, keep swallowing tokens as part of the default
-    SmallVectorSized<Token, 64> tokens;
-    SmallVectorSized<TokenKind, 16> delimPairStack;
+std::span<Token> Preprocessor::MacroParser::parseTokenList(bool allowNewlines) {
+    SmallVector<Token, 16> tokens;
+    SmallVector<TokenKind> delimPairStack;
+
+    // Comma and right parenthesis only end the default token list if they are
+    // not inside a nested pair of (), [], or {}, otherwise, keep swallowing
+    // tokens as part of the list.
     while (true) {
         auto kind = peek().kind;
         if (kind == TokenKind::EndOfFile || (!allowNewlines && !peek().isOnSameLine())) {
@@ -766,22 +817,23 @@ span<Token> Preprocessor::MacroParser::parseTokenList(bool allowNewlines) {
         }
 
         if (delimPairStack.empty()) {
-            if ((kind == TokenKind::Comma || kind == TokenKind::CloseParenthesis))
+            if (kind == TokenKind::Comma || kind == TokenKind::CloseParenthesis)
                 break;
         }
-        else if (delimPairStack.back() == kind)
-            delimPairStack.pop();
+        else if (delimPairStack.back() == kind) {
+            delimPairStack.pop_back();
+        }
 
-        tokens.append(consume());
+        tokens.push_back(consume());
 
         TokenKind closeKind = SyntaxFacts::getDelimCloseKind(kind);
         if (closeKind != TokenKind::Unknown)
-            delimPairStack.append(closeKind);
+            delimPairStack.push_back(closeKind);
     }
     return tokens.copy(pp.alloc);
 }
 
-void Preprocessor::MacroParser::setBuffer(span<Token const> newBuffer) {
+void Preprocessor::MacroParser::setBuffer(std::span<Token const> newBuffer) {
     buffer = newBuffer;
     currentIndex = 0;
 }
@@ -817,6 +869,11 @@ Token Preprocessor::MacroParser::expect(TokenKind kind) {
     return next();
 }
 
+void Preprocessor::splitTokens(Token sourceToken, size_t offset, SmallVectorBase<Token>& results) {
+    Lexer::splitTokens(alloc, diagnostics, sourceManager, sourceToken, offset,
+                       getCurrentKeywordVersion(), results);
+}
+
 static bool isSameToken(Token left, Token right) {
     if (left.kind != right.kind || left.rawText() != right.rawText())
         return false;
@@ -875,4 +932,4 @@ bool Preprocessor::isSameMacro(const DefineDirectiveSyntax& left,
     return isSameTokenList(left.body, right.body);
 }
 
-} // namespace slang
+} // namespace slang::parsing
