@@ -2,17 +2,21 @@
 // Token.cpp
 // Contains Token class and related helpers
 //
-// File is under the MIT license; see LICENSE for details
+// SPDX-FileCopyrightText: Michael Popoloski
+// SPDX-License-Identifier: MIT
 //------------------------------------------------------------------------------
 #include "slang/parsing/Token.h"
 
 #include "slang/diagnostics/ParserDiags.h"
 #include "slang/diagnostics/PreprocessorDiags.h"
 #include "slang/parsing/LexerFacts.h"
+#include "slang/syntax/SyntaxNode.h"
 #include "slang/syntax/SyntaxPrinter.h"
 #include "slang/util/BumpAllocator.h"
 
-namespace slang {
+namespace slang::parsing {
+
+using namespace syntax;
 
 // Heap-allocated info block. This structure is variably sized based on the
 // actual type of token. Type-specific data is stored at the end, followed
@@ -30,7 +34,7 @@ struct Token::Info {
     logic_t& bit() { return *reinterpret_cast<logic_t*>(extra()); }
     double& real() { return *reinterpret_cast<double*>(extra()); }
     SVIntStorage& integer() { return *reinterpret_cast<SVIntStorage*>(extra()); }
-    string_view& stringText() { return *reinterpret_cast<string_view*>(extra()); }
+    std::string_view& stringText() { return *reinterpret_cast<std::string_view*>(extra()); }
     SyntaxKind& directiveKind() { return *reinterpret_cast<SyntaxKind*>(extra()); }
 };
 
@@ -39,7 +43,7 @@ static constexpr size_t getExtraSize(TokenKind kind) {
     switch (kind) {
         case TokenKind::StringLiteral:
         case TokenKind::IncludeFileName:
-            size = sizeof(string_view);
+            size = sizeof(std::string_view);
             break;
         case TokenKind::RealLiteral:
         case TokenKind::TimeLiteral:
@@ -76,15 +80,15 @@ void NumericTokenFlags::setOutOfRange(bool value) {
     raw |= uint8_t(value) << 6;
 }
 
-Trivia::Trivia() : rawText{ "", 0 }, kind(TriviaKind::Unknown) {
+Trivia::Trivia() : rawText{"", 0}, kind(TriviaKind::Unknown) {
 }
 
-Trivia::Trivia(TriviaKind kind, string_view rawText) :
-    rawText{ rawText.data(), (uint32_t)rawText.size() }, kind(kind) {
+Trivia::Trivia(TriviaKind kind, std::string_view rawText) :
+    rawText{rawText.data(), (uint32_t)rawText.size()}, kind(kind) {
 }
 
-Trivia::Trivia(TriviaKind kind, span<Token const> tokens) :
-    tokens{ tokens.data(), (uint32_t)tokens.size() }, kind(kind) {
+Trivia::Trivia(TriviaKind kind, std::span<Token const> tokens) :
+    tokens{tokens.data(), (uint32_t)tokens.size()}, kind(kind) {
 }
 
 Trivia::Trivia(TriviaKind kind, SyntaxNode* syntax) : syntaxNode(syntax), kind(kind) {
@@ -109,13 +113,13 @@ Trivia Trivia::withLocation(BumpAllocator& alloc, SourceLocation location) const
     return result;
 }
 
-optional<SourceLocation> Trivia::getExplicitLocation() const {
+std::optional<SourceLocation> Trivia::getExplicitLocation() const {
     switch (kind) {
         case TriviaKind::Directive:
         case TriviaKind::SkippedSyntax:
             return syntaxNode->getFirstToken().location();
         case TriviaKind::SkippedTokens:
-            ASSERT(tokens.len);
+            SLANG_ASSERT(tokens.len);
             return tokens.ptr[0].location();
         default:
             if (hasFullLocation)
@@ -131,7 +135,7 @@ SyntaxNode* Trivia::syntax() const {
     return nullptr;
 }
 
-string_view Trivia::getRawText() const {
+std::string_view Trivia::getRawText() const {
     switch (kind) {
         case TriviaKind::Directive:
         case TriviaKind::SkippedSyntax:
@@ -140,66 +144,98 @@ string_view Trivia::getRawText() const {
         default:
             if (hasFullLocation)
                 return fullLocation->text;
-            return { rawText.ptr, rawText.len };
+            return {rawText.ptr, rawText.len};
     }
 }
 
-span<Token const> Trivia::getSkippedTokens() const {
+std::span<Token const> Trivia::getSkippedTokens() const {
     if (kind != TriviaKind::SkippedTokens)
         return {};
-    return { tokens.ptr, tokens.len };
+    return {tokens.ptr, tokens.len};
+}
+
+Trivia Trivia::clone(BumpAllocator& alloc, bool deep) const {
+    Trivia result;
+    result.kind = kind;
+    result.hasFullLocation = hasFullLocation;
+
+    switch (kind) {
+        case TriviaKind::Directive:
+        case TriviaKind::SkippedSyntax:
+            if (deep)
+                result.syntaxNode = syntax::deepClone(*syntaxNode, alloc);
+            else
+                result.syntaxNode = syntaxNode;
+            break;
+        case TriviaKind::SkippedTokens:
+            result.tokens = tokens;
+            break;
+        default:
+            if (hasFullLocation) {
+                result.fullLocation = alloc.emplace<FullLocation>();
+                result.fullLocation->text = fullLocation->text;
+                result.fullLocation->location = fullLocation->location;
+            }
+            else {
+                result.rawText = rawText;
+            }
+            break;
+    }
+
+    return result;
 }
 
 Token::Token() :
     kind(TokenKind::Unknown), missing(false), triviaCountSmall(0), reserved(0), numFlags() {
 }
 
-Token::Token(BumpAllocator& alloc, TokenKind kind, span<Trivia const> trivia, string_view rawText,
-             SourceLocation location) {
+Token::Token(BumpAllocator& alloc, TokenKind kind, std::span<Trivia const> trivia,
+             std::string_view rawText, SourceLocation location) {
     init(alloc, kind, trivia, rawText, location);
 }
 
-Token::Token(BumpAllocator& alloc, TokenKind kind, span<Trivia const> trivia, string_view rawText,
-             SourceLocation location, string_view strText) {
-    ASSERT(kind == TokenKind::StringLiteral || kind == TokenKind::IncludeFileName);
+Token::Token(BumpAllocator& alloc, TokenKind kind, std::span<Trivia const> trivia,
+             std::string_view rawText, SourceLocation location, std::string_view strText) {
+    SLANG_ASSERT(kind == TokenKind::StringLiteral || kind == TokenKind::IncludeFileName);
     init(alloc, kind, trivia, rawText, location);
     info->stringText() = strText;
 }
 
-Token::Token(BumpAllocator& alloc, TokenKind kind, span<Trivia const> trivia, string_view rawText,
-             SourceLocation location, SyntaxKind directive) {
-    ASSERT(kind == TokenKind::Directive || kind == TokenKind::MacroUsage);
+Token::Token(BumpAllocator& alloc, TokenKind kind, std::span<Trivia const> trivia,
+             std::string_view rawText, SourceLocation location, SyntaxKind directive) {
+    SLANG_ASSERT(kind == TokenKind::Directive || kind == TokenKind::MacroUsage);
     init(alloc, kind, trivia, rawText, location);
     info->directiveKind() = directive;
 }
 
-Token::Token(BumpAllocator& alloc, TokenKind kind, span<Trivia const> trivia, string_view rawText,
-             SourceLocation location, logic_t bit) {
-    ASSERT(kind == TokenKind::UnbasedUnsizedLiteral);
+Token::Token(BumpAllocator& alloc, TokenKind kind, std::span<Trivia const> trivia,
+             std::string_view rawText, SourceLocation location, logic_t bit) {
+    SLANG_ASSERT(kind == TokenKind::UnbasedUnsizedLiteral);
     init(alloc, kind, trivia, rawText, location);
     info->bit() = bit;
 }
 
-Token::Token(BumpAllocator& alloc, TokenKind kind, span<Trivia const> trivia, string_view rawText,
-             SourceLocation location, const SVInt& value) {
-    ASSERT(kind == TokenKind::IntegerLiteral);
+Token::Token(BumpAllocator& alloc, TokenKind kind, std::span<Trivia const> trivia,
+             std::string_view rawText, SourceLocation location, const SVInt& value) {
+    SLANG_ASSERT(kind == TokenKind::IntegerLiteral);
     init(alloc, kind, trivia, rawText, location);
 
     SVIntStorage storage(value.getBitWidth(), value.isSigned(), value.hasUnknown());
     if (value.isSingleWord())
         storage.val = *value.getRawPtr();
     else {
-        storage.pVal =
-            (uint64_t*)alloc.allocate(sizeof(uint64_t) * value.getNumWords(), alignof(uint64_t));
+        storage.pVal = (uint64_t*)alloc.allocate(sizeof(uint64_t) * value.getNumWords(),
+                                                 alignof(uint64_t));
         memcpy(storage.pVal, value.getRawPtr(), sizeof(uint64_t) * value.getNumWords());
     }
 
     info->integer() = storage;
 }
 
-Token::Token(BumpAllocator& alloc, TokenKind kind, span<Trivia const> trivia, string_view rawText,
-             SourceLocation location, double value, bool outOfRange, optional<TimeUnit> timeUnit) {
-    ASSERT(kind == TokenKind::RealLiteral || kind == TokenKind::TimeLiteral);
+Token::Token(BumpAllocator& alloc, TokenKind kind, std::span<Trivia const> trivia,
+             std::string_view rawText, SourceLocation location, double value, bool outOfRange,
+             std::optional<TimeUnit> timeUnit) {
+    SLANG_ASSERT(kind == TokenKind::RealLiteral || kind == TokenKind::TimeLiteral);
     init(alloc, kind, trivia, rawText, location);
     info->real() = value;
 
@@ -208,18 +244,18 @@ Token::Token(BumpAllocator& alloc, TokenKind kind, span<Trivia const> trivia, st
         numFlags.set(*timeUnit);
 }
 
-Token::Token(BumpAllocator& alloc, TokenKind kind, span<Trivia const> trivia, string_view rawText,
-             SourceLocation location, LiteralBase base, bool isSigned) {
+Token::Token(BumpAllocator& alloc, TokenKind kind, std::span<Trivia const> trivia,
+             std::string_view rawText, SourceLocation location, LiteralBase base, bool isSigned) {
     init(alloc, kind, trivia, rawText, location);
     numFlags.set(base, isSigned);
 }
 
-string_view Token::valueText() const {
+std::string_view Token::valueText() const {
     switch (kind) {
         case TokenKind::StringLiteral:
             return info->stringText();
         case TokenKind::Identifier: {
-            string_view result = rawText();
+            std::string_view result = rawText();
             if (!result.empty() && result[0] == '\\')
                 result = result.substr(1);
             return result;
@@ -229,14 +265,13 @@ string_view Token::valueText() const {
     }
 }
 
-string_view Token::rawText() const {
-    string_view text = LexerFacts::getTokenKindText(kind);
+std::string_view Token::rawText() const {
+    std::string_view text = LexerFacts::getTokenKindText(kind);
     if (!text.empty())
         return text;
 
     // not a simple token, so extract info from our data pointer
     switch (kind) {
-        case TokenKind::Unknown:
         case TokenKind::Identifier:
         case TokenKind::SystemIdentifier:
         case TokenKind::IncludeFileName:
@@ -250,12 +285,16 @@ string_view Token::rawText() const {
         case TokenKind::MacroUsage:
         case TokenKind::EmptyMacroArgument:
         case TokenKind::LineContinuation:
-            return string_view(info->rawTextPtr, rawLen);
+            return std::string_view(info->rawTextPtr, rawLen);
+        case TokenKind::Unknown:
+            if (info)
+                return std::string_view(info->rawTextPtr, rawLen);
+            return "";
         case TokenKind::Placeholder:
         case TokenKind::EndOfFile:
             return "";
         default:
-            THROW_UNREACHABLE;
+            SLANG_UNREACHABLE;
     }
 }
 
@@ -264,24 +303,26 @@ SourceRange Token::range() const {
 }
 
 SourceLocation Token::location() const {
+    if (!info)
+        return SourceLocation::NoLocation;
     return info->location;
 }
 
-span<Trivia const> Token::trivia() const {
+std::span<Trivia const> Token::trivia() const {
     if (triviaCountSmall == 0)
         return {};
 
     const Trivia* trivia;
     byte* ptr = info->extra() + getExtraSize(kind);
-    memcpy(&trivia, ptr, sizeof(trivia));
+    memcpy(reinterpret_cast<void*>(&trivia), ptr, sizeof(trivia));
 
     if (triviaCountSmall == MaxTriviaSmallCount + 1) {
         size_t size;
         memcpy(&size, ptr + sizeof(trivia), sizeof(size_t));
-        return { trivia, size };
+        return {trivia, size};
     }
 
-    return { trivia, triviaCountSmall };
+    return {trivia, triviaCountSmall};
 }
 
 std::string Token::toString() const {
@@ -289,28 +330,28 @@ std::string Token::toString() const {
 }
 
 SVInt Token::intValue() const {
-    ASSERT(kind == TokenKind::IntegerLiteral);
+    SLANG_ASSERT(kind == TokenKind::IntegerLiteral);
     return info->integer();
 }
 
 double Token::realValue() const {
-    ASSERT(kind == TokenKind::RealLiteral || kind == TokenKind::TimeLiteral);
+    SLANG_ASSERT(kind == TokenKind::RealLiteral || kind == TokenKind::TimeLiteral);
     return info->real();
 }
 
 logic_t Token::bitValue() const {
-    ASSERT(kind == TokenKind::UnbasedUnsizedLiteral);
+    SLANG_ASSERT(kind == TokenKind::UnbasedUnsizedLiteral);
     return info->bit();
 }
 
 NumericTokenFlags Token::numericFlags() const {
-    ASSERT(kind == TokenKind::IntegerBase || kind == TokenKind::TimeLiteral ||
-           kind == TokenKind::RealLiteral);
+    SLANG_ASSERT(kind == TokenKind::IntegerBase || kind == TokenKind::TimeLiteral ||
+                 kind == TokenKind::RealLiteral);
     return numFlags;
 }
 
 SyntaxKind Token::directiveKind() const {
-    ASSERT(kind == TokenKind::Directive || kind == TokenKind::MacroUsage);
+    SLANG_ASSERT(kind == TokenKind::Directive || kind == TokenKind::MacroUsage);
     return info->directiveKind();
 }
 
@@ -340,7 +381,7 @@ bool Token::isOnSameLine() const {
     return true;
 }
 
-Token Token::withTrivia(BumpAllocator& alloc, span<Trivia const> trivia) const {
+Token Token::withTrivia(BumpAllocator& alloc, std::span<Trivia const> trivia) const {
     return clone(alloc, trivia, rawText(), location());
 }
 
@@ -348,11 +389,11 @@ Token Token::withLocation(BumpAllocator& alloc, SourceLocation location) const {
     return clone(alloc, trivia(), rawText(), location);
 }
 
-Token Token::withRawText(BumpAllocator& alloc, string_view rawText) const {
+Token Token::withRawText(BumpAllocator& alloc, std::string_view rawText) const {
     return clone(alloc, trivia(), rawText, location());
 }
 
-Token Token::clone(BumpAllocator& alloc, span<Trivia const> trivia, string_view rawText,
+Token Token::clone(BumpAllocator& alloc, std::span<Trivia const> trivia, std::string_view rawText,
                    SourceLocation location) const {
     Token result(alloc, kind, trivia, rawText, location);
     result.missing = missing;
@@ -363,8 +404,21 @@ Token Token::clone(BumpAllocator& alloc, span<Trivia const> trivia, string_view
     return result;
 }
 
-void Token::init(BumpAllocator& alloc, TokenKind kind_, span<Trivia const> trivia,
-                 string_view rawText, SourceLocation location) {
+Token Token::deepClone(BumpAllocator& alloc) const {
+    if (!info) {
+        // No extra information, don't alloc extra info
+        // If allocated it, the valid() function would fail
+        return *this;
+    }
+
+    SmallVector<Trivia> triviaBuffer(trivia().size(), UninitializedTag());
+    for (const auto& t : trivia())
+        triviaBuffer.push_back(t.clone(alloc, true));
+    return clone(alloc, triviaBuffer.copy(alloc), rawText(), location());
+}
+
+void Token::init(BumpAllocator& alloc, TokenKind kind_, std::span<Trivia const> trivia,
+                 std::string_view rawText, SourceLocation location) {
     kind = kind_;
     missing = false;
     triviaCountSmall = 0;
@@ -373,7 +427,7 @@ void Token::init(BumpAllocator& alloc, TokenKind kind_, span<Trivia const> trivi
     rawLen = uint32_t(rawText.size());
 
     size_t extra = getExtraSize(kind);
-    ASSERT(extra % alignof(void*) == 0);
+    SLANG_ASSERT(extra % alignof(void*) == 0);
 
     size_t size = sizeof(Info) + extra;
     if (!trivia.empty()) {
@@ -394,7 +448,7 @@ void Token::init(BumpAllocator& alloc, TokenKind kind_, span<Trivia const> trivi
     if (!trivia.empty()) {
         const Trivia* triviaPtr = trivia.data();
         byte* dest = info->extra() + extra;
-        memcpy(dest, &triviaPtr, sizeof(triviaPtr));
+        memcpy(dest, reinterpret_cast<const void*>(&triviaPtr), sizeof(triviaPtr));
 
         if (trivia.size() > MaxTriviaSmallCount) {
             size = trivia.size();
@@ -491,4 +545,4 @@ Token Token::createExpected(BumpAllocator& alloc, Diagnostics& diagnostics, Toke
     return Token::createMissing(alloc, expected, location);
 }
 
-} // namespace slang
+} // namespace slang::parsing
