@@ -2,15 +2,16 @@
 // SyntaxFacts.cpp
 // Various syntax-related utility methods
 //
-// File is under the MIT license; see LICENSE for details
+// SPDX-FileCopyrightText: Michael Popoloski
+// SPDX-License-Identifier: MIT
 //------------------------------------------------------------------------------
 #include "slang/syntax/SyntaxFacts.h"
 
 #include "slang/syntax/AllSyntax.h"
 
-namespace slang {
+namespace slang::syntax {
 
-string_view SyntaxFacts::getSimpleTypeName(const DataTypeSyntax& syntax) {
+std::string_view SyntaxFacts::getSimpleTypeName(const DataTypeSyntax& syntax) {
     if (syntax.kind == SyntaxKind::NamedType) {
         auto& namedType = syntax.as<NamedTypeSyntax>();
         if (namedType.name->kind == SyntaxKind::IdentifierName)
@@ -420,7 +421,7 @@ bool SyntaxFacts::isPossibleStatement(TokenKind kind) {
         case TokenKind::RandSequenceKeyword:
         case TokenKind::RandCaseKeyword:
         case TokenKind::ExpectKeyword:
-        case TokenKind::OpenParenthesisStar:
+        case TokenKind::OpenParenthesis:
         case TokenKind::Semicolon:
             return true;
         default:
@@ -490,7 +491,7 @@ SyntaxKind SyntaxFacts::getModuleDeclarationKind(TokenKind kind) {
     }
 }
 
-TokenKind SyntaxFacts::getModuleEndKind(TokenKind kind) {
+parsing::TokenKind SyntaxFacts::getModuleEndKind(TokenKind kind) {
     switch (kind) {
         case TokenKind::ModuleKeyword: return TokenKind::EndModuleKeyword;
         case TokenKind::MacromoduleKeyword: return TokenKind::EndModuleKeyword;
@@ -501,24 +502,22 @@ TokenKind SyntaxFacts::getModuleEndKind(TokenKind kind) {
     }
 }
 
-TokenKind SyntaxFacts::getDelimCloseKind(TokenKind kind) {
+parsing::TokenKind SyntaxFacts::getDelimCloseKind(TokenKind kind) {
     switch (kind) {
         case TokenKind::OpenParenthesis: return TokenKind::CloseParenthesis;
         case TokenKind::OpenBrace: return TokenKind::CloseBrace;
         case TokenKind::OpenBracket: return TokenKind::CloseBracket;
         case TokenKind::ApostropheOpenBrace: return TokenKind::CloseBrace;
-        case TokenKind::OpenParenthesisStar: return TokenKind::StarCloseParenthesis;
         default: return TokenKind::Unknown;
     }
 }
 
-TokenKind SyntaxFacts::getSkipToKind(TokenKind kind) {
+parsing::TokenKind SyntaxFacts::getSkipToKind(TokenKind kind) {
     switch (kind) {
         case TokenKind::OpenParenthesis: return TokenKind::CloseParenthesis;
         case TokenKind::OpenBrace: return TokenKind::CloseBrace;
         case TokenKind::OpenBracket: return TokenKind::CloseBracket;
         case TokenKind::ApostropheOpenBrace: return TokenKind::CloseBrace;
-        case TokenKind::OpenParenthesisStar: return TokenKind::StarCloseParenthesis;
         case TokenKind::BeginKeyword: return TokenKind::EndKeyword;
         case TokenKind::CaseKeyword: return TokenKind::EndCaseKeyword;
         case TokenKind::CheckerKeyword: return TokenKind::EndCheckerKeyword;
@@ -642,7 +641,7 @@ bool SyntaxFacts::isPossibleAnsiPort(TokenKind kind) {
         case TokenKind::InOutKeyword:
         case TokenKind::RefKeyword:
         case TokenKind::VarKeyword:
-        case TokenKind::OpenParenthesisStar:
+        case TokenKind::OpenParenthesis:
             return true;
         default:
             return isNetType(kind) || isPossibleDataType(kind);
@@ -656,7 +655,7 @@ bool SyntaxFacts::isPossibleUdpPort(TokenKind kind) {
         case TokenKind::OutputKeyword:
         case TokenKind::RegKeyword:
         case TokenKind::Comma:
-        case TokenKind::OpenParenthesisStar:
+        case TokenKind::OpenParenthesis:
             return true;
         default:
             return false;
@@ -665,7 +664,7 @@ bool SyntaxFacts::isPossibleUdpPort(TokenKind kind) {
 
 bool SyntaxFacts::isPossibleModportPort(TokenKind kind) {
     switch (kind) {
-        case TokenKind::OpenParenthesisStar:
+        case TokenKind::OpenParenthesis:
         case TokenKind::InputKeyword:
         case TokenKind::OutputKeyword:
         case TokenKind::InOutKeyword:
@@ -690,7 +689,8 @@ bool SyntaxFacts::isPossibleFunctionPort(TokenKind kind) {
         case TokenKind::RefKeyword:
         case TokenKind::VarKeyword:
         case TokenKind::ConstKeyword:
-        case TokenKind::OpenParenthesisStar:
+        case TokenKind::OpenParenthesis:
+        case TokenKind::DefaultKeyword:
             return true;
         default:
             return isPossibleDataType(kind);
@@ -737,10 +737,11 @@ bool SyntaxFacts::isPossibleForInitializer(TokenKind kind) {
     return kind == TokenKind::Comma || kind == TokenKind::VarKeyword || isPossibleExpression(kind);
 }
 
-bool SyntaxFacts::isPossibleOpenRangeElement(TokenKind kind) {
+bool SyntaxFacts::isPossibleValueRangeElement(TokenKind kind) {
     switch (kind) {
         case TokenKind::OpenBracket:
         case TokenKind::Comma:
+        case TokenKind::DefaultKeyword:
             return true;
         default:
             return isPossibleExpression(kind);
@@ -750,9 +751,10 @@ bool SyntaxFacts::isPossibleOpenRangeElement(TokenKind kind) {
 bool SyntaxFacts::isPossiblePattern(TokenKind kind) {
     switch (kind) {
         case TokenKind::Dot:
-        case TokenKind::DotStar:
         case TokenKind::ApostropheOpenBrace:
             return true;
+        case TokenKind::TripleAnd:
+            return false;
         default:
             return isPossibleExpression(kind);
     }
@@ -831,8 +833,7 @@ bool SyntaxFacts::isPossibleParameter(TokenKind kind) {
 
 bool SyntaxFacts::isPossiblePortConnection(TokenKind kind) {
     switch (kind) {
-        case TokenKind::OpenParenthesisStar:
-        case TokenKind::DotStar:
+        case TokenKind::OpenParenthesis:
         case TokenKind::Dot:
         case TokenKind::Comma:
             return true;
@@ -892,7 +893,6 @@ bool SyntaxFacts::isCloseDelimOrKeyword(TokenKind kind) {
         case TokenKind::CloseBrace:
         case TokenKind::CloseBracket:
         case TokenKind::CloseParenthesis:
-        case TokenKind::StarCloseParenthesis:
             return true;
         default:
             return isEndKeyword(kind);
@@ -1081,7 +1081,8 @@ bool SyntaxFacts::isEndOfConditionalPredicate(TokenKind kind) {
 
 bool SyntaxFacts::isEndOfAttribute(TokenKind kind) {
     switch (kind) {
-        case TokenKind::StarCloseParenthesis:
+        case TokenKind::Star:
+        case TokenKind::CloseParenthesis:
             // these indicate a missing *) somewhere
         case TokenKind::Semicolon:
         case TokenKind::PrimitiveKeyword:
@@ -1147,7 +1148,7 @@ bool SyntaxFacts::isNotInConcatenationExpr(TokenKind kind) {
 }
 
 bool SyntaxFacts::isPossibleLetPortItem(TokenKind kind) {
-    return kind == TokenKind::OpenParenthesisStar || kind == TokenKind::UntypedKeyword || isPossibleDataType(kind);
+    return kind == TokenKind::OpenParenthesis || kind == TokenKind::UntypedKeyword || isPossibleDataType(kind);
 }
 
 bool SyntaxFacts::isNotInParameterList(TokenKind kind) {
@@ -1156,7 +1157,7 @@ bool SyntaxFacts::isNotInParameterList(TokenKind kind) {
 
 bool SyntaxFacts::isPossiblePropertyPortItem(TokenKind kind) {
     switch (kind) {
-        case TokenKind::OpenParenthesisStar:
+        case TokenKind::OpenParenthesis:
         case TokenKind::LocalKeyword:
         case TokenKind::PropertyKeyword:
         case TokenKind::SequenceKeyword:
@@ -1229,7 +1230,6 @@ static bool isModuleOrPackageDecl(SyntaxKind kind) {
         case SyntaxKind::NetTypeDeclaration:
         case SyntaxKind::TypedefDeclaration:
         case SyntaxKind::ForwardTypedefDeclaration:
-        case SyntaxKind::ForwardInterfaceClassTypedefDeclaration:
         case SyntaxKind::PackageImportDeclaration:
         case SyntaxKind::DataDeclaration:
         case SyntaxKind::TaskDeclaration:
@@ -1370,11 +1370,25 @@ bool SyntaxFacts::isAllowedInProgram(SyntaxKind kind) {
     }
 }
 
+bool SyntaxFacts::isAllowedInAnonymousProgram(SyntaxKind kind) {
+    switch (kind) {
+        case SyntaxKind::TaskDeclaration:
+        case SyntaxKind::FunctionDeclaration:
+        case SyntaxKind::ClassDeclaration:
+        case SyntaxKind::CovergroupDeclaration:
+        case SyntaxKind::EmptyMember:
+            return true;
+        default:
+            return false;
+    }
+}
+
 bool SyntaxFacts::isAllowedInPackage(SyntaxKind kind) {
     switch (kind) {
         case SyntaxKind::TimeUnitsDeclaration:
         case SyntaxKind::PackageExportDeclaration:
         case SyntaxKind::PackageExportAllDeclaration:
+        case SyntaxKind::AnonymousProgram:
             return true;
         default:
             return isModuleOrPackageDecl(kind);
@@ -1422,7 +1436,6 @@ bool SyntaxFacts::isAllowedInChecker(SyntaxKind kind) {
         case SyntaxKind::NetTypeDeclaration:
         case SyntaxKind::TypedefDeclaration:
         case SyntaxKind::ForwardTypedefDeclaration:
-        case SyntaxKind::ForwardInterfaceClassTypedefDeclaration:
         case SyntaxKind::PackageImportDeclaration:
         case SyntaxKind::DataDeclaration:
         case SyntaxKind::FunctionDeclaration:
@@ -1462,6 +1475,28 @@ bool SyntaxFacts::isStrength1(TokenKind kind) {
     }
 }
 
+bool SyntaxFacts::isAssignmentOperator(SyntaxKind kind) {
+    switch (kind) {
+        case SyntaxKind::AssignmentExpression:
+        case SyntaxKind::AddAssignmentExpression:
+        case SyntaxKind::SubtractAssignmentExpression:
+        case SyntaxKind::MultiplyAssignmentExpression:
+        case SyntaxKind::DivideAssignmentExpression:
+        case SyntaxKind::ModAssignmentExpression:
+        case SyntaxKind::AndAssignmentExpression:
+        case SyntaxKind::OrAssignmentExpression:
+        case SyntaxKind::XorAssignmentExpression:
+        case SyntaxKind::LogicalLeftShiftAssignmentExpression:
+        case SyntaxKind::LogicalRightShiftAssignmentExpression:
+        case SyntaxKind::ArithmeticLeftShiftAssignmentExpression:
+        case SyntaxKind::ArithmeticRightShiftAssignmentExpression:
+        case SyntaxKind::NonblockingAssignmentExpression:
+            return true;
+        default:
+            return false;
+    }
+}
+
 // clang-format on
 
-} // namespace slang
+} // namespace slang::syntax
