@@ -2,15 +2,19 @@
 // SyntaxPrinter.cpp
 // Support for printing syntax nodes and tokens
 //
-// File is under the MIT license; see LICENSE for details
+// SPDX-FileCopyrightText: Michael Popoloski
+// SPDX-License-Identifier: MIT
 //------------------------------------------------------------------------------
 #include "slang/syntax/SyntaxPrinter.h"
 
-#include "slang/parsing/Parser.h"
+#include "slang/parsing/ParserMetadata.h"
+#include "slang/syntax/SyntaxNode.h"
 #include "slang/syntax/SyntaxTree.h"
 #include "slang/text/SourceManager.h"
 
-namespace slang {
+namespace slang::syntax {
+
+using namespace parsing;
 
 SyntaxPrinter::SyntaxPrinter(const SourceManager& sourceManager) : sourceManager(&sourceManager) {
 }
@@ -21,7 +25,8 @@ SyntaxPrinter& SyntaxPrinter::print(Trivia trivia) {
             if (includeDirectives)
                 print(*trivia.syntax());
             else if (includePreprocessed) {
-                for (const auto& t : trivia.syntax()->getFirstToken().trivia())
+                auto nestedTrivia = trivia.syntax()->getFirstToken().trivia();
+                for (const auto& t : nestedTrivia)
                     print(t);
             }
             break;
@@ -35,6 +40,10 @@ SyntaxPrinter& SyntaxPrinter::print(Trivia trivia) {
                     print(t);
             }
             break;
+        case TriviaKind::DisabledText:
+            if (includeSkipped)
+                append(trivia.getRawText());
+            break;
         case TriviaKind::LineComment:
         case TriviaKind::BlockComment:
             if (!includeComments)
@@ -61,9 +70,9 @@ SyntaxPrinter& SyntaxPrinter::print(Token token) {
             // Exclude any trivia that is from a preprocessed location as well. In order
             // to know that we need to skip over any trivia that is implicitly located
             // relative to something ahead of it (a directive or the token itself).
-            SmallVectorSized<const Trivia*, 8> pending;
+            SmallVector<const Trivia*> pending;
             for (const auto& trivia : token.trivia()) {
-                pending.append(&trivia);
+                pending.push_back(&trivia);
                 auto loc = trivia.getExplicitLocation();
                 if (loc) {
                     if (!sourceManager->isPreprocessedLoc(*loc)) {
@@ -111,11 +120,12 @@ std::string SyntaxPrinter::printFile(const SyntaxTree& tree) {
         .setIncludeSkipped(true)
         .setIncludeTrivia(true)
         .setIncludePreprocessed(false)
+        .setSquashNewlines(false)
         .print(tree)
         .str();
 }
 
-SyntaxPrinter& SyntaxPrinter::append(string_view text) {
+SyntaxPrinter& SyntaxPrinter::append(std::string_view text) {
     if (!squashNewlines) {
         buffer.append(text);
         return *this;
@@ -155,4 +165,4 @@ SyntaxPrinter& SyntaxPrinter::append(string_view text) {
     return *this;
 }
 
-} // namespace slang
+} // namespace slang::syntax
